/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: appio.c
  function: Application top level
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 13 1996

 ********************************************************************/

#include "libogg.h"
#include "oggapp.h"

int OGG_verbose=0;

void reporterror(char *s){
  fprintf(stderr,"\nERROR: %s\n",s);
  exit(-1);
}

void flatreport(char *s){
  if(OGG_verbose){
    fprintf(stderr,"%s",s);
    fflush(stderr);
  }
}

void report(char *s){
  flatreport(s);
  flatreport("\n");
}

char *static_buffer=NULL;      /* When we read headers from a       */
long static_buffer_size=0;     /* non-seekable stream, more than one */
long static_buffer_alloced=0;  /* function will need to check the same
			       data; we can't put it back, so we need
			       to buffer it */

/* static buffer hack maint */
/* clean pipe behavior */

long blocking_read(char *buffer, long size, long num, FILE *inf){
  long words=0;

  while (words<num){
    if(feof(inf))break;
    errno=0;
    words+=fread(buffer+words,size,num-words,inf);
    if(errno){
      report("\nERROR while reading input stream:");
      reporterror(strerror(errno));
    }
    if(OGG_verbose)
      if(words<num && !feof(inf))
	flatreport("\r=[ Waiting for input ]==>");
      
  }
  return(words);
}

long blocking_write(char *buffer, long size, long num, FILE *outf){
  long words=0;

  while (words<num){
    if(feof(outf))break;
    errno=0;
    words+=fwrite(buffer+words,size,num-words,outf);
    if(errno){
      report("\nERROR while writing output stream:");
      reporterror(strerror(errno));
    }
    if(OGG_verbose)
      if(words<num && !feof(outf))
	flatreport("\r==[ Waiting to write ]==>");
      
  }
  return(words);
}

void static_buffer_verify(long bytes){
  if(bytes>static_buffer_alloced){
    if(static_buffer_alloced>0)
      static_buffer=(char *)realloc(static_buffer,bytes);
    else
      static_buffer=(char *)malloc(bytes);
    if(static_buffer==NULL)reporterror("Unable to malloc.");
    static_buffer_alloced=bytes;
  }
}

void static_buffer_read(FILE *f,long bytes){
  if(static_buffer_size==0 && bytes && feof(f))
    reporterror("Unexpected end of data!");
  if(bytes>static_buffer_size){
    int readb;

    /* alloc more space if necessary */
    static_buffer_verify(bytes);

    /* read data into buffer */
    readb=blocking_read(static_buffer+static_buffer_size,1,
			bytes-static_buffer_size,f);
    static_buffer_size+=readb;
  }
}

void static_buffer_shift(long bytes){
  if(bytes>=static_buffer_size)
    static_buffer_size=0;
  else{
    memmove(static_buffer,static_buffer+bytes,static_buffer_size-bytes);
    static_buffer_size-=bytes;
  }
}

