/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: header.c
  function: Read/Write various PCM format headers
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 08 1996

 ********************************************************************/

#include "oggapp.h"
#include "appio.h"
#define MAXWAV 1024  /* A WAV header better not be more
			than 1k :) */
extern long bytesout;
extern long bytesin;

long GetNum(char *s,int endianness,int bytes){
  long ret;
  if(!endianness)s+=bytes-1;
  ret=*s;
  while(--bytes){
    if(endianness)
      s++;
    else
      s--;
    ret<<=8;
    ret|=(*s)&0xff;
  }
  return(ret);
}

void PutNum(long num,FILE *f,int endianness,int bytes){
  int i;
  if(!endianness)
    i=0;
  else
    i=bytes-1;
  while(bytes--){
    if(fputc((num>>(i<<3))&0xff,f)==EOF)
      reporterror("Could not write to output.");
    if(endianness)
      i--;
    else
      i++;
  }
}

int CheckWav(FILE *infile,long *channels,long *rate,long *bits,
	     long *bigendian,long *osigned,long *end){
  
  /* Do it clumsy-like until Mike gets his lib version of this
     finished. */
  
  static_buffer_read(infile,MAXWAV); 
  if(static_buffer_size<40)
    reporterror("Unexpected end of file!");

  /* Izzit Da Shoez? */    
  if(!memcmp("RIFF",static_buffer,4) && !memcmp("WAVE",static_buffer+8,4)){
    /* It's a WAV file */
    int ptr=12;
    report("Input is a RIFF/WAVE bitstream.");
    /* Look for the format chunk; none of the other chunks are worth
       reading */
    while(ptr<static_buffer_size-30){
      if(!strncmp("fmt",static_buffer+ptr,3)){
	/* OK, this is our audio specs chunk.  Slurp it up. */
	ptr+=8;  /* past the chunk bytes */

	/* Check that this is PCM */
	if(memcmp(static_buffer+ptr,"\001\000",2)) 
	  /* The data is stored as little endian :) */
	  /* This is not a PCM file. */
	  reporterror("This WAV file contains a proprietary data format.\n"
		      "Can't continue reading.\n");
	
	ptr+=2;
	*channels=GetNum(static_buffer+ptr,0,2);
	ptr+=2;
	*rate=GetNum(static_buffer+ptr,0,4);
	ptr+=4;
	/* Don't need 'ave bytes' or 'block size' for most files */
	ptr+=6;
	/* bits per sample */
	*bits=GetNum(static_buffer+ptr,0,2);
	ptr+=2;
	*bigendian=0; /* WAVE is for little endian PCs */
	if(*bits<9)
	  *osigned=0;
	else
	  *osigned=-1;  /* WAVE is always signed when >8 bits*/
	
	/* Now, align things to the beginning of the data */
	/* Look for 'dataxxxx' */
	while(ptr<static_buffer_size-8){
	  if(!strncmp("data",static_buffer+ptr,4)){
	    ptr+=4;
	    /* We're there.  How big? */
	    *end=GetNum(static_buffer+ptr,0,4);
	    ptr+=4;
	    *end/=*channels;
	    *end/=((*bits)-1)/8+1;
	    static_buffer_read(infile,ptr);
	    static_buffer_shift(ptr);
	    bytesin+=ptr;

	    /* OK, it's incorrect to assume only one chunk.  Mike can
               fix this later. */

	    return(-1);
	  }
	  ptr++;
	}
	/* No data? */
	reporterror("Cannot find beginning of data in WAVE file.\n");
      }
      ptr++;
    }
    /* Well then.  We can't read it */
    reporterror("File has a WAVE header, but is missing the\n"
		"required audio format chunk.");
  }
			
  /* Well, then, it's not a WAVE file */
  return(0);
}

long header_seekpos=0;
void WriteWav(FILE *f,long channels,long rate,long bits,
	     int *bigendian,int *osigned){

  header_seekpos=ftell(f);
  /* Again, quick and dirty.  To be replaced by Mike */

  fprintf(f,"RIFF    WAVEfmt ");
  PutNum(16,f,0,4);
  PutNum(1,f,0,2);
  PutNum(channels,f,0,2);
  PutNum(rate,f,0,4);
  PutNum(rate*channels*((bits-1)/8+1),f,0,4);
  PutNum(((bits-1)/8+1)*channels,f,0,2);
  PutNum(bits,f,0,2);
  fprintf(f,"data    ");
  *bigendian=0;
  if(bits>8)
    *osigned=-1;
  else
    *osigned=0;

  /* Have to finish setting the size fields later :( */
}

void FinishWav(FILE *f,long channels,long rate,long bits){
  long size=ftell(f)-header_seekpos;
  if(fseek(f,header_seekpos+4,SEEK_SET))
    reporterror("Couldn't seek to finish the .WAV header!");
  PutNum(size-8,f,0,4);
  
  if(fseek(f,header_seekpos+40,SEEK_SET))
    reporterror("Couldn't seek to finish the .WAV header!");
  PutNum(size-44,f,0,4);
}

/* bigendian IEEE-754 80 bit float to double */
/* Double is usually 64 bits tho */

double IEEE80toDOUBLE(char *stream){
  int s=stream[0]&0x80;
  int e=((stream[0]&0x7f)<<8)|(stream[1]&0xff);
  double f=((unsigned long)(stream[2]&0xff)<<24)|
    ((stream[3]&0xff)<<16)|
      ((stream[4]&0xff)<<8)|
	(stream[5]&0xff);

  if(e==32767)
    if(stream[2]&0x80)
#ifdef NAN
      return(NAN);
#else
      reporterror("Can't handle NaN!");
#endif
    else
      if(s)
	return(-HUGE_VAL);
      else
	return(HUGE_VAL);

  f=ldexp(f,32);
  f+=((stream[6]&0xff)<<24)|
    ((stream[7]&0xff)<<16)|
      ((stream[8]&0xff)<<8)|
	(stream[9]&0xff);
  return(ldexp(f,e-16446));
}
	  
/* To deal with Apple silliness... */
void PutFloat10(long num,FILE *f){
  char buffer[10];
  long shifts=0,temp;
  memset(buffer,0,10);

  if(num<0){
    buffer[0]='\200';
    num=-num;
  }
  temp=num;

  while(temp>1){
    shifts++;
    temp>>=1;
  }
  
  buffer[1]=(char)shifts;
  num<<=31-shifts;
  buffer[2]=(num>>24)&0xff;
  buffer[3]=(num>>16)&0xff;
  buffer[4]=(num>>8)&0xff;
  buffer[5]=num&0xff;
  blocking_write(buffer,1,10,f);
}
	  

int CheckAiff(FILE *infile,long *channels,long *rate,long *bits,
	     long *bigendian,long *osigned,long *end){
  /* Do both AIFF and AIFC here.  We can't do the AIFC compression
     tho; Mike is doing that too. */
  int aifc=0;

  static_buffer_read(infile,MAXWAV);
  if(static_buffer_size<30)
    reporterror("Unexpected end of file!");
  if(!memcmp("FORM",static_buffer,4) && !memcmp("AIF",static_buffer+8,3)){
    int ptr;
    if(static_buffer[11]=='C'){
      aifc=-1;
      report("Input is an AIFF-C bitstream...");
    }else{
      if(static_buffer[11]!='F')return(0);
      report("Input is an AIFF bitstream...");
    }

    /* It's an AIFF/C file */
    ptr=12;

    while(ptr<static_buffer_size-10){
      /* What chunk is this?.... */
      if(!memcmp("SSND",static_buffer+ptr,4))
	reporterror("No COMM chunk preceeds the sound data!");
      if(!memcmp("COMM",static_buffer+ptr,4)){
	if(ptr>=static_buffer_size-GetNum(static_buffer+ptr+4,1,4)-8)
	  reporterror("Unexpected end of file!");
	*channels=GetNum(static_buffer+ptr+8,1,2);
	*rate=IEEE80toDOUBLE(static_buffer+ptr+16);
	*bits=GetNum(static_buffer+ptr+14,1,2);
	*bigendian=-1; /* AIFF/C always bigendian */
	*osigned=-1;  /* AIFF/C is always signed */
	*end=GetNum(static_buffer+ptr+10,1,4);
	
	if(aifc) /* we need to make certain this isn't compressed data;
		    can't handle that yet */
	  if(memcmp(static_buffer+ptr+26,"NONE",4))
	    reporterror("This AIFF-C uses compression; Ogg can't read \n"
			"it (yet).\n");
	  
	/* OK!  Now find the SSND chunk... */
	ptr+=8+GetNum(static_buffer+ptr+4,1,4);
	if(ptr%2)ptr++;

	while(ptr<static_buffer_size-12){
	  if(!memcmp("SSND",static_buffer+ptr,4)){
	    /* Whamo.  Here it is. */
	    ptr+=8;
	    ptr+=GetNum(static_buffer+ptr,1,4);
	    ptr+=4; /* ignore block size field */
	    if(ptr%2)ptr++;
	    
	    static_buffer_read(infile,ptr);
	    static_buffer_shift(ptr);    
	    bytesin+=ptr;
	    return(-1);
	  }
	  ptr+=8+GetNum(static_buffer+ptr+4,1,4);
	  if(ptr%2)ptr++;
	}
	reporterror("Couldn't find sound data in AIFF[-C] file.");

      }
      ptr+=8+GetNum(static_buffer+ptr+4,1,4);
      if(ptr%2)ptr++;

    }
    reporterror("Couldn't find COMM chunk in AIFF[-C] header.");
  }
  return(0);
}

void WriteAifc(FILE *f,long channels,long rate,long bits,
	     int *bigendian,int *osigned){
  header_seekpos=ftell(f);

  /* Again, quick and dirty.  To be replaced by Mike */

  fprintf(f,"FORM    AIFC");                        /* 12 */

  fprintf(f,"FVER");                                /* 16 */
  PutNum(4,f,1,4);                                  /* 20 */
  PutNum(2726318400UL,f,1,4);                        /* 24 */

  fprintf(f,"COMM");                                
  PutNum(38,f,1,4);
  PutNum(channels,f,1,2);                           /* 34 */
  PutNum(0,f,1,4); /* Catch on the other side */    
  PutNum(bits,f,1,2);                                
  PutFloat10(rate,f);                               /* 50 */

  PutNum(14,f,1,1);                                 /* 51 */
  fprintf(f,"not compressed");                      /* 65 */
  PutNum(0,f,1,1);                                  /* 66 */

  fprintf(f,"SSND");                                /* 70 */
  PutNum(0,f,1,4); /* later */
  PutNum(0,f,1,4);
  PutNum(0,f,1,4);                                  /* 82 */
}

void FinishAifc(FILE *f,long channels,long rate,long bits){
  long size=ftell(f)-header_seekpos;
  long frames;

  /* Have to pad if not word aligned */
  if(size%2){
    fputc(0,f);
    size++;
  }
  frames=(size-82)/channels/((bits-1)/8+1);

  /* FORM size */
  if(fseek(f,header_seekpos+4,SEEK_SET))
    reporterror("Couldn't seek to finish the AIFF-C header!");
  PutNum(size-8,f,1,4);

  /* COMM frames */
  if(fseek(f,header_seekpos+34,SEEK_SET))
    reporterror("Couldn't seek to finish the AIFF-C header!");
  PutNum(frames,f,1,4);

  /* SSND size */
  if(fseek(f,header_seekpos+70,SEEK_SET))
    reporterror("Couldn't seek to finish the AIFF-C header!");
  PutNum(size-74,f,1,4);
}

int CheckVoc(FILE *infile,long *channels,long *rate,long *bits,
	     long *bigendian,long *osigned){
  report("Oops...  Don't know how to check VOC headers yet.");
  return(0);
}

int CheckSnd(FILE *infile,long *channels,long *rate,long *bits,
	     long *bigendian,long *osigned){
  report("Oops...  Don't know how to check SND headers yet.");
  return(0);
}

