/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: main.c
  function: Application top level
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 18 1996

 ********************************************************************/

/* Most of the code that makes direct use of libogg is in oggstuff.c;
   setup.c generates OggScripts and talks to the Ogg parser. */

/* The header parsing code for WAV and AIFF[C] is the ugliest stuff
   I've written in a while.  It's soon to be replaced by the finished
   product by Mike Whitson. */

#include <signal.h>
#include "../libogg/libogg.h"
#include "oggapp.h"
#include "oggstuff.h"
#include "setup.h"
#include "appio.h"
#include "header.h"

/* command line values and globals */

char OPTFORMAT[]="qvW:A:V:R:S:blc:C:r:suQ:T:LB:E:N:h**";
long bytesout=0;
long bytesin=0;

int hasoption(char c){
  int i=0;
  while(OPTFORMAT[i]!='*'){
    if(OPTFORMAT[i]==c)
      if(OPTFORMAT[i+1]==':')
	return(1);
      else
	return(0);
    i++;
  }
  return(0);
}

/* This finds a lone (non argument) string; '1' is the first string
   after the actual executable name */
char *stringoption(int argc,char *argv[],int pos){
  int i;
  for(i=1;i<argc;i++){
    if(argv[i][0]=='-'){
      if(argv[i][1]=='\n' || !strcmp(argv[i-1],"--"))pos--;
    }else
      if(argv[i-1][0]=='-'){
	if(!hasoption(argv[i-1][strlen(argv[i-1])-1]))
	  pos--;
      }else
	pos--;
    if(pos==0)return(argv[i]);
  }
  return(NULL);
}

int checkoption(int argc,char *argv[],int c){
  int i;
  for(i=1;i<argc;i++){
    if(argv[i][0]=='-'){
      if(strcmp(argv[i-1],"--")){
	int j;
	for(j=1;j<strlen(argv[i]);j++){
	  if(hasoption(argv[i][j]) && argv[i][j]!=c)break;
	  if(argv[i][j]==c)return(-1);
	}
      }
    }
  }
  return(0);
}

int optionnum(int argc,char *argv[],int pos,long *num,char **s){
  int i;
  for(i=1;i<argc;i++){
    if(argv[i][0]=='-'){
      if(strcmp(argv[i-1],"--")){
	int j;
	for(j=1;j<strlen(argv[i]);j++){
	  if(argv[i][j]!='-')pos--;
	  if(pos<0){
	    if(argv[i][j+1]=='\0'){
	      *s=argv[i+1];
	      if(*s)
		*num=atoi(*s);
	      else
		*num=0;
	    }else{
	      *s=argv[i]+j+1;
	      *num=atoi(*s);
	    }
	    return(argv[i][j]);
	  }
	  if(hasoption(argv[i][j]))break;
	}
      }
    }
  }
  return(0);
}

int main(int argc,char *argv[]){
  long i;
  long opnum;
  char *infilename=NULL, *outfilename=NULL, opchar, *opstring, buffer[160];
  char *antileak=NULL;
  FILE *infile=NULL, *outfile=NULL;

  long channels=2;
  long polyphonic=-1;
  long rate=44100;
  long bits=16;

  int quality=9;
  int time=5;

  long begin=0;
  long end=0;
  long frames=0;

  long bigendian=0;
  long osigned=-1;

  int oggin=0;
  int rawin=0; /* Only need rawin; the header can simply be stripped
		after it is checked. */

  int rawout=0;
  int wavout=0;
  int aiffout=0;
  int vocout=0;
  int sndout=0; 
  int mutexout=0;

  signal(SIGPIPE,SIG_IGN);

  /* first look for the verbose flag; the following steps might want to 
     report actions */
  
  OGG_verbose=checkoption(argc,argv,'v');
  if(checkoption(argc,argv,'h')){
    OGG_verbose=-1;
    report("\nUsage: ogg oggfile [-q -v] [-W|A|R outputfile]\n"
	   "       ogg inputfile [oggfile] [-ULZ] [-qv] [comp opts]\n");
    report("Options: (see man page for more details)\n"
	   "  -A file   Output AIFC format to 'file'\n"
	   "  -B<fr>    Compress beginning at frame <fr>\n"
	   "  -b        Bigendian input (compression only)\n"
	   "  -c<ch>    input channels (compression only)\n"
	   "  -E<fr>    End compression at frame <fr>\n"
	   "  -h -help  Display this help information\n"
	   "  -L        Lossless mode (same as -Q10)\n"
	   "  -m        Disable polyphonic analysis (multichannel only)\n"
	   "  -N<fr>    Compress <fr> frames of sound\n"
	   "  -p        Enable polyphonic analysis\n"
	   "  -q        Query input header and exit\n"
	   "  -Q<qu>    Compression quality (1=low 8=high 9,10=lossless)\n"
	   "  -R file   Output raw audio data to 'file'\n"
	   "  -r<freq>  Sample rate of input data (compression only)  \n"
	   "  -s        Signed input (compression only)\n"
	   "  -T<ti>    Compression time (1=fastest 5=smallest)\n"
	   "  -U        Create an uncompressed Ogg file\n"
	   "  -u        Unsigned input (compression only)\n"
	   "  -v        Verbose\n"
	   "  -W file   Output .wav to 'file'\n"
	   "  -w        Force a raw input word size (in bits)\n"
	   "  -Z        Quasi-lossless mode (same as -Q9)\n"
	   "The input and/or output file may be specified as '-', which\n"
	   "specifies stdin or stdout respectively.  Not specifying an \n"
	   "output file will produce an output file with the base name of\n"
	   "the input file, plus the appropriate extention.\n");
    report("For the most up to date information about OggSQUISH, check\n"
	   "URL http://deskfish.cs.titech.ac.jp:8001/squish/index.html or\n"
	   "contact xiphmont@mit.edu\n");
    exit(0);
  }
  OGG_verbose=checkoption(argc,argv,'q')?-1:OGG_verbose;

  sprintf(buffer,"\nXiphophorus OggSQUISH"
	 " file converter/player (libogg v%d.%drev%d)\n(c) "
	 "1994-1996 Xiphophorus\n",
	 MAJOR,MINOR,REVISION);
  report(buffer);
	  
  /* find the input file name */
  /* (later we should also accept URLs and such, for WWW and OggStream
     connects */

  infilename=stringoption(argc,argv,1); /* get first standalone string */

  /* open the file */
  if(infilename==NULL || !strcmp(infilename,"-")){
    /* piped stdin input */
    infile=stdin;
  }else{
    /* regular file, it seems... */
    infile=fopen(infilename,"rb");
    if(infile==NULL)reporterror("Cannot open input file");
  }

  /* Check for file type.... */

  /* Ogg archive? */
  if(CheckOgg(infile,&channels,&rate,&bits))
    oggin=-1;
  else
    if(CheckWav(infile,&channels,&rate,&bits, &bigendian,&osigned,&end))
      rawin=-1;
    else
      if(CheckAiff(infile,&channels,&rate,&bits, &bigendian,&osigned,&end))
	rawin=-1;
      else
	if(CheckVoc(infile,&channels,&rate,&bits, &bigendian,&osigned,&end))
	  rawin=-1;
	else
	  if(CheckSnd(infile,&channels,&rate,&bits, &bigendian,&osigned,&end))
	    rawin=-1;
	  else{
	    report("Assuming raw data input file (no header detected)...");
	    report("Using default raw settings...");
	  }

  sprintf(buffer,"%ld bits, %ld channel(s), %ld Hz",bits,channels,rate);
  flatreport(buffer);

  if(!oggin)
    if(osigned)
      flatreport(", signed");
    else
      flatreport(", unsigned");

  if(!oggin)
    if(bigendian)
      flatreport(", bigendian");
    else
      flatreport(", littleendian");
  flatreport("\n");

  if(checkoption(argc,argv,'q'))exit(0);
  
  /* The rest of the command line is options and numbers; we also
     might have an output file, catch that at the end. */
  
  i=0;
  while((opchar=optionnum(argc,argv,i++,&opnum,&opstring))){
    switch(opchar){
    case 'W':
      if(mutexout)reporterror("Specify only one of W, A, V, S, R.");
      if(!oggin)reporterror("W, A, V, S and R can only be used on Ogg input.");
      wavout=mutexout=-1;
      outfilename=opstring;
      report("Outputting RIFF/WAV audio file.");
      break;
    case 'A':
      if(mutexout)reporterror("Specify only one of W, A, V, S, R.");
      if(!oggin)reporterror("W, A, V, S and R can only be used on Ogg input.");
      aiffout=mutexout=-1;
      outfilename=opstring;
      report("Outputting AIFF-C audio file (no compression).");
      break;
    case 'V':
      if(mutexout)reporterror("Specify only one of W, A, V, S, R.");
      if(!oggin)reporterror("W, A, V, S and R can only be used on Ogg input.");
      vocout=mutexout=-1;
      outfilename=opstring;
      report("Outputting VOC audio file.");
      break;
    case 'S':
      if(mutexout)reporterror("Specify only one of W, A, V, S, R.");
      if(!oggin)reporterror("W, A, V, S and R can only be used on Ogg input.");
      sndout=mutexout=-1;
      outfilename=opstring;
      report("Outputting SND audio file.");
      break;
    case 'R':
      if(mutexout)reporterror("Specify only one of W, A, V, S, R.");
      if(!oggin)reporterror("W, A, V, S and R can only be used on Ogg input.");
      rawout=mutexout=-1;
      outfilename=opstring;
      report("Outputting raw data audio file.");
      break;
    case 'b':
      if(oggin)reporterror("-b cannot be used on an OggSQUISH input file.");
      bigendian=-1;
      report("Bigendian input specified.");
      break;
    case 'l':
      if(oggin)reporterror("-l cannot be used on an OggSQUISH input file.");
      bigendian=0;
      report("Littleendian input specified.");
      break;
    case 'm':
      if(oggin)reporterror("-m cannot be used on an OggSQUISH input file.");
      polyphonic=0;
      report("Multichannel (non-polyphonic) analysis requested.");
      break;
    case 'p':
      if(oggin)reporterror("-p cannot be used on an OggSQUISH input file.");
      polyphonic=-1;
      report("Polyphonic analysis requested.");
      break;
    case 'c':
      if(oggin)reporterror("-c cannot be used on an OggSQUISH input file.");
      if(opnum<1 || opnum>8)reporterror("Number of channels out of bounds.");
      channels=opnum;
      report("Multichannel input specified.");
      break;
    case 'w':
      if(oggin)reporterror("-w cannot be used on an OggSQUISH input file.");
      if(opnum<1 || opnum>24)reporterror("Word bit size out of bounds.");
      bits=opnum;
      sprintf(buffer,"Sample size override: %ld bits",bits);
      report(buffer);
      break;
    case 'r':
      if(oggin)reporterror("-r cannot be used on an OggSQUISH input file.");
      switch(opnum){
      case 8000:case 11025:case 22050:case 44100:case 48000:
	sprintf(buffer,"Sampling rate override: %ld Hz",opnum);
	report(buffer);
	break;
      default:
	sprintf(buffer,"Warning: Strange sampling rate specified (%ld Hz).",
		opnum);
	report(buffer);
      }
      rate=opnum;
      break;
    case 's':
      if(oggin)reporterror("-s cannot be used on an OggSQUISH input file.");
      osigned=-1;
      report("Signed input specified.");
      break;
    case 'u':
      if(oggin)reporterror("-u cannot be used on an OggSQUISH input file.");
      osigned=0;
      report("Unsigned input specified.");
      break;
    case 'Q':
      if(oggin)reporterror("-Q cannot be used on an OggSQUISH input file.");
      if(opnum<1 || opnum>10)reporterror("Quality setting must be between"
					 " 1 and 10 inclusive.");
      quality=opnum;
      if(quality==10)
	report("Lossless compression requested.");
      else{
	sprintf(buffer,"Compression quality setting: %ld.",opnum);
	report(buffer);
      }
      break;
    case 'T':
      if(oggin)reporterror("-T cannot be used on an OggSQUISH input file.");
      if(opnum<1 || opnum>5)reporterror("Time/speed setting must be between"
					 " 1 and 5 inclusive.");
      sprintf(buffer,"Compression time setting: %ld.",opnum);
      report(buffer);
      time=opnum;
      break;
    case 'U':
      if(oggin)reporterror("-U cannot be used on an OggSQUISH input file.");
      time=0;
      quality=10;
      report("Non compressed bitstream requested.");
      break;
    case 'L':
      if(oggin)reporterror("-L cannot be used on an OggSQUISH input file.");
      quality=10;
      report("Lossless compression requested.");
      break;
    case 'Z':
      if(oggin)reporterror("-Z cannot be used on an OggSQUISH input file.");
      quality=9;
      report("Quasi-lossless compression requested.");
      break;
    case 'B':
      if(oggin)reporterror("-B cannot be used on an OggSQUISH input file.");
      begin=opnum;
      break;
    case 'E':
      if(oggin)reporterror("-E cannot be used on an OggSQUISH input file.");
      end=opnum;
      break;
    case 'N':
      if(oggin)reporterror("-N cannot be used on an OggSQUISH input file.");
      frames=opnum;
      break;
    case 'v':
      break;
    default:
      sprintf(buffer,"Unknown option: %c.",(char)opchar);
      reporterror(buffer);
      break;
    }
  }
  if(!frames)frames=end-begin;
  if(channels<2)polyphonic=0;

  if(oggin){
    /* do we have an output we need to open for conversion, or are we
       playing audio directly? */

    if(outfilename){
      int type=0;

      /* OK; conversion */
      if(sndout || vocout)reporterror("Can't output to SND or VOC yet");

      if(!rawout){
	/* WAV/etc only work if the output is seekable */
	if(!strcmp(outfilename,"-"))
	  reporterror("Outputting WAV and AIFF[-C] requires a seekable "
		      "output stream.");
	/* Can't always tell; the output could be a named pipe for all
           we know */
	report("NOTE: Outputting WAV and AIFF[-C] requires a seekable "
	       "output stream.");

      }

      if(wavout)type=1;
      if(aiffout)type=2;
      if(sndout)type=3;
      if(vocout)type=4;

      decompression_loop(infile,outfilename,type,bits,channels,rate);
      flatreport("\rDone.              "
		 "                                                    \n\n");
    }else{
      /* OK; direct playback */
      /* Our header is already loaded */
      /* Oooo.... hardwired /dev/dsp... Mike needs to fix this too :) */

      decompression_loop(infile,"/dev/dsp",-1,bits,channels,rate);
      flatreport("\rDone.              "
		 "                                                      \n\n");
    }
  }else{
    /* Convert input to Ogg; our input stream is already positioned. */
    /* Look for output file in command line */

    outfilename=stringoption(argc,argv,2);
    if(outfilename==NULL){
      char *temp=strrchr(infilename,'.');
      antileak=outfilename=calloc(strlen(infilename)+10,1);
      if(outfilename==NULL)reporterror("Couldn't alloc memory");
      if(temp==NULL)
	strcpy(outfilename,infilename);
      else
	strncpy(outfilename,infilename,temp-infilename);
      strcat(outfilename,".ogg");
      flatreport("No output file specified; writing to ");
      flatreport(outfilename);
      flatreport("\n");
    }

    /* Open output file */
    
    if(strcmp(outfilename,"-"))
      outfile=fopen(outfilename,"wb");
    else
      outfile=stdout;
    if(outfile==NULL)reporterror("Unable to open output file.");

    /* Save decompression header and load compression OggScript */

    script_setup(outfile,channels,polyphonic,rate,bits,quality,time);

    /* Compression loop */

    compression_loop(infile,outfile,bits,channels,bigendian,osigned,begin,
		     frames);

    flatreport("\rDone.              "
	   "                                                        ");
    if((float)bytesin/bytesout>1.)
      sprintf(buffer,"\rDone.  Bytes out: %ld  Compression ratio: %#.3g:1\n\n",
	      bytesout,(float)bytesin/bytesout);
    else
      sprintf(buffer,"\n\n");
    
    flatreport(buffer);

  }

  /* Clean up */
  if(static_buffer)free(static_buffer);
  if(antileak)free(antileak);
  return(0);
}

