/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH PLAYER/CONVERTER SOURCE CODE. *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: oggstuff.c
  function: Application libogg usage code
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 08 1996

 ********************************************************************/

#include "oggapp.h"
#include "appio.h"
#include "libogg.h"
#include "header.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef __linux__
#include <sys/soundcard.h>
#include <sys/ioctl.h>
#endif

extern long bytesin;
extern long bytesout;

int CheckOgg(FILE *infile,int *channels,long *rate,int *bits){

  /* work in 1k blocks */

  /* normally, we wouldn't do it as below, but we have to keep bytes
     from the beginning if this turns out not to be an Ogg
     bitstream. */

  long bytes=0,read;
  while(1){

    /* Normally, we'd just send a new block, but it's more
       straightforward right now to resubmit from the beginning.  This
       means we have to reset the parser each time */

    ogg_script_reset();

    if(static_buffer_size<bytes)reporterror("Unexpected end of file!");
    /* tops off the static buffer */
    static_buffer_read(infile,bytes+=1024);

    /* Inspect the input; if it's an Ogg header, load it */
    read=ogg_script_loadheader(static_buffer,static_buffer_size);

    /* Check results */
    switch(ogg_errno){
    case 0:
      /* No error; did we finish? */
      if(ogg_script_ready){
	char buffer[160];
	struct ogg_dspinfo info;

	/* OK; libogg says it got a complete header.  Commit to
	   reading an ogg file and shift the static read buffer */
	static_buffer_shift(read);
	bytesin+=read;
	sprintf(buffer,"Input is an OggSQUISH version %d.%d (rev %d) "
		"bitstream.",ogg_bitstream_major,ogg_bitstream_minor,
		ogg_bitstream_revision);
	report(buffer);

	/* Send the script to the Ogg interpreter */
	ogg_script_use(0);

	switch(ogg_errno){
	case 0:
	  if(ogg_run_loaded && ogg_run_ready){

	    /* Query the primary dsp output buffer; this bitstream might
	       not have an audio component!  We also need the sample
	       information to report */
	    
	    if(!ogg_buffer_used(ogg_dsp_out,ogg_primary))
	      reporterror("Could not locate an audio component in the "
			  "bitstream.");
	    ogg_buffer_dsp(ogg_dsp_out,ogg_primary,&info);

	    /* This *really* shouldn't generate an error, but it's
               still good to check */
				
	    if(ogg_errno){
	      report("Problem checking OggSQUISH bitstream information:");
	      reporterror(ogg_strerr);
	    }
				
	    *channels=info.channels;
	    *rate=info.rate;
	    *bits=info.dynrange;

	    /* All done.  Successful header load. */
	    return(-1);
	  }
	  
	  /* No errors, but ogg_run_ready is not set.  This
	     should never happen. */
	  reporterror("Internal error.");
	case 101:
	  /* Argh; the ogg bitstream header could not be built for
	     some reason.  This is probably a bug in the bitstream
	     decompression script in the header. */
	  /* Report the returned error */
	  reporterror(ogg_script_err_string);
	default:
	  /* Some other error; report it */
	  reporterror(ogg_strerr);
	}
      }
      
      /* No errors, but we didn't manage to read a whole header
	 either.  Try more data */

      break;
    case 7:
      /* Ah, OK.  This isn't an Ogg header */
      return(0);
    case 6:
      /* Oops; this version is too old */
      {
	char buffer[160];
	sprintf(buffer,"Input bitstream is from Ogg version %dalpha%d.",
		ogg_bitstream_major,ogg_bitstream_minor);
	report(buffer);
	reporterror("\nThe specified input file is an old OggSQUISH\n"
		    "bitstream that cannot be loaded by 0.98+ OggSQUISH;\n"
		    "bitstreams from before 0.98 are incompatible.");
      }
    case 8:
      reporterror("OggSQUISH input file header is corrupted.");
    case 100:
      /* This is probably due to a header requesting a feature we
         don't have */
      {
	char buffer[160];
	sprintf(buffer,"Input is an OggSQUISH version %d.%d (rev %d) "
		"bitstream.",ogg_bitstream_major,ogg_bitstream_minor,
		ogg_bitstream_revision);
	report(buffer);

	if(ogg_script_err_number==9)
	  if(MAJOR<ogg_bitstream_major || 
	     (MAJOR==ogg_bitstream_major && MINOR<ogg_bitstream_minor) ||
	     (MAJOR==ogg_bitstream_major && MINOR==ogg_bitstream_minor &&
	      REVISION<ogg_bitstream_revision))
	    reporterror("The input bitstream has requested a module "
			"function that seems to be newer than this version "
			"of OggSQUISH.  Consider upgrading to a more recent\n"
			"version of ogg.");
	  else
	    reporterror("The input bitstream has requested a module "
			"function that this Ogg decoder does not \n"
			"understand.\n");
	report("Header corruption!");
	reporterror(ogg_script_err_string);
      }
    default:
      reporterror(ogg_strerr);
    }
  }
  /* never get here */
}

void compression_loop(FILE *infile,FILE *outfile,long bits,long channels,
		      int endianness,int sign,long begin,long frames){
  int bytesperelement;
  int bytesperframe=(bytesperelement=(bits-1)/8+1)*channels;
  char buffer[160];
  int flag=0;
  long bytesinP=0;
  long bytesoutP=0;

  long FRAMECOUNTER=0,i;
  FILE *analysis;
  
  for(i=0;i<10;i++)
    if(ogg_buffer_used(ogg_dsp_out,ogg_secondary1+i)){
      sprintf(buffer,"Analysis output on secondary dsp output buffer %ld",i+1);
      report(buffer);
    }

  /* Before the loop, we might need to fast forward. */ 
  /* Right now, for sake of simplicity, we always assume that the
     input and output streams are not seekable. that means we gotta
     read and discard alot of bytes */

  report("");
  begin*=bytesperframe;
  if(OGG_verbose)
    sprintf(buffer,"\r=[  Fast Forwarding  ]==> %ldk to go...   ",begin>>10);
  flatreport(buffer);

  while(begin){
    long toread=1024;
    if(toread>begin)toread=begin;
    static_buffer_read(infile,toread);
    if(begin<static_buffer_size){
      static_buffer_shift(begin);
      begin=0;
    }else{
      begin-=static_buffer_size;
      static_buffer_shift(static_buffer_size);
    }
    if(++flag % 19 ==0 && OGG_verbose){
      sprintf(buffer,"\r=[  Fast Forwarding  ]==> %ldk to go...   ",begin>>10);
      flag=0;
    }
    flatreport(buffer);
  }
    
  /* The compression loop is very simple; first we determine how much
     input the engine wants, then we read and submit that much data.
     Make one pass and fetch the output data, then write the output.
     Repeat until termination.  At the moment, this compression engine
     just puts everything after a single header.  Multiblocking with
     tuning support is for a later revision (Once again, Wrong Mike's
     problem :) */

  /* Also as a matter of convenience, there's some new signal analysis
     code here.  If we use secondary dsp output buffers in the
     compression code, the loop assumes that these are for output of
     analysis waveforms; the buffers are drained, and each is sent to
     a file, unique to the buffer and the frame number */

  flatreport("\r=[    Processing     ]==>                                ");
  if(!frames)
    frames=-1;
  else
    frames*=bytesperframe;
  while(ogg_run_ready){
    int elements;

    /* This is a cute, simple way of dealing with minimum input size
       requests. */
    while(ogg_input_query(ogg_dsp_in,ogg_primary) && frames && !feof(infile)){
      int toread=1024*bytesperframe;
      int bread;

      /* We might have an explicit stopping point */
      if(frames!=-1 && toread>frames)toread=frames;

      if(!feof(infile))static_buffer_read(infile,toread);

      /* So long as we submit from the beginning of the static buffer,
         our alignment is OK */
      elements=static_buffer_size/bytesperelement;
      ogg_input_dsp(ogg_dsp_in,ogg_primary,static_buffer,elements,bits,
		    endianness?OGG_BIGENDIAN:OGG_LITTLEENDIAN,sign,-1);

      /* Check for errors */
      switch(ogg_errno){
      case 0:
	bread=elements*bytesperelement;
	if(frames!=-1)frames-=bread;
	static_buffer_shift(bread);
	bytesin+=bread;
	break;
      default:
	reporterror(ogg_strerr);
      }
    }

    /* All set, call the engine on this frame */
    flatreport("\r=[    Processing     ]==> ");
    ogg_run();
    if(ogg_errno)reporterror(ogg_strerr);

    /* Get the output data and write to our output stream */
    /* If we're only doing analysis, this buffer might not even be in use */
    if(ogg_buffer_used(ogg_bit_out,ogg_primary)){
      char *outdata;
      long bread;
      ogg_output_bit(ogg_bit_out,ogg_primary,&outdata,&bread);
      
      /* Errors? */
      if(ogg_errno)reporterror(ogg_strerr);
      
      if(outdata){
	/* Write output */
	blocking_write(outdata,1,bread,outfile);
	bytesout+=bread;
	free(outdata);
      }
    }

    /* Waveform analysis output */
    for(i=0;i<10;i++)
      if(ogg_buffer_used(ogg_dsp_out,ogg_secondary1+i)){	
	void *outdata;
	long bread;
	ogg_output_dsp(ogg_dsp_out,ogg_secondary1+i,&outdata,&bread,32,
		       OGG_SAMEENDIAN,-1,0);
	
	/* Errors? */
	if(ogg_errno)reporterror(ogg_strerr);
	
	if(outdata){
	  struct ogg_dspinfo info;
	  int ch;

	  /* We need to know some things about our data... */
	  ogg_buffer_dsp(ogg_dsp_out,ogg_secondary1+i,&info);

	  /* Write output */
	  for(ch=0;ch<info.channels;ch++){
	    long j;

	    /* analysis.frame.buffer.channel */
	    sprintf(buffer,"in.%ld.%ld.%d",FRAMECOUNTER,i,ch);
	    analysis=fopen(buffer,"w");

	    for(j=0;j<bread;j+=info.channels)
	      fprintf(analysis,"%ld\n",(long)(((size32 *)outdata)[j+ch]));
	    fclose(analysis);
	  }
	  free(outdata);
	}
      }

    if(OGG_verbose){
      if(bytesin-bytesinP>5*1024){
	sprintf(buffer,"%ldk",bytesin>>10);
	flatreport(buffer);
	sprintf(buffer," -> %ldk",bytesout>>10);
	flatreport(buffer);
	
	if(bytesout-bytesoutP && bytesin>bytesout){
	  sprintf(buffer," (%#.3g:1",(double)bytesin/bytesout);
	  flatreport(buffer);
	  
	  sprintf(buffer,", IR %#.3g:1)        ",(double)(bytesin-bytesinP)/
		  (bytesout-bytesoutP));
	  bytesinP=bytesin;
	  bytesoutP=bytesout;
	  flatreport(buffer);
	}
      }
    }
    FRAMECOUNTER++;
  }
}

/* This function is intended to eventually be ultra-robust; if it
   can't set the proper playback mode for a sample, it should take the
   best it can get and run the sample output through a post-processing
   filter in libogg to convert to the best available playback mode. */

void decompression_loop(FILE *infile,char *outfilename,int type,
			int bits,int channels,long rate){
  int bytesperelement;
  int bytesperframe;
  int endian,sign;
  long bytespersecond;
  FILE *outfile,*analysis;
  long realbytes=bytesout,verbflag,i,FRAMECOUNTER=0;
  char buffer[160];

  for(i=0;i<10;i++)
    if(ogg_buffer_used(ogg_dsp_out,ogg_secondary1+i)){
      sprintf(buffer,"Analysis output on secondary dsp output buffer %ld",i+1);
      report(buffer);
    }

  report("");

  switch(type){
  case 0:case 1:case 2:case 3:
    outfile=fopen(outfilename,"wb");
    if(outfile==NULL)
      reporterror("Could not open output file.");
  }

  /* Ogg bitstreams can terminate and resume with a new header; keep
     going until we actually get an EOF... */

  while(static_buffer_size || !feof(infile)){
    int audio,temp;
    char buffer[160];
    long playblocksize=32768,playblockptr;
    char *playblock;

    bytesperframe=(bytesperelement=(bits-1)/8+1)*channels;
    bytesout*=bytesperframe; /* time counter hack */

    switch(type){
    case -1:
      /* Direct playback mode */

      outfile=fopen(outfilename,"wb");
      if(outfile==NULL)
	reporterror("Unable to open audio playback device.");

      /* Can we set a playback mode exactly matching the sample? */
      /* If not, set the best we can.  Playback speed is a problem; as
	 soon as we add effects to the application, modifying playback
	 speed is trivial by using libogg to do it. */
      
#ifdef __linux__
      /* In reality, the following code would likely work with minimal
	 effort on any UNIX box using Hannu's Unix Sound System code,
	 but I've not the other hardware to test it; also I don't know
	 how Hannu abstracts over endianness */

      audio=fileno(outfile);
      if(audio==-1)
	reporterror("Unable to open audio playback device");
    
      endian=0;
      temp=bits;
      ioctl(audio,SNDCTL_DSP_SAMPLESIZE,&temp);
      if(temp!=bits){
	/* Uh oh...  Let's see what we can manage to set... */
	int try;
	for(try=16;try>3;try--){
	  temp=try;
	  ioctl(audio,SNDCTL_DSP_SAMPLESIZE,&temp);
	  if(temp==try){
	    sprintf(buffer,"Warning: Unable to set %d bit sample playback.\n"
		    "         Using %d bit playback instead.",bits,try);
	    report(buffer);
	    bits=try;
	    break;
	  }
	}
	if(temp!=try)
	  reporterror("Unable to set any useful playback mode.");
      }
      
      if(bits>8)
	sign=-1;
      else
	sign=0;
      
      switch(channels){
      case 1:
	temp=0;
	if(ioctl(audio,SNDCTL_DSP_STEREO,&temp)==-1)
	  reporterror("Unable to set monophonic playback mode.");
	break;
      case 2:
	temp=1;
	if(ioctl(audio,SNDCTL_DSP_STEREO,&temp)==-1)
	  reporterror("Unable to set stereo playback mode.");
	break;
      default:
	sprintf(buffer,"Unable to set %d channel playback mode",channels);
	reporterror(buffer);
      }
      
      temp=rate;
      if(ioctl(audio,SNDCTL_DSP_SPEED,&temp)==-1){
	sprintf(buffer,"Unable to set %ldHz playback speed",rate);
	reporterror(buffer);
      }
      
      ioctl(audio,SNDCTL_DSP_GETBLKSIZE,&playblocksize);
      if(playblocksize==-1)
	reporterror("Playback error reading sound blocksize");
      
#else
      /* Mike, add SGI libaudio support */
      /* We'll do some /dev/audio code later too when the playback
	 conversion speed effect support is added (/dev/audio==8kHz)... */
      reporterror("Direct playback is currently supported only under Linux");
#endif
      playblockptr=0;
      playblock=(char *)malloc(playblocksize);
      if(playblock==NULL)reporterror("Unable to allocate memory!");
      break;
    case 1:
      WriteWav(outfile,channels,rate,bits,&endian,&sign);
      break;
    case 2:
      WriteAifc(outfile,channels,rate,bits,&endian,&sign);
      break;
    } 

    /* OK, we're set... */
    
    flatreport("\r=[    Processing     ]==> ");
    bytespersecond=rate*((bits-1)/8+1)*channels;
    while(ogg_run_ready){
      char *temp_output;
      long temp_outbytes=0;

      /* Submit data for a pass with libogg */
      /* Just grab fixed size blocks and submit until we've grabbed enough */
      while(ogg_input_query(ogg_bit_in,ogg_primary) && !feof(infile)){
	if(!feof(infile))static_buffer_read(infile,4096);
	
	/* Submit the grabbed bitstream to libogg */
	
	ogg_input_bit(ogg_bit_in,ogg_primary,static_buffer,
		      static_buffer_size);
	
	/* Check for errors */
	switch(ogg_errno){
	case 0:
	  static_buffer_shift(static_buffer_size);
	  bytesin+=static_buffer_size;
	  break;
	default:
	  reporterror(ogg_strerr);
	}
      }
      
      /* All set, call the engine on this frame */
      ogg_run();
      if(ogg_errno)reporterror(ogg_strerr);
      
      /* Get the output data */
      
      ogg_output_dsp(ogg_dsp_out,ogg_primary,(void **)&temp_output,
		     &temp_outbytes,bits,OGG_SAMEENDIAN,sign,-1);
      bytesout+=temp_outbytes*=bytesperelement;
      realbytes+=temp_outbytes;

      /* Errors? */
      if(ogg_errno)reporterror(ogg_strerr);
      
      verbflag=0;
      if(type==-1){
	long tempptr=0;
	if(playblockptr+temp_outbytes>=playblocksize){
	  if(playblockptr){
	    memcpy(playblock+playblockptr,temp_output,
		   playblocksize-playblockptr);
	    blocking_write(playblock,1,playblocksize,outfile);
	    tempptr+=playblocksize-playblockptr;
	    playblockptr=0;
	    verbflag=-1;
	  }
	  while(temp_outbytes-tempptr>=playblocksize){
	    blocking_write(temp_output+tempptr,1,playblocksize,outfile);
	    tempptr+=playblocksize;
	    verbflag=-1;
	  }
	}
	if(tempptr<temp_outbytes){
	  memcpy(playblock+playblockptr,temp_output+tempptr,
		 temp_outbytes-tempptr);
	  playblockptr+=temp_outbytes-tempptr;
	}
	free(temp_output);

	if(!ogg_run_ready){
	  blocking_write(playblock,1,playblockptr,outfile);
	  verbflag=-1;
	}
      }else{
	verbflag=-1;
	blocking_write(temp_output,1,temp_outbytes,outfile);
	free(temp_output);
      }

      /* Waveform analysis output */
      for(i=0;i<10;i++)
	if(ogg_buffer_used(ogg_dsp_out,ogg_secondary1+i)){	
	  void *outdata;
	  long bread;
	  ogg_output_dsp(ogg_dsp_out,ogg_secondary1+i,&outdata,&bread,32,
			 OGG_SAMEENDIAN,-1,0);
	  
	  /* Errors? */
	  if(ogg_errno)reporterror(ogg_strerr);
	  
	  if(outdata){
	    struct ogg_dspinfo info;
	    int ch;
	    
	    /* We need to know some things about our data... */
	    ogg_buffer_dsp(ogg_dsp_out,ogg_secondary1+i,&info);
	    
	    /* Write output */
	    for(ch=0;ch<info.channels;ch++){
	      long j;
	      
	      /* analysis.frame.buffer.channel */
	      sprintf(buffer,"out.%ld.%ld.%d",FRAMECOUNTER,i,ch);
	      analysis=fopen(buffer,"w");
	      
	      for(j=0;j<bread;j+=info.channels)
		fprintf(analysis,"%ld\n",(long)(((size32 *)outdata)[j+ch]));
	      fclose(analysis);
	    }
	    free(outdata);
	  }
      }

      FRAMECOUNTER++;

      /* Cutesy hour:minute:second.hundredths output */

      if(verbflag && OGG_verbose){
	long temp=bytesout;
	long seconds=temp/bytespersecond;
	long hundredths=(temp-seconds*bytespersecond)*100/bytespersecond;
	long minutes=seconds/60;
	long hours=0;
	seconds-=minutes*60;
	hours=minutes/60;
	minutes-=hours*60;

	flatreport("\r=[    Processing     ]==> ");
	if(hours)
	  sprintf(buffer," %02ld:%02ld:%02ld.%02ld (%ldkB out)     ",hours,minutes,
		  seconds,hundredths,realbytes>>10);
	else
	  sprintf(buffer," %02ld:%02ld.%02ld (%ldkB out)     ",minutes,
		  seconds,hundredths,realbytes>>10);

	flatreport(buffer);
      }
    }

    /* The bitstream has terminated */
    switch(type){
    case -1:
      fclose(outfile);
      free(playblock);
      break;
    case 1:
      FinishWav(outfile,channels,rate,bits);
      break;
    case 2:
      FinishAifc(outfile,channels,rate,bits);
      break;
    }

    /* The static buffer at this point should be empty */
    {
      /* Recover buffered data */
      char *temp;
      long bread;
      ogg_output_bit(ogg_bit_in,ogg_primary,&temp,&bread);
      if(temp){
	static_buffer_verify(bread);
	memcpy(static_buffer,temp,bread);
	static_buffer_size=bread;
	free(temp);
      }
    }

    if(feof(infile) && static_buffer_size==0) break; /* Done! */

    /* Hmmm....  look to see if we have another header. */
    
    bytesout/=bytesperframe; /* time counter hack */
    ogg_run_reset();
    temp=OGG_verbose;
    flatreport("\r=[Loading new header ]=\n");
    CheckOgg(infile,&channels,&rate,&bits);
    if(ogg_errno)
      reporterror("Garbage at end of file\n");
    
    sprintf(buffer,"Sample changes loaded! %d bits, %d channel(s), %ld Hz",
	    bits,channels,rate);
    report(buffer);
  }
  if(type>=0)fclose(outfile);
}
