/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: setup.c
  function: Virtual machine setup and configuration
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 09 1996

 ********************************************************************/

#include "oggapp.h"
#include "libogg.h"
#include "appio.h"

extern long bytesout;

void parse_errors(){
  char buffer[80];
  switch(ogg_errno){
  case 0:
    return;
  case 1: 
    reporterror("Out of memory."); 
  case 3:
    reporterror("Segfault; Out of memory?");
  case 100:case 101:
    report("");
    report("Internal error in script:");
    if(ogg_script_err_line!=-1)
      sprintf(buffer,"  line: %d",ogg_script_err_line);
    else
      sprintf(buffer,"  statement: %d",ogg_script_err_statement);
    report(buffer);
    sprintf(buffer,"  argument: %d",ogg_script_err_argument);
    report(buffer);
    reporterror(ogg_script_err_string);
  default:
    report("Internal error:");
    reporterror(ogg_strerr);
  }
}

void standard_define(char *s,long val){
  char buffer[80];
  
  sprintf(buffer,"= %s %ld;",s,val);
  ogg_script_parse(buffer);
  parse_errors();
}

void script_setup(FILE *outfile,int channels,int polyphonic,long rate,
		  int bits,int quality,int time){
  char *h;
  long bytes;

  /*if(bits<12)bits=16;*/
  /* First, set up defines with our passed in information */
  standard_define("channels",channels);
  standard_define("poly",polyphonic);
  standard_define("rate",rate);
  standard_define("dyn",bits); /* Yes!  We really can handle
				  arbitrary bit widths
				  efficiently! */

  switch(quality){
    /* Lossless and quasi-lossless are split out of the pack early */
  case 10:case 9:
    switch(time){
    case 0:

      /**********************************************************************/
      /* Uncompressed encoding                                              */
      /**********************************************************************/

      /* Just do the uncompressed case in one fell swoop. */
      standard_define("hint",1024);
      
      ogg_script_parse(
	    "Aid     ogg_dsp_in ogg_primary inP hint rate channels dyn;"
	    "Sof     inP mainP outB hint frameV;"
	    "b:      B;"

      	    "Covr    mainP dyn bM outB;"

	    "Aob     ogg_bit_out ogg_primary outB TRUE;"
	    "Shalt   frameV hint%");
      
      parse_errors();
      ogg_script_use(0); /* Just stick it in interpreter bank 0 */
      parse_errors();
      
      /* Decode script is next */
      /* Our parser has been reset; gotta do the defines thing over again */
      
      standard_define("channels",channels);
      standard_define("poly",polyphonic);
      standard_define("rate",rate);
      standard_define("dyn",bits);
      standard_define("hint",1024);
      
      ogg_script_parse(
	 "Aib     ogg_bit_in ogg_primary inB hint*channels*dyn/8+8 TRUE;"
	 "Sif     inB hint frameV;"
	 "b:      B;"
	 
	 "Civrs   outP dyn frameV channels bM inB;"
	 "Aod     ogg_dsp_out ogg_primary outP rate channels dyn;"
	 "Shalt   frameV hint%");
      parse_errors();
      
      /* Pack the script into an archive header */
      ogg_script_makeheader(&h,&bytes);
      parse_errors();
      
      /* Save that header! (Catch that rabbit?) */
      blocking_write(h, 1, bytes, outfile);
      bytesout+=bytes;
      
      return;
    default:

      /*********************************************************************/
      /* Lossless and quasi-lossless compression                           */
      /*********************************************************************/
      
      standard_define("hint",2048);
      if(bits>16)
	ogg_script_parse("= ddr Dr64;");
      else
	ogg_script_parse("= ddr Dr32;");
      parse_errors();
      
      switch(time){
      case 2:
	ogg_script_parse("= coeffs 2;");
	break;
      case 3: case 4:case 5:
	ogg_script_parse("= coeffs 4;");
	break;
      }
      parse_errors();
      
      ogg_script_parse(
	     "Aid     ogg_dsp_in ogg_primary inP hint rate channels dyn;"
	     "Sof     inP mainP outB hint frameV;"
	     "a:   B;");
	
      parse_errors();
      
      if(time>=4){
	ogg_script_parse("b:   Bh;"
			 "b1:  Bdpl 4 bM 4 128 0;"
			 "b2:  Bdpl 4 bM 4 128 0;"
			 "b3:  Bdpl 4 bM 4 128 0;"
			 "b4:  Bdpl 4 bM 4 128 0;");
	parse_errors();
	if(time>=5){
	  ogg_script_parse("c:   Bdpc 4 4 bM 4 128 0;");
	  parse_errors();
	}
      }
      
      if(polyphonic){
	
	/****** Polyphonics go here *************************************/

	/* Set up A/B/center pipes and priming */
	ogg_script_parse("Ez   centerP frameV channels/2 0;"
			 "Sv   0 loopV;"
			 
	     "loopone:    Jge  loopV+1 channels outoneL;"
			 "Ecs  mainP loopV aP;"
			 "Ecs  mainP loopV+1 bP;"
			 "Eas  aP bP;"
			 "Ecr  centerP loopV>1 aP;"
			 "Jfor loopV channels 2 looponeL;"

	     "outone:     ");
	parse_errors();

	switch(time){
	case 1:
	  ogg_script_parse("Ddf     mainP primeP outP;"
			   "Ddf     centerP cprimeP coutP;");
	  parse_errors();
	  break;
	case 2:case 3:case 4:case 5:
	  if(quality==10)
	    ogg_script_parse("Dca    mainP coeffP nullV coeffs;"
			     "ddr    mainP primeP coeffP outP;"
			     "Dca    centerP ccoeffP nullV coeffs;"
			     "ddr    centerP cprimeP ccoeffP coutP;");
	  else
	    ogg_script_parse(
		      "Dca  mainP coeffP nullV coeffs;"
		      "Dr   mainP primeP coeffP outP 0 32 shiftP;"
		      "Dca  centerP ccoeffP nullV coeffs;"
		      "Dr   centerP cprimeP ccoeffP coutP 0 32 shiftP;");
	  parse_errors();
	}
	
	ogg_script_parse("Sv   0 loopV;"
	     "looptwo:    Jge  loopV+1 channels outtwoL;"
			 
			 "Eav  outP loopV aV;"
			 "Eav  outP loopV+1 bV;"
			 "Eav  coutP loopV>1 cV;"

			 "Jg   bV cV oneL;"
			 "Jg   aV cV twoL;"
			 /* A + B */
			 "Coa  0 1 polyB;"
			 "J    threeL;"
			 /* B + Center */
		  "two:   Ecs  coutP loopV>1 tempP;"
			 "Ecr  outP loopV tempP;"
			 "Ecs  ccoeffP loopV>1 tempP;"
			 "Ecr  coeffP loopV tempP;"
			 "Coa  1 1 polyB;"
			 "Coa  1 1 polyB;"
			 "J    threeL;"
		  "one:   Jg   aV bV twoL;"
			 /* A + Center */
			 "Coa  1 1 polyB;"
			 "Coa  0 1 polyB;"
			 "Ecs  coutP loopV>1 tempP;"
			 "Ecr  outP loopV+1 tempP;"
			 "Ecs  ccoeffP loopV>1 tempP;"
			 "Ecr  coeffP loopV+1 tempP;"
			 "J    threeL;"
	     "three:      Jfor loopV channels 2 looptwoL;"
	     "outtwo:     ");
	parse_errors();

	/************ end of polyphonic analysis ********************/
      }else{
	switch(time){
	case 1:
	  ogg_script_parse("Ddf     mainP primeP outP;");
	  parse_errors();
	  break;
	case 2:case 3:case 4:case 5:
	  if(quality==10)
	    ogg_script_parse("Dca    mainP coeffP nullV coeffs;"
			     "ddr    mainP primeP coeffP outP;");
	  else
	    ogg_script_parse("Dca    mainP coeffP nullV coeffs;"
			     "Dr     mainP primeP coeffP outP 0 32 shiftP;");
	  parse_errors();
	}
      }
      
      switch(time){
      case 2:case 3:case 4:case 5:
	
	ogg_script_parse("Covtl   coeffP compB 1 # aM;"
			 "Conrl   coeffP compB;");
	parse_errors();
	break;
      }
      
      switch(time){
      case 1:case 2: case 3:
	ogg_script_parse("Covtl   outP compB 1 # aM;"
			 "Conrl   outP compB;");
	break;
      case 4:
	ogg_script_parse("Covtl   outP compB 4 # b1M b2M b3M b4M;"
			 "Conrl   outP compB;");
	break;
      case 5:
	ogg_script_parse("Covtl   outP compB 4 # b1M b2M b3M b4M;"
			 "Conc    outP compB 4 4 cM 1;");
	break;
      }
      parse_errors();
      
      /* Make sure we aren't bigger than the original */
      
      ogg_script_parse("Sbt  compB testV;"
		       "Sbt  outB test2V;"
		       "Jwle  test2V+channels*dyn*frameV testV bailL outB;"
		       "Sba  compB outB;");
      parse_errors();
      if(polyphonic){
	ogg_script_parse("Sba  polyB outB;");
	parse_errors();
      }

      switch(time){
      case 5:
	ogg_script_parse("Bxc  cM;");
	parse_errors();
      case 4:
	ogg_script_parse("Bxc  b1M;"
			 "Bxc  b2M;"
			 "Bxc  b3M;"
			 "Bxc  b4M;");
	parse_errors();
      }
      ogg_script_parse("J    finishL;"
		       "bail: ");
      parse_errors();
      switch(time){
      case 5:
	ogg_script_parse("Bxu  cM;");
	parse_errors();
      case 4:
	ogg_script_parse("Bxu  b1M;"
			 "Bxu  b2M;"
			 "Bxu  b3M;"
			 "Bxu  b4M;");
	parse_errors();
      }
      ogg_script_parse("Covr mainP dyn aM outB;"
		       "finish: Aob     ogg_bit_out ogg_primary outB FALSE;"
		       "Spm     mainP primeP;");
      parse_errors();
      if(polyphonic)
	ogg_script_parse("Spm     centerP cprimeP;"
			 "Shalt   frameV hint%");
      else
	ogg_script_parse("Shalt   frameV hint%");

      parse_errors();
      ogg_script_use(0); /* Just stick it in interpreter bank 0 */
      parse_errors();
      
      /* Decompression script is next */
      
      standard_define("channels",channels);
      standard_define("poly",polyphonic);
      standard_define("rate",rate);
      standard_define("dyn",bits);
      standard_define("hint",2048);    
      if(bits>16)
	ogg_script_parse("= ddp Dp64;");
      else
	ogg_script_parse("= ddp Dp32;");
      switch(time){
      case 2:
	ogg_script_parse("= coeffs 2;");
	break;
      case 3: case 4:case 5:
	ogg_script_parse("= coeffs 4;");
	break;
      }
      parse_errors();
      
      ogg_script_parse(
	    "Aib     ogg_bit_in ogg_primary inB hint*channels*dyn/8+8 FALSE;"
	    "Sif     inB hint frameV;"
	    "a:      B;");
      parse_errors();
      if(time>=4){
	ogg_script_parse(
			 "b:      Bh;"
			 "b1:     Bdpl 4 bM 4 128 1;"
			 "b2:     Bdpl 4 bM 4 128 1;"
			 "b3:     Bdpl 4 bM 4 128 1;"
			 "b4:     Bdpl 4 bM 4 128 1;");
	parse_errors();
	if(time>=5){
	  ogg_script_parse("c:      Bdpc 4 4 bM 4 128 1;");
	  parse_errors();
	}
      }
      
      ogg_script_parse("Jr bailL inB;");	
      parse_errors();
      
      switch(time){
      case 2:case 3:case 4:case 5:
	ogg_script_parse("Civtl   coeffP inB 1 coeffs channels # aM;"
			 "Cinrl   coeffP inB;");
	parse_errors();
	break;
      }
      
      switch(time){
      case 1:case 2:case 3:
	ogg_script_parse("Civtl   inP inB 1 frameV channels # aM;"
			 "Cinrl   inP inB;");
	parse_errors();
	break;
      case 4:
	ogg_script_parse(
			 "Civtl   inP inB 4 frameV channels # b1M b2M b3M b4M;"
			 "Cinrl   inP inB;");
	parse_errors();
	break;
      case 5:
	ogg_script_parse(
			 "Civtl   inP inB 4 frameV channels # b1M b2M b3M b4M;"
			 "Cinc    inP inB 4 4 cM 1;");
	parse_errors();
	break;
      }

      if(polyphonic){
	
	/****** Polyphonics go here *************************************/

	ogg_script_parse("Sv    0 loopV;"
			 
	     "loopone:    Jge   loopV+1 channels outoneL;"

			 "Jr    centerL inB;"
			 /* A + B */
			 "Spush 0 fifoP;"
			 "J     firstoutL;"
		 "center: "
			 "Spush 1 fifoP;"
			 "Ecs   primeP loopV aP;"
			 "Ecs   primeP loopV+1 bP;"
			 "Eas   aP bP;"			 
			 "Jr    bzoneL inB;"
			 "Ecr   primeP loopV+1 aP;"
			 "Spush 0 fifoP;"
			 "J     firstoutL;"
		 "bzone:  Spush 1 fifoP;"
			 "Ecr   primeP loopV aP;"
		 "firstout: "
			 "Jfor loopV channels 2 looponeL;"
	     "outone:     ");
	parse_errors();
      }

      switch(time){
      case 1:
	ogg_script_parse("Ddi     inP primeP mainP;");
	break;
      case 2:case 3:case 4: case 5:
	if(quality==10)
	  ogg_script_parse("ddp     mainP primeP coeffP inP;");
	else
	  ogg_script_parse("Dp      mainP primeP coeffP inP shiftP;");
	break;
      }
      parse_errors();

      if(polyphonic){
	
	/****** Polyphonics go here *************************************/

	ogg_script_parse("Sv   0 loopV;"
			 
	     "loopone2:  Jge   loopV+1 channels outone2L;"

			 "Spull fifoP jV;"
			 "Jz   jV firstout2L;"
			 "Spull fifoP jV;"
			 "Jnz  jV bzone2L;"
			 /* A+C */
			 "Ecs   mainP loopV aP;"
			 "Ecs   mainP loopV+1 bP;"
			 "Eas   bP aP;"
			 "Ean   bP;"
			 "Ecr   mainP loopV+1 bP;"
			 "J     firstout2L;"
		 "bzone2: "
			 /* C+B */
			 "Ecs   mainP loopV aP;"
			 "Ecs   mainP loopV+1 bP;"
			 "Eaa   aP bP;"
			 "Ecr   mainP loopV aP;"
		 "firstout2: "
			 "Jfor loopV channels 2 loopone2L;"
	     "outone2:     ");
	parse_errors();
      }

      ogg_script_parse(
	       "J       finishL;"
	 "bail: Civrs   mainP dyn frameV channels aM inB;" 
       "finish: Spc     mainP primeP;"     
	       "Aod     ogg_dsp_out ogg_primary mainP rate channels dyn;"
	       "Shalt   frameV hint%");
      parse_errors();
      
      /* Pack the script into an archive header */
      ogg_script_makeheader(&h,&bytes);
      parse_errors();
      
      /* Save that header! (Catch that rabbit?) */
      blocking_write(h, 1, bytes, outfile);
      bytesout+=bytes;
	
    }
    return;
  default:
    reporterror("\nThis OggSquish release is for lossless compression only\n"
		"(options -U, -L and -Z, or -Q10 and -Q9).  Codec releases\n"
		"that include lossy compression will be available soon.\n\n");
  }
}
