/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: A.c
  function: I/O buffer routines and A module 'links'
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 11 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "bitwise.h"
#include "buffer.h"
#include "parser.h"
#include "virtual.h"
#include "A.h"

/* At the immediate moment, we only do dsp and compressed buffers.
   Text and video will need to wait until later */

/* The buffer structures and input buffer data, etc, come from
   OM_MACHINE. */

/* The A buffers serve, for the most part, only the purpose of
   attaching the external I/O buffer to an internal pipe.  Input
   buffers are sent straight to the pipe before the frame is even
   processed; output buffers read and buffer internal information at
   the appropriate point in the execution order. */

extern char *ogg_comment;
extern int OM_MACHINE;

extern dspb **__ogg_dspin;
extern dspb **__ogg_dspout;
extern bitb **__ogg_bitin;
extern bitb **__ogg_bitout;

extern long __ogg_dspinN;
extern long __ogg_dspoutN;
extern long __ogg_bitinN;
extern long __ogg_bitoutN;

/* Aid action */

static void __ogga_Aid(module *m){
  dspb *ptr=(dspb *)(m->state);
  __ogg_sendtoP(ptr->pipe,ptr->ptr);
  ptr->ptr=NULL;
}

/* soft reset */
static void __oggs_Aid(module *m){
  dspb *ptr=(dspb *)(m->state);
  __ogg_free(ptr->ptr);
  ptr->ptr=NULL;
  __ogg_resetP(ptr->pipe);
}

/* create */
void __oggc_Aid(module *m){
  int group=__ogg_fixedARG(m,0);
  int name=__ogg_fixedARG(m,1);
  dspb *ptr;
  
  switch(group){
  case 0: /* Group 0 is the only dsp input group right now */
    m->state=(ptr=__ogg_registerDSP(&__ogg_dspin,&__ogg_dspinN,name));
    break;
  default:
    __ogg_Cerror(100,1);
  }

  ptr->pipe=__ogg_fixedARG(m,2);
  ptr->hint=__ogg_fixedARG(m,3);
  ptr->rate=__ogg_fixedARG(m,4);
  ptr->channels=__ogg_fixedARG(m,5);
  ptr->dynrange=__ogg_fixedARG(m,6);
  m->action=__ogga_Aid;
  m->softreset=__oggs_Aid;
}

/* Aib action */

static void __ogga_Aib(module *m){
  bitb *ptr=(bitb *)(m->state);

  /* bitpipes don't append like mpipes do; however, postprocessing
     moves all remaining input back into temp */

  __ogg_Bfill(ptr->bpipe,ptr->ptr,ptr->bytes);
  __ogg_free(ptr->ptr);
  ptr->ptr=NULL;
  ptr->bytes=0;
}

/* soft reset */
static void __oggs_Aib(module *m){
  bitb *ptr=(bitb *)(m->state);
  __ogg_free(ptr->ptr);
  ptr->ptr=NULL;
  ptr->bytes=0;
  __ogg_Breset(ptr->bpipe);
}

/* create */
void __oggc_Aib(module *m){
  int group=__ogg_fixedARG(m,0);
  int name=__ogg_fixedARG(m,1);
  bitb *ptr;
  
  switch(group){
  case 2: /* Group 2 is the only bit input group right now */
    m->state=(ptr=__ogg_registerBIT(&__ogg_bitin,&__ogg_bitinN,name));
    break;
  default:
    __ogg_Cerror(100,1);
  }

  ptr->bpipe=__ogg_fixedARG(m,2);
  ptr->hint=__ogg_fixedARG(m,3);
  ptr->seekable=__ogg_fixedARG(m,4);
  m->action=__ogga_Aib;
  m->softreset=__oggs_Aib;
}

/* Aod action */
static void __ogga_Aod(module *m){
  dspb *ptr=(dspb *)(m->state);
  ptr->ptr=__ogg_readfromP(ptr->pipe);
  __ogg_ripmoveMB(ptr->ptr,OM_MACHINE);
}

/* soft reset */
static void __oggs_Aod(module *m){
  dspb *ptr=(dspb *)(m->state);
  __ogg_freeMB(ptr->ptr);
  ptr->ptr=NULL;
}

/* create */
void __oggc_Aod(module *m){
  int group=__ogg_fixedARG(m,0);
  int name=__ogg_fixedARG(m,1);
  dspb *ptr;
  
  switch(group){
  case 1: /* Group 1 is the only dsp output group right now */
    m->state=(ptr=__ogg_registerDSP(&__ogg_dspout,&__ogg_dspoutN,name));
    break;
  default:
    __ogg_Cerror(100,1);
  }

  ptr->pipe=__ogg_fixedARG(m,2);
  ptr->rate=__ogg_fixedARG(m,3);
  ptr->channels=__ogg_fixedARG(m,4);
  ptr->dynrange=__ogg_fixedARG(m,5);
  m->action=__ogga_Aod;
  m->softreset=__oggs_Aod;
}

/* Aob action */

static void __ogga_Aob(module *m){
  bitb *ptr=(bitb *)(m->state);
  __ogg_Bwflush(ptr->bpipe,&(ptr->ptr),&(ptr->bytes));
  __ogg_ripmove(ptr->ptr,OM_MACHINE);
}

/* soft reset */
static void __oggs_Aob(module *m){
  bitb *ptr=(bitb *)(m->state);
  __ogg_free(ptr->ptr);
  ptr->ptr=NULL;
  ptr->bytes=0;
}

/* create */
void __oggc_Aob(module *m){
  int group=__ogg_fixedARG(m,0);
  int name=__ogg_fixedARG(m,1);
  bitb *ptr;
  
  switch(group){
  case 3: /* Group 3 is the only bit output group right now */
    m->state=(ptr=__ogg_registerBIT(&__ogg_bitout,&__ogg_bitoutN,name));
    break;
  default:
    __ogg_Cerror(100,1);
  }

  ptr->bpipe=__ogg_fixedARG(m,2);
  ptr->seekable=__ogg_fixedARG(m,3);
  m->action=__ogga_Aob;
  m->softreset=__oggs_Aob;
}
