/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: C.c
  function: Toplevel structure encodings
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 06 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "bitwise.h"
#include "B.h"
#include "C.h"
#include <limits.h>
#include <math.h>
#include "ilog.h"

#define AUL (0x01000000UL)  /* DO NOT ALTER! */
#define AULSHIFT 24         /* DO NOT ALTER! */

extern int OM_MACHINE;

extern long __ogg_Bmax;
extern long __ogg_Bbits;
extern void *__ogg_Bstate;


typedef struct Cxvr{
  int pipe;
  int bpipe;

  module *backend;
  void (*write)(long);
  long (*read)(void);
} Cxvr;

static void __ogga_Covr(module *m){
  Cxvr *c=(Cxvr *)(m->state);
  void *buffer=__ogg_readfromP(c->pipe);
  int size=__ogg_sizeMB(buffer),i;
  int channels=__ogg_chanMB(buffer),ch;
  int bits=__ogg_fixedARG(m,1);
  if(bits<0)bits=0;
  
  __ogg_Bstate=c->backend->state;
  if(buffer){
    __ogg_Bset(c->bpipe);
    for(ch=0;ch<channels;ch++){
      long *ptr=__ogg_ptrMB(buffer,ch);
      __ogg_Bbits=bits;
      __ogg_Bmax=0;
      for(i=0;i<size;i++)
	(*(c->write))(ptr[i]);
    }
  }

  __ogg_appendtoP(c->pipe,buffer);
}

/* reset is a nop */
  
void __oggc_Covr(module *m){
  Cxvr *c=(Cxvr *)(m->state=__ogg_calloc(1,sizeof(Cxvr),OM_MACHINE));
  c->pipe=__ogg_fixedARG(m,0);
  c->backend=__ogg_gethandleM(__ogg_fixedARG(m,2));
  c->bpipe=__ogg_fixedARG(m,3);
  c->write=__ogg_B_write_p(c->backend);
  m->action=__ogga_Covr;
}

static void __ogga_Civrs(module *m){
  Cxvr *c=(Cxvr *)(m->state);
  int size=__ogg_fixedARG(m,2),i;
  int channels=__ogg_fixedARG(m,3),ch;
  int bits=__ogg_fixedARG(m,1);
  void *buffer;

  __ogg_resetP(c->pipe);
  if(size<=0 || channels<=0 || bits<=0) return;

  buffer=__ogg_allocMB(channels,size);
  
  __ogg_Bstate=c->backend->state;
  __ogg_Bset(c->bpipe);
  for(ch=0;ch<channels;ch++){
    long *ptr=__ogg_ptrMB(buffer,ch);
    __ogg_Bmax=0;
    __ogg_Bbits=bits;
    for(i=0;i<size;i++)
      ptr[i]=((size32)((*(c->read))()))<<(32-bits)>>(32-bits);
  }
  __ogg_appendtoP(c->pipe,buffer);

}

/* reset is a nop */
  
void __oggc_Civrs(module *m){
  Cxvr *c=(Cxvr *)(m->state=__ogg_calloc(1,sizeof(Cxvr),OM_MACHINE));
  c->pipe=__ogg_fixedARG(m,0);
  c->bpipe=__ogg_fixedARG(m,5);
  c->backend=__ogg_gethandleM(__ogg_fixedARG(m,4));
  c->read=__ogg_B_read_p(c->backend);
  m->action=__ogga_Civrs;
}

/* Conr  inP outB */

static void __ogga_Conr(module *m){
  int ip=__ogg_fixedARG(m,0);
  int ob=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);

  if(b){
    int chan=__ogg_chanMB(b),ch;
    int size=__ogg_sizeMB(b),i;
    __ogg_Bset(ob);
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
	for(i=0;i<size;i++)
	  if(ptr[i]<0)
	    __ogg_Bout(1,1);
	  else
	    __ogg_Bout(0,1);
	  
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Conr(module *m){
  m->action=__ogga_Conr;
}

/* Cinr  modP inB */

static void __ogga_Cinr(module *m){
  int ip=__ogg_fixedARG(m,0);
  int ib=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  size32 buf;

  if(b){
    int chan=__ogg_chanMB(b),ch;
    int size=__ogg_sizeMB(b),i,j;
    __ogg_Bset(ib);
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<size;){
	buf=__ogg_Blook(32);
	for(j=0;j<32 && i<size;j++){
	  if(buf&(1L<<j))ptr[i]=-ptr[i];
	  i++;
	}
	__ogg_Badv(j);
      }
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Cinr(module *m){
  m->action=__ogga_Cinr;
}

/* Conrl  inP outB */

static void __ogga_Conrl(module *m){
  int ip=__ogg_fixedARG(m,0);
  int ob=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);

  if(b){
    int chan=__ogg_chanMB(b),ch;
    int size=__ogg_sizeMB(b),i;
    __ogg_Bset(ob);
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
	for(i=0;i<size;i++)
	  if(ptr[i]<0)
	    __ogg_Bout(1,1);
	  else
	    if(ptr[i])__ogg_Bout(0,1);
	  
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Conrl(module *m){
  m->action=__ogga_Conrl;
}

/* Cinrl  modP inB */

static void __ogga_Cinrl(module *m){
  int ip=__ogg_fixedARG(m,0);
  int ib=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  size32 buf;

  if(b){
    int chan=__ogg_chanMB(b),ch;
    int size=__ogg_sizeMB(b),i,j;
    __ogg_Bset(ib);
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<size;){
	buf=__ogg_Blook(32);
	for(j=0;j<32 && i<size;){
	  if(ptr[i]){
	    if(buf&(1L<<j))ptr[i]=-ptr[i];
	    j++;
	  }
	  i++;
	}
	__ogg_Badv(j);
      }
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Cinrl(module *m){
  m->action=__ogga_Cinrl;
}

/* Unsigned op; you have to save out a negation bitstream later too */
/* This op saves out a vector encoded in a binary tree; it can only
   operate on structures that are sized in powers of two, so if the
   input is a different size, it will save int power of two blocks. */ 

/* Each 'layer' (grown from the bottom of the tree up) can have a
   limit (as a maxval) on the number of bits to be used in a delta.
   Maxvals are only useful in the range 0 <= n <= 255.
   This pays attention to overflow */

/* New restriction; the 'backends' must either be all stateless, or
   come from the same encoder 'state' */

typedef struct Cxvt{
  int pipe;
  int bpipe;
  int layers;

  void **backendstate;
  void (**write)(long);
  long (**read)(void);
  long *bits;

  long maxbits;
} Cxvt;

static int power_of_2(long in){
  long ret=1;
  int power=1;
  
  if(in<0)return 0;
  while(ret<<=1){
    if(ret==in)return(power);
    power++;
  }
  return(0);
}

static int closest_power_of_2(long in){
  long ret=(LONG_MAX>>1)+1;
  
  if(in<0)return 0;
  while(ret){
    if(ret<=in)return(ret);
    ret>>=1;
  }
  return(0);
}

/* Covt    : fulldelta vector tree; goes integer if will not overflow
             and precision is set low.  Currently obsolete except for a 
	     few clients that used prereleases in non-ogg applications.
   Covta   : same as Covt, but adptive (also obsolete)
   Covtl   : lossless encoding; part of the beta spec

   input versions mirror the above (right now, only input versions actually
   go integer; input is more time sensitive) 

   Cfvt, etc, 'fake' the transaction for limiting purposes */

static void *fill_vt_struct(module *m,int maxp){
  int op=__ogg_fixedARG(m,0);
  int ib=__ogg_fixedARG(m,1);
  int layers=__ogg_fixedARG(m,2);
  int i,varptr;
  module *backend;
  Cxvt *c=(Cxvt *)(m->state=__ogg_calloc(1,sizeof(Cxvt),OM_MACHINE));

  c->pipe=op;
  c->bpipe=ib;
  c->layers=layers;
  if(layers<1)layers=1;

  c->backendstate=__ogg_malloc(sizeof(void *)*layers,OM_MACHINE);
  c->write=__ogg_malloc(sizeof(void *)*layers,OM_MACHINE);
  c->read=__ogg_malloc(sizeof(void *)*layers,OM_MACHINE);
  c->bits=(long *)__ogg_malloc(sizeof(long)*layers,OM_MACHINE);
  
  /* Grab the various backends */

  varptr=0;
  for(i=0;i<layers;i++){
    if(__ogg_variableBOUND(m,varptr))__ogg_Cerror(102,3+i);

    backend=__ogg_gethandleM(__ogg_variableARG(m,varptr++));
    c->backendstate[layers-i-1]=backend->state;
    c->write[layers-i-1]=__ogg_B_write_p(backend);
    c->read[layers-i-1]=__ogg_B_read_p(backend);
  }
  if(maxp){

    if(!__ogg_variableBOUND(m,varptr++))__ogg_Cerror(102,3+i);
    
    for(i=0;i<layers;i++){
      if(__ogg_variableBOUND(m,varptr))__ogg_Cerror(102,3+i+layers);
      c->bits[layers-i-1]=__ogg_variableARG(m,varptr++);
    c->maxbits=c->maxbits>c->bits[layers-i-1]?c->maxbits:c->bits[layers-i-1];
    }
    
    if(!__ogg_variableEND(m,varptr++))__ogg_Cerror(102,layers+4+i);
  }else
    if(!__ogg_variableEND(m,varptr++))__ogg_Cerror(102,3+i);

  return((void *)c);
}

#ifdef PRE_COMPAT
#include "Cold.c"
#endif

/* C{i|o}vtl     dataP bpipeB layersA # backends */

static void __ogga_Covtl(module *m){
  Cxvt *c=(Cxvt *)(m->state);
  void *buf=__ogg_readfromP(c->pipe);
  int insize=__ogg_sizeMB(buf),sofar=0,size;
  int chan=__ogg_chanMB(buf),ch;
  int reallayers=ilog(insize);
  int *recallS=(int *)__ogg_malloc(sizeof(int)*reallayers,OM_TEMP);
  long *hackS=(long *)__ogg_malloc(sizeof(long)*reallayers,OM_TEMP);

  while((size=closest_power_of_2(insize))){
    reallayers=ilog(size);

    /* build then write */

    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(buf,ch)+sofar,i;
      long *work=__ogg_malloc(sizeof(double)*size*2,OM_TEMP);
      size32 power; /* signed! */

      for(i=0;i<size;i++)work[i]=labs(ptr[i]);
      
      if(size>1){
	long j,base;
	
	/* Build up the binary tree */
	
	i=size;
	base=0;
	while(i>1){
	  j=0;
	  while(j<i){
	    power=(work[base+i+(j>>1)]=(work[base+j]>work[base+j+1]?
					work[base+j]:work[base+j+1]));
	    j+=2;
	  }
	  base+=i;
	  i>>=1;
	}
      }else
	power=work[0];

      __ogg_Bset(c->bpipe);
      /* Write the power output and get ready */
    
      __ogg_Bout(ilog(power),5);
      if(power){
	__ogg_Bout(power,ilog(power));
	__ogg_Bbits=0;

	if(size>1){ /* write */
	  int current=reallayers-1,trunc,i;
	  unsigned long ihi,ilo,index=0,values;

	  for(i=0;i<reallayers;i++)hackS[i]=(size<<1)-(1<<(reallayers-i));

	  recallS[current]=-1;
	  while(1){
	    ihi=(hackS[current]++);
	    ilo=hackS[current-1];

	    trunc=(current<c->layers?current:c->layers)-1;

	    if((values=work[ihi])){
	      /* OK, we do have a value to read */
	      
	      __ogg_Bmax=values<<1;
	      __ogg_Bstate=c->backendstate[trunc];
	      if(work[ilo]==values)
		(*(c->write[trunc]))((values<<1)-work[ilo+1]);
	      else
		(*(c->write[trunc]))(work[ilo]);
	    }
	  
	    if(current>1){
	    
	      /* left tree branch */
	      /* store data */

	      recallS[current-1]=current;
	      current--;
	    
	    }else{
	      
	      /* right branch */
	      
	      hackS[current-1]+=2;
	      index+=2;
	      if(index>=size)break;

	      /* retrieve data */
	      current=recallS[current];
	      current--;
	      recallS[current]=recallS[current+1];

	    }
	  }
	}
      }
      __ogg_free(work);
    }
    insize-=size;
    sofar+=size;
  }
  __ogg_free(recallS);
  __ogg_free(hackS);
  __ogg_appendtoP(c->pipe,buf);
}

static void __ogga_Civtl(module *m){
  Cxvt *c=(Cxvt *)(m->state);
  int insize=__ogg_fixedARG(m,3),sofar=0,size;
  int chan=__ogg_fixedARG(m,4),ch;
  void *buf=__ogg_allocMB(chan,insize);
  long reallayers=ilog(insize);
  int *recallS=(int *)__ogg_malloc(sizeof(int)*reallayers,OM_TEMP);
  long *powerS=(long *)__ogg_malloc(sizeof(long)*reallayers,OM_TEMP);

  __ogg_resetP(c->pipe);
  __ogg_Bset(c->bpipe);
  __ogg_Bbits=0;

  while((size=closest_power_of_2(insize))){
    reallayers=ilog(size);

    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(buf,ch)+sofar;
      int bits=__ogg_Bread(5);
      long power;
      
      if(bits){
	power=__ogg_Bread(bits);

	if(size==1)
	  ptr[0]=power;
	else{
	  long current=reallayers-1,trunc;
	  long index=0,left,right;
	  void **backendstate=c->backendstate;
	  long (**readptrs)()=c->read;
	    
	  recallS[current]=-1;
	  while(1){
	      
	    trunc=(current<c->layers?current:c->layers)-1;
	      
	    if((__ogg_Bmax=(power<<1))){
	      /* OK, we do have a value to read */
		
	      __ogg_Bstate=backendstate[trunc];
	      left=(*(readptrs[trunc]))();
	      if(left>power){
		right=(power<<1)-left;
		left=power;
	      }else
		right=power;
		
	    }else
	      left=(right=0);
	    
	    if(current>1){
		
	      /* left tree branch */
	      /* store data */
	      
	      power=left;
	      powerS[current]=right;
	      recallS[current-1]=current;
	      current--;
	      
	    }else{
	      
	      /* Two leaves, then right branch */
	      
	      ptr[index++]=left;
	      ptr[index++]=right;
	      if(index>=size)break;
	      
	      current=recallS[current];
	      /* retrieve data */
	      power=powerS[current];
	      
	      current--;
	      recallS[current]=recallS[current+1];
	      
	    }      
	  }
	}
      }
    }
    insize-=size;
    sofar+=size;
  }
  __ogg_free(powerS);
  __ogg_free(recallS);  
  __ogg_appendtoP(c->pipe,buf);
}

void __oggc_Civtl(module *m){

  m->state=fill_vt_struct(m,0);
  m->action=__ogga_Civtl;
}

void __oggc_Covtl(module *m){

  m->state=fill_vt_struct(m,0);
  m->action=__ogga_Covtl;
}

/****************************************************************/
/* C{i|o}nc      dataP bpipeB keysizeA wordsizeA backendM compA */

/* 0 bits  < key/word size <= 8 bits */

typedef struct Cxnc{
  int pipe;
  int bpipe;

  void *backendstate;
  void (*write)(long);
  long (*read)(void);

  long codebits;
  long wordbits;
  long keymask;
  int noncompact;
} Cxnc;

static void __ogga_Conc(module *m){
  Cxnc *c=(Cxnc *)m->state;
  void *in=__ogg_readfromP(c->pipe);
  long size=__ogg_sizeMB(in),i;
  long chan=__ogg_chanMB(in),ch;

  __ogg_Bset(c->bpipe);
  __ogg_Bbits=0;
  __ogg_Bstate=c->backendstate;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(in,ch);
    unsigned long key=0;
    unsigned long val=0;
    int accbits=0;
    i=0;

    /* First, we seed */
    
    while(accbits<c->codebits && i<size){
      if(ptr[i] || c->noncompact){
	if(ptr[i]<0){
	  key=(key<<1)|1;
	  __ogg_Bout(1,1);
	}else{
	  key<<=1;
	  __ogg_Bout(0,1);
	}
	accbits++;
      }
      i++;
    }

    while(i<size){
      /* Get a word */

      accbits=0;
      val=0;
      while(i<size && accbits<c->wordbits){
	if(ptr[i] || c->noncompact){
	  if(ptr[i]<0)
	    val=(val<<1)|1;
	  else
	    val<<=1;
	  accbits++;
	}
	i++;
      }

      if(accbits<c->wordbits){
	int j;
	for(j=0;j<accbits;j++)
	  __ogg_Bout((val>>(accbits-j-1))&1,1);
      }else{
	__ogg_Bmax=key;
	(*(c->write))(val);
	key=((key<<accbits)|val)&c->keymask;
      }
    }
  }
  __ogg_appendtoP(c->pipe,in);
}

static void __ogga_Cinc(module *m){
  Cxnc *c=(Cxnc *)m->state;
  void *in=__ogg_readfromP(c->pipe);
  long size=__ogg_sizeMB(in),i,compsize;
  long chan=__ogg_chanMB(in),ch;

  __ogg_Bset(c->bpipe);
  __ogg_Bbits=0;
  __ogg_Bstate=c->backendstate;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(in,ch);
    unsigned long key=0;
    unsigned long val=0;
    int accbits=0;

    if(c->noncompact==0){
      compsize=0;
      for(i=0;i<size;i++)
	if(ptr[i])compsize++;
    }else
      compsize=size;

    /* First, we get the seed */
    
    i=0;
    while(accbits<c->codebits && i<size){
      if(ptr[i] || c->noncompact){
	if(__ogg_Bread(1)){
	  key=(key<<1)|1;
	  ptr[i]=-ptr[i];
	}else
	  key<<=1;
	accbits++;
      }
      i++;
    }

    compsize-=accbits;

    while(i<size){

      if(compsize<c->wordbits){
	/* Can only get a word piece */

	for(;i<size;i++)
	  if(ptr[i] || c->noncompact)
	    if(__ogg_Bread(1))
	      ptr[i]=-ptr[i];
       
      }else{
	/* Get a full word */
	int mask=1<<(c->wordbits-1);

	__ogg_Bmax=key;
	val=(*(c->read))();
	key=((key<<accbits)|val)&c->keymask;


	for(accbits=0;accbits<c->wordbits;i++)
	  if(ptr[i] || c->noncompact){
	    if(val&mask)
	      ptr[i]=-ptr[i];
	    mask>>=1;
	    accbits++;
	  }
      }
      
      compsize-=c->wordbits;

    }
  }
  __ogg_appendtoP(c->pipe,in);
}

static void *fill_nc_struct(module *m){
  int op=__ogg_fixedARG(m,0);
  int ib=__ogg_fixedARG(m,1);
  int key=__ogg_fixedARG(m,2);
  int word=__ogg_fixedARG(m,3);

  module *backend;
  Cxnc *c=(Cxnc *)(m->state=__ogg_calloc(1,sizeof(Cxnc),OM_MACHINE));

  c->pipe=op;
  c->bpipe=ib;
  c->codebits=key;
  c->wordbits=word;
  c->keymask=(1<<key)-1;
  c->noncompact=!__ogg_fixedARG(m,5);

  /* Grab the backend */

  backend=__ogg_gethandleM(__ogg_fixedARG(m,4));
  c->backendstate=backend->state;
  c->write=__ogg_B_write_p(backend);
  c->read=__ogg_B_read_p(backend);

  return((void *)c);
}

void __oggc_Conc(module *m){
  m->state=fill_nc_struct(m);
  m->action=__ogga_Conc;
}

void __oggc_Cinc(module *m){
  m->state=fill_nc_struct(m);
  m->action=__ogga_Cinc;
}

static void __ogga_Coa(module *m){
  long val=__ogg_fixedARG(m,0);
  long bits=__ogg_fixedARG(m,1);
  int pipeB=__ogg_fixedARG(m,2);

  __ogg_Bset(pipeB);
  __ogg_Bout(val,bits);
}

static void __ogga_Cia(module *m){
  long valV=__ogg_fixedARG(m,0);
  long bits=__ogg_fixedARG(m,1);
  int pipeB=__ogg_fixedARG(m,2);

  __ogg_Bset(pipeB);
  __ogg_writetoV(valV,__ogg_Bread(bits));
}

void __oggc_Coa(module *m){
  m->action=__ogga_Coa;
}

void __oggc_Cia(module *m){
  m->action=__ogga_Cia;
}
