/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: E.c
  function: Vector and small matrix transforms
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 04 1996

 ********************************************************************/

#include <math.h>
#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "quantize.h"
#include "E.h"

static void __ogga_Es(module *m){
  int ip=__ogg_fixedARG(m,0);
  int ap=__ogg_fixedARG(m,1);
  int bp=__ogg_fixedARG(m,2);
  long split=__ogg_fixedARG(m,3);
  void *b=__ogg_readnfromP(ip,split);
  __ogg_sendtoP(bp,__ogg_readfromP(ip));
  __ogg_sendtoP(ap,b);
}

void __oggc_Es(module *m){
  m->action=__ogga_Es;
}

static void __ogga_Ez(module *m){
  int op=__ogg_fixedARG(m,0);
  long size=__ogg_fixedARG(m,1);
  long chan=__ogg_fixedARG(m,2);
  long val=__ogg_fixedARG(m,3);
  void *b=__ogg_allocMB(chan,size);
  if(val)__ogg_setMB(b,val);
  __ogg_sendtoP(op,b);
}

void __oggc_Ez(module *m){
  m->action=__ogga_Ez;
}

static void __ogga_Ezo(module *m){
  int p=__ogg_fixedARG(m,0);
  long from=__ogg_fixedARG(m,1);
  long to=__ogg_fixedARG(m,2);
  long val=__ogg_fixedARG(m,3);
  void *b=__ogg_readfromP(p);
  long chan=__ogg_chanMB(b),ch;
  long size=__ogg_sizeMB(b),i;

  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    for(i=from;i<size && i<=to;i++)
      ptr[i]=val;
  }
  __ogg_appendtoP(p,b);
}

void __oggc_Ezo(module *m){
  m->action=__ogga_Ezo;
}

static void __ogga_Ec(module *m){
  int ap=__ogg_fixedARG(m,0);
  int bp=__ogg_fixedARG(m,1);
  int op=__ogg_fixedARG(m,2);
  void *temp=__ogg_readfromP(bp);
  __ogg_sendtoP(op,__ogg_readfromP(ap));
  __ogg_appendtoP(op,temp);
}

void __oggc_Ec(module *m){
  m->action=__ogga_Ec;
}

static void __ogga_Eq(module *m){
  int p=__ogg_fixedARG(m,0);
  int bits=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(p);
  int chan=__ogg_chanMB(b),ch;
  long size=__ogg_chanMB(b);

  if(bits<1 || bits>31)return;
  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    switch(bits){
    case 1:
      __ogg_msb_quantize(ptr,size);
      break;
    default:
      __ogg_msb_quantizeN(ptr,size,bits);
      break;
    }
  }
  __ogg_appendtoP(p,b);
}

void __oggc_Eq(module *m){
  m->action=__ogga_Eq;
}

static void __ogga_Eqi(module *m){
  int p=__ogg_fixedARG(m,0);
  int bits=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(p);
  int chan=__ogg_chanMB(b),ch;
  long size=__ogg_chanMB(b);

  if(bits<1 || bits>31)return;
  for(ch=0;ch<chan;ch++){
    long *ptr=__ogg_ptrMB(b,ch);
    switch(bits){
    case 1:
      __ogg_msb_unquantize(ptr,size);
      break;
    default:
      __ogg_msb_unquantizeN(ptr,size,bits);
      break;
    }
  }
  __ogg_appendtoP(p,b);
}

void __oggc_Eqi(module *m){
  m->action=__ogga_Eqi;
}

static void __ogga_El(module *m){
  int ref=__ogg_fixedARG(m,0);
  int data=__ogg_fixedARG(m,1);
  void *refbuf=__ogg_readfromP(ref);
  void *buf=__ogg_readfromP(data);
  int chan=__ogg_chanMB(refbuf),ch;
  long size=__ogg_sizeMB(refbuf),i;
  int chan2=__ogg_chanMB(buf);
  long size2=__ogg_sizeMB(buf);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *refptr=__ogg_ptrMB(refbuf,ch);
    long *ptr=__ogg_ptrMB(buf,ch);

    for(i=0;i<size;i++)
      if(ptr[i]<0){
	if(-ptr[i]>labs(refptr[i]))ptr[i]=-labs(refptr[i]);
      }else{
	if(ptr[i]>labs(refptr[i]))ptr[i]=labs(refptr[i]);
      }
  }
  __ogg_appendtoP(ref,refbuf);
  __ogg_appendtoP(data,buf);
}

void __oggc_El(module *m){
  m->action=__ogga_El;
}

static void __ogga_Edi(module *m){
  int in=__ogg_fixedARG(m,0);
  int out=__ogg_fixedARG(m,1);
  int drift=__ogg_fixedARG(m,2);
  void *b=__ogg_readfromP(in);
  void *d=__ogg_readfromP(drift);
  int chan=__ogg_chanMB(b),ch;
  long size=__ogg_sizeMB(b),i;
  void *o=__ogg_allocMB(chan,size);

  if(b && d){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      long *optr=__ogg_ptrMB(o,ch);
      long drift=__ogg_ptrMB(d,ch)[0];
      long next;

      optr[0]=drift-ptr[size-1];
      next=2;

      while(next<=size){
	for(i=next-1;i>2;i-=2){
	  optr[i]=optr[i>>1];
	  optr[i-1]=(optr[i]+optr[(i>>1)-1])>>1;
	}
	optr[1]=optr[0];
	optr[0]>>=1;

	next<<=1;
      }

      for(i=0;i<size;i++)
	optr[i]+=ptr[i];

    }
  }
  __ogg_appendtoP(in,b);
  __ogg_appendtoP(drift,d);
  __ogg_sendtoP(out,o);
}

void __oggc_Edi(module *m){
  m->action=__ogga_Edi;
}

static void __ogga_Eaa(module *m){
  int outP=__ogg_fixedARG(m,0);
  int inP=__ogg_fixedARG(m,1);
  void *out=__ogg_readfromP(outP);
  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;
  int chan2=__ogg_chanMB(in);
  long size2=__ogg_sizeMB(in);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);

    for(i=0;i<size;i++)
      outptr[i]+=inptr[i];
  }
  __ogg_appendtoP(outP,out);
  __ogg_appendtoP(inP,in);
}

void __oggc_Eaa(module *m){
  m->action=__ogga_Eaa;
}

static void __ogga_Eas(module *m){
  int outP=__ogg_fixedARG(m,0);
  int inP=__ogg_fixedARG(m,1);
  void *out=__ogg_readfromP(outP);
  void *in=__ogg_readfromP(inP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;
  int chan2=__ogg_chanMB(in);
  long size2=__ogg_sizeMB(in);

  if(chan2<chan)chan=chan2;
  if(size2<size)size=size2;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);
    long *inptr=__ogg_ptrMB(in,ch);

    for(i=0;i<size;i++)
      outptr[i]-=inptr[i];
  }
  __ogg_appendtoP(outP,out);
  __ogg_appendtoP(inP,in);
}

void __oggc_Eas(module *m){
  m->action=__ogga_Eas;
}

static void __ogga_Ean(module *m){
  int outP=__ogg_fixedARG(m,0);
  void *out=__ogg_readfromP(outP);
  int chan=__ogg_chanMB(out),ch;
  long size=__ogg_sizeMB(out),i;

  for(ch=0;ch<chan;ch++){
    long *outptr=__ogg_ptrMB(out,ch);

    for(i=0;i<size;i++)
      outptr[i]= -outptr[i];
  }
  __ogg_appendtoP(outP,out);
}

void __oggc_Ean(module *m){
  m->action=__ogga_Ean;
}

/* Split a single channel vector out of a multichannel vector pack */

static void __ogga_Ecs(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int channelp=__ogg_fixedARG(m,1);
  int outP=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);
  int chan=__ogg_chanMB(data);
  long size=__ogg_sizeMB(data);
  void *out=__ogg_allocMB(1,size);

  if(channelp>=0 && channelp<chan){
    long *inptr=__ogg_ptrMB(data,channelp);
    long *outptr=__ogg_ptrMB(out,0);
    memcpy(outptr,inptr,size*sizeof(long));
  }
  __ogg_appendtoP(dataP,data);
  __ogg_sendtoP(outP,out);
}

void __oggc_Ecs(module *m){
  m->action=__ogga_Ecs;
}

/* replace a channel in an input bundle with data from an input */

static void __ogga_Ecr(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int channelp=__ogg_fixedARG(m,1);
  int inP=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);
  void *in=__ogg_readfromP(inP);

  int chan=__ogg_chanMB(data);
  long size=__ogg_sizeMB(data);
  long size2=__ogg_sizeMB(in);
  if(size2<size)size=size2;

  if(channelp>=0 || channelp<chan){
    long *dataptr=__ogg_ptrMB(data,channelp);
    long *inptr=__ogg_ptrMB(in,0);
    memcpy(dataptr,inptr,size*sizeof(long));
  }
  __ogg_appendtoP(dataP,data);
  __ogg_appendtoP(inP,in);
}

void __oggc_Ecr(module *m){
  m->action=__ogga_Ecr;
}

static void __ogga_Eav(module *m){
  int dataP=__ogg_fixedARG(m,0);
  int ch=__ogg_fixedARG(m,1);
  int outV=__ogg_fixedARG(m,2);

  void *data=__ogg_readfromP(dataP);

  int chan=__ogg_chanMB(data);
  long size=__ogg_sizeMB(data),i;

  double acc=0;
  
  if(ch>=0 && ch<chan){
    long *ptr=__ogg_ptrMB(data,ch);
    for(i=0;i<size;i++)
      acc+=labs(ptr[i]);
  }

  if(size>0)
    acc/=size;
  
  __ogg_appendtoP(dataP,data);
  __ogg_writetoV(outV,(long)rint(acc));
}

void __oggc_Eav(module *m){
  m->action=__ogga_Eav;
}

