/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: J.c
  function: Branching and flow control modules
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 04 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "bitwise.h"
#include "J.h"

/*      
	J      jump
	Jle    jump if <=
        Je     jump if ==
	Jnz    jump if != 0
	etc...

	Jwle   jump and write 1 bit if <=   (else write 0)
	Jwe    jump and write 1 bit if ==   ""
	Jwnz   jump and write 1 bit if != 0 ""
	etc...

	Jr     read bit, jump if set

	Jfor   aV bA locL;   test A < B, jump if true, inc A
*/

static void __ogga_J(module *m){
  int inL=__ogg_fixedARG(m,0);
  __ogg_virtual_machine_jump(inL);
}

void __oggc_J(module *m){
  m->action=__ogga_J;
}

static void __ogga_Jle(module *m){
  long a=__ogg_fixedARG(m,0);
  long b=__ogg_fixedARG(m,1);
  int inL=__ogg_fixedARG(m,2);
  if(a<=b)__ogg_virtual_machine_jump(inL);
  if(m->name[1]=='w'){
    __ogg_Bset(__ogg_fixedARG(m,3));
    if(a<=b)
      __ogg_Bout(1,1);
    else
      __ogg_Bout(0,1);
  }
}

static void __ogga_Jl(module *m){
  long a=__ogg_fixedARG(m,0);
  long b=__ogg_fixedARG(m,1);
  int inL=__ogg_fixedARG(m,2);
  if(a<b)__ogg_virtual_machine_jump(inL);
  if(m->name[1]=='w'){
    __ogg_Bset(__ogg_fixedARG(m,3));
    if(a<b)
      __ogg_Bout(1,1);
    else
      __ogg_Bout(0,1);
  }
}

static void __ogga_Jge(module *m){
  long a=__ogg_fixedARG(m,0);
  long b=__ogg_fixedARG(m,1);
  int inL=__ogg_fixedARG(m,2);
  if(a>=b)__ogg_virtual_machine_jump(inL);
  if(m->name[1]=='w'){
    __ogg_Bset(__ogg_fixedARG(m,3));
    if(a>=b)
      __ogg_Bout(1,1);
    else
      __ogg_Bout(0,1);
  }
}

static void __ogga_Jg(module *m){
  long a=__ogg_fixedARG(m,0);
  long b=__ogg_fixedARG(m,1);
  int inL=__ogg_fixedARG(m,2);
  if(a>b)__ogg_virtual_machine_jump(inL);
  if(m->name[1]=='w'){
    __ogg_Bset(__ogg_fixedARG(m,3));
    if(a>b)
      __ogg_Bout(1,1);
    else
      __ogg_Bout(0,1);
  }
}

static void __ogga_Je(module *m){
  long a=__ogg_fixedARG(m,0);
  long b=__ogg_fixedARG(m,1);
  int inL=__ogg_fixedARG(m,2);
  if(a==b)__ogg_virtual_machine_jump(inL);
  if(m->name[1]=='w'){
    __ogg_Bset(__ogg_fixedARG(m,3));
    if(a==b)
      __ogg_Bout(1,1);
    else
      __ogg_Bout(0,1);
  }
}

static void __ogga_Jne(module *m){
  long a=__ogg_fixedARG(m,0);
  long b=__ogg_fixedARG(m,1);
  int inL=__ogg_fixedARG(m,2);
  if(a!=b)__ogg_virtual_machine_jump(inL);
  if(m->name[1]=='w'){
    __ogg_Bset(__ogg_fixedARG(m,3));
    if(a!=b)
      __ogg_Bout(1,1);
    else
      __ogg_Bout(0,1);
  }
}


static void __ogga_Jz(module *m){
  long a=__ogg_fixedARG(m,0);
  int inL=__ogg_fixedARG(m,1);
  if(a==0)__ogg_virtual_machine_jump(inL);
  if(m->name[1]=='w'){
    __ogg_Bset(__ogg_fixedARG(m,2));
    if(a==0)
      __ogg_Bout(1,1);
    else
      __ogg_Bout(0,1);
  }
}

static void __ogga_Jnz(module *m){
  long a=__ogg_fixedARG(m,0);
  int inL=__ogg_fixedARG(m,1);
  if(a)__ogg_virtual_machine_jump(inL);
  if(m->name[1]=='w'){
    __ogg_Bset(__ogg_fixedARG(m,2));
    if(a)
      __ogg_Bout(1,1);
    else
      __ogg_Bout(0,1);
  }
}

static void __ogga_Jr(module *m){
  int inL=__ogg_fixedARG(m,0);
  __ogg_Bset(__ogg_fixedARG(m,1));

  if(__ogg_Bread(1))__ogg_virtual_machine_jump(inL);
}

void __oggc_Jle(module *m){
  m->action=__ogga_Jle;
}

void __oggc_Jl(module *m){
  m->action=__ogga_Jl;
}

void __oggc_Jge(module *m){
  m->action=__ogga_Jge;
}

void __oggc_Jg(module *m){
  m->action=__ogga_Jg;
}

void __oggc_Je(module *m){
  m->action=__ogga_Je;
}

void __oggc_Jne(module *m){
  m->action=__ogga_Jne;
}

void __oggc_Jz(module *m){
  m->action=__ogga_Jz;
}

void __oggc_Jnz(module *m){
  m->action=__ogga_Jnz;
}

void __oggc_Jr(module *m){
  m->action=__ogga_Jr;
}

/* Jfor   aV bA locL;   test A < B, jump if true, inc A */

static void __ogga_Jfor(module *m){
  long aV=__ogg_fixedARG(m,0),a;
  long b=__ogg_fixedARG(m,1);
  long step=__ogg_fixedARG(m,2);
  int jumpL=__ogg_fixedARG(m,3);

  if((a=(__ogg_readfromV(aV)+step))<b)__ogg_virtual_machine_jump(jumpL);
  __ogg_writetoV(aV,a);
}

void __oggc_Jfor(module *m){
  m->action=__ogga_Jfor;
}
