/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: W.c
  function: 1D vector windowing transforms
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jul 07 1996

 ********************************************************************/

#include <math.h>
#include "ogg.h"
#include "memory.h"
#include "parser.h"
#include "virtual.h"
#include "W.h"

static float oPI2 = 6.28318530717958647693;
static float oPI  = 3.14159265358979323846;

static void __ogga_Whl(module *m){
  int ip=__ogg_fixedARG(m,0);
  int amount=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;
  int actual;
  
  if(amount<0)amount=size>>1;
  actual=amount>size?size:amount;
  
  amount++;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<actual;i++)
	ptr[i]*=.5+cos(oPI*(1.-(i+1.)/amount))*.5;
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Whl(module *m){
  m->action=__ogga_Whl;
}

static void __ogga_Whr(module *m){
  int ip=__ogg_fixedARG(m,0);
  int amount=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;
  int actual;
  
  if(amount<0)amount=size>>1;
  actual=amount>size?size:amount;
  
  amount++;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<actual;i++)
	ptr[size-i-1]*=.5+cos(oPI*(1.-(i+1.)/amount))*.5;
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Whr(module *m){
  m->action=__ogga_Whr;
}

static void __ogga_Whli(module *m){
  int ip=__ogg_fixedARG(m,0);
  int amount=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;
  int actual;
  
  if(amount<0)amount=size>>1;
  actual=amount>size?size:amount;
  
  amount++;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<actual;i++)
	ptr[i]/=.5+cos(oPI*(1.-(i+1.)/amount))*.5;
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Whli(module *m){
  m->action=__ogga_Whli;
}

static void __ogga_Whri(module *m){
  int ip=__ogg_fixedARG(m,0);
  int amount=__ogg_fixedARG(m,1);
  void *b=__ogg_readfromP(ip);
  long size=__ogg_sizeMB(b),i;
  long chan=__ogg_chanMB(b),ch;
  int actual;
  
  if(amount<0)amount=size>>1;
  actual=amount>size?size:amount;
  
  amount++;

  if(b){
    for(ch=0;ch<chan;ch++){
      long *ptr=__ogg_ptrMB(b,ch);
      for(i=0;i<actual;i++)
	ptr[size-i-1]/=.5+cos(oPI*(1.-(i+1.)/amount))*.5;
    }
    __ogg_appendtoP(ip,b);
  }
}

void __oggc_Whri(module *m){
  m->action=__ogga_Whri;
}
