/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 *                                                                  *
 ********************************************************************

  file: lpc.c
  function: LPC low level routines
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 11 1996

 ********************************************************************/

/* Some of these routines (autocorrelator, LPC coefficient estimator)
   are directly derived from and/or modified from code written by
   Jutta Degener and Carsten Bormann; thus we include their copyright
   below.  The entirety of this file is freely redistributable on the
   condition that both of these copyright notices are preserved
   without modification.  */

/* Preserved Copyright: *********************************************/

/* Copyright 1992, 1993, 1994 by Jutta Degener and Carsten Bormann,
Technische Universita"t Berlin

Any use of this software is permitted provided that this notice is not
removed and that neither the authors nor the Technische Universita"t
Berlin are deemed to have made any representations as to the
suitability of this software for any purpose nor are held responsible
for any defects of this software. THERE IS ABSOLUTELY NO WARRANTY FOR
THIS SOFTWARE.

As a matter of courtesy, the authors request to be informed about uses
this software has found, about bugs in this software, and about any
improvements that may be of general interest.

Berlin, 28.11.1994
Jutta Degener
Carsten Bormann

*********************************************************************/

#define NEED_ILOG
#include "ogg.h"
#include "memory.h"
#include "lpc.h"

/* Algorithm Invented by N. Levinson in 1947, modified by J. Durbin in
   1959. */

double levinson_durbin(double *autoc,double *lpccoeff,int p){
  /*  in: [0...p] autocorrelation values */
  /* out: [0...p-1] LPC coefficients + a[0] constant as return val */
  int i,j;  
  double r,*ref,error=autoc[0];

  if(autoc[0]==0)return 0;

  ref=(double *)__ogg_malloc(sizeof(double)*p,OM_TEMP);
  for(i=0;i<p;i++){

    /* Sum up this iteration's reflection coefficient. */
    r=-autoc[i+1];
    for(j=0;j<i;j++)r-=lpccoeff[j]*autoc[i-j];
    ref[i]=(r/=error);    

    /* Update LPC coefficients and total error. */
    lpccoeff[i]=r;
    for(j=0;j<(i>>1);j++){
      double tmp=lpccoeff[j];
      lpccoeff[j]+=r*lpccoeff[i-1-j];
      lpccoeff[i-1-j]+=r*tmp;
    }
    if(i%2)lpccoeff[j]+=lpccoeff[j]*r;

    error*=1.0-r*r;
  }
  __ogg_free(ref);
  return error;
}

/* Compute the autocorrelation for lags between 0 and lag-1, and x==0
   outside 0...n-1 */

void autocorrelation(int n, long *x, int lag, double *ac){
  /*  in: [0...n-1] samples x   */
  /* out: [0...lag-1] ac values */
  double d; 
  int i;
  while(lag--){
    for(i=lag,d=0;i<n;i++)d+=(double)x[i]*x[i-lag];
    ac[lag]=d;
  }
}

