/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: ogg.h
  function: toplevel internal header for libogg
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jul 31 1996

 ********************************************************************/

#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <setjmp.h>
#include <limits.h>

#if defined ( WIN32 ) || defined ( MACVER )
#define size16 short
#define size32 long
#define size64 LONGLONG
#endif

#ifdef __BOUNDS_CHECKING_ON
/* Insert your bounds gcc include dir here */
#  include</usr/lib/gcc-lib/i486-linux/2.7.2/bounds/run-includes/unchecked.h>
#endif

#ifdef __GNUC__
#  define STIN static inline
#else
#  define STIN static
#endif
#  define lDEC_MAX 1000000000

/* Do we report things like ripcording nonempty blocks? */
#undef DEBUG

#define MAJOR 0
#define MINOR 98
#define REVISION 7

/* Memory block groups for ripcords */

#define NOGROUP    -1
#define OM_TEMP    0

#define OM_PARSE   1
#define OM_ASSEM   2
#define OM_BANK    3

#ifdef SELFSUFF
#  define OM_PIPE    4
#  define OM_VAR     5
#  define OM_BPIPE   6
#  define OM_HUFF    7
#  define OM_MACHINE 8
#else
#  define OM_BASE    4

#  define POM_PIPE    0
#  define POM_VAR     1
#  define POM_BPIPE   2
#  define POM_HUFF    3
#  define POM_MACHINE 4

#  define OM_OFF     5
#endif

struct ogg_settings{
  int ogg_memory;
  int ogg_precision;
};

struct ogg_dspinfo{
  long rate;
  long channels;
  long dynrange;
};

struct ogg_bitinfo{
  long seekable;
};

typedef struct module{
  char *name;
  char *argtypes;
  char **fixed;
  char **variable;

  void *state;  /* module state and configuration */
  void (*action)(struct module *m); 
                /* What action to call when it's the module's turn */
  void (*softreset)(struct module *m); 
                /* What to call when reinitializing the module */
  int handle;
} module;

#define ogg_dsp_in     0
#define ogg_dsp_out    1
#define ogg_bit_in     2
#define ogg_bit_out    3
#define ogg_vid_in     4
#define ogg_vid_out    5
#define ogg_txt_in     6
#define ogg_txt_out    7
#define ogg_primary    0
#define ogg_secondary1 1
#define ogg_secondary2 2
#define ogg_secondary3 3
#define ogg_secondary4 4
#define ogg_secondary5 5
#define ogg_secondary6 6
#define ogg_secondary7 7
#define ogg_secondary8 8
#define ogg_secondary9 9

#define OGG_LITTLEENDIAN 0
#define OGG_BIGENDIAN 1
#define OGG_SAMEENDIAN -1
#define OGG_SWAPENDIAN 2

extern void  ogg_script_reset(void);
extern void  ogg_script_parse(char *script);
extern void  ogg_script_use(int scriptnum);
extern void  ogg_script_makeheader(char **ret,long *bytes);
extern long  ogg_script_loadheader(char *header,long bytes);

extern void ogg_settings_get(struct ogg_settings *s);
extern void ogg_settings_submit(struct ogg_settings *s);

extern int ogg_buffer_used(int group,int name);
extern void ogg_buffer_dsp(int group,int name,struct ogg_dspinfo *s);
extern void ogg_buffer_bit(int group,int name,struct ogg_bitinfo *s);
extern long ogg_input_query(int group,int name);

extern void ogg_input_reset(int group,int name);
extern void ogg_input_dsp(int group,int name,void *data,long elements,
			  int width,int endian, int sign, int scale);
extern void ogg_output_dsp(int group,int name, void **data, long *elements, 
			   int width,int endian, int sign, int scale);
extern void ogg_input_bit(int group,int name, char *data, long elements);
extern void ogg_output_bit(int group,int name, char **data, long *elements);

extern void ogg_run_set(int num);
extern void ogg_run(void);
extern void ogg_run_reset(void);
extern void ogg_run_clear(void);
