/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: parser.c
  function: Script and header handling routines
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 13 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "bitwise.h"
#include "registry.h"
#include "strerrs.h"
#include "buffer.h"
#include "virtual.h"
#include "parser.h"

extern int ogg_errno;
extern char *ogg_strerr;
extern jmp_buf __ogg_jmp_env;
extern int __ogg_jmp_env_set;

extern int ogg_script_ready;
extern int __ogg_script_complete;

extern int ogg_script_err_line;
extern int ogg_script_err_statement;
extern int ogg_script_err_argument;
extern int ogg_script_err_number;
extern char *ogg_script_err_string;

extern int ogg_run_ready;
extern int ogg_run_scriptnum;

int ogg_bitstream_major=-1;
int ogg_bitstream_minor=-1;
int ogg_bitstream_revision=-1;
char *ogg_bitstream_vendor=NULL;

extern int OM_PIPE;
extern int OM_VAR;
extern int OM_BPIPE;
extern int OM_HUFF;
extern int OM_MACHINE;

/* String support for the parser */

static void strcatA(char **in,char *add){
  if(*in==NULL){
    *in=(char *)__ogg_calloc(strlen(add)+1,1,OM_TEMP);
  }else{
    *in=(char *)__ogg_realloc(*in,strlen(*in)+strlen(add)+1,OM_TEMP);
  }
  strcat(*in,add);
}

/* Good for MIN_LONG+1 to MAX_LONG */
static void *aA(char **ret,long in){
  long try=lDEC_MAX,num,digits=0;
  char buffer[80];

  buffer[0]='\0';
  if(in<0){
    in= -in;
    buffer[digits++]='-';
  }

  for(;try>in && try>1;try/=10);

  for(;try>0;try/=10){
    for(num=0;in>=try;num++)
      in-=try;
    buffer[digits++]=(char)(48+num);
  }
  buffer[digits]='\0';
  strcatA(ret,buffer);
  return(ret);
}

static void strncatA(char **in,char *add,int num){
  if(*in==NULL){
    *in=(char *)__ogg_calloc(num+1,1,OM_TEMP);
  }else{
    *in=(char *)__ogg_realloc(*in,strlen(*in)+num+1,OM_TEMP);
  }
  strncat(*in,add,num);
}

/****  The parser/assembler  *********************************************/

/* Parsing makes a first pass to parse and forward resolve some
   symbols, then a second pass to replace symbols and simplify algebra */

static char *O_partial=NULL;
static char **O_lines=NULL;

static char **O_vars=NULL;
static char **O_pipes=NULL;
static char **O_bpipes=NULL;
static char **O_labels=NULL;
static long  *O_locs=NULL;

static char **Vd=NULL; /* global macro storage */
static char **Ve=NULL;

typedef struct parsed{
  char *name;      /* this will be a mod name *or* a redirect */
  char **fixed;    /* We can use the list handlers */
  char **variable;
  mod *regentry;
} parsed;

static parsed **O_link=NULL;
static long numparsed=0;

static int listsize(char **list){
  int count=0;
  if (list==NULL)return(0);
  while(list[count]!=NULL)count++;
  return(count);
}

static int searchlist(char **list,char *key){
  int count,size=listsize(list);
  for(count=0;count<size;count++)
    if(!strcmp(list[count],key))return(count);
  return(-1);
}

/* This allocs everything from grp; make sure assembled mem is
 in OM_ASSEM */
static int expandlist(char ***list,char *newentry,int grp){
  int size=listsize(*list);
  *list=(char **)__ogg_realloc(*list,sizeof(char *)*(size+2),grp);
  (*list)[size]=(char *)__ogg_calloc(strlen(newentry)+1,1,grp);
  strcpy((*list)[size],newentry);
  (*list)[size+1]=NULL;
  return(size);
}

static int expandlistN(long **list,long newentry){
  int size=0;

  if (*list!=NULL)
    while((*list)[size]!=-1)size++;

  *list=(long *)__ogg_realloc(*list,sizeof(long)*(size+2),OM_PARSE);
  (*list)[size]=newentry;
  (*list)[size+1]=-1;
  return(0);
}

static int registerlist(char ***list,char *newkey){
  int temp;
  temp=searchlist(*list,newkey);
  if(temp!=-1)return(temp);
  return(expandlist(list,newkey,OM_PARSE));
}

/* Don't need freelist anymore, we pull the ripcords */

static void Trim(char *i){
  int begin,end,length=strlen(i);

  for(begin=0;begin<length;begin++)
    if(!isspace(i[begin]))break;

  for(end=length;end>0;end--)
    if(!isspace(i[end-1]))break;
  
  memmove(i,i+begin,end-begin);
  i[end-begin]='\0';
}

/*static void RTrim(char *i){
  int end,length=strlen(i);

  for(end=length;end>0;end--)
    if(!isspace(*(i+end-1)))break;
  *(i+end)='\0';
}*/

/* Tokens are all temp space */
static char *tokenR(char **i){
  char *ret,temp;
  int index,index2;
  if(*i==NULL)return(0);
  index=strspn(*i," \n\f\t\r\v\"");

  /* We can't put a quote and spaces in a string at the same time in a
     script.  Eit.  fix it if it ever becomes a problem. */
  if(index && (*i)[index-1]=='\"'){
    index2=strcspn((*i)+index,"\"");
    index--;
    index2+=2;
  }else
    index2=strcspn((*i)+index," \n\f\t\r\v");
  if(index2==0)return(NULL);
  temp=(*i)[index+index2];
  (*i)[index+index2]='\0';
  ret=(char *)__ogg_calloc(strlen((*i)+index)+1,1,OM_TEMP);
  strcpy(ret,(*i)+index);
  (*i)[index+index2]=temp;
  (*i)+=index+index2;
  return(ret);
}

static char *tokenS(char **i){
  char *ret=tokenR(i);
  if(ret==NULL)return(NULL);
  if(ret[0]=='"'){
    memmove(ret,ret+1,strlen(ret));
    ret[strlen(ret)-1]='\0';
  }
  return(ret);
}

static int line=-1;
static int statement=-1;
static int argument=-1;

static void VOMerror(int num){
  /* Set the error flags */
  ogg_errno=100;
  ogg_strerr=top100_errors[ogg_errno-100];
  ogg_script_err_number=num;
  ogg_script_err_line=line;
  ogg_script_err_statement=statement;
  ogg_script_err_argument=argument;
  ogg_script_err_string=parse_errors[num];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
  /* Toplevel cleans house */
}

void __ogg_inputTOlines(char *in){
  char *delim,*temp,*token,*token2,*ptr;
  line=0;
  statement=-1;
  /* look for a line delimiter */

  while((delim=strpbrk(in,";%"))!=NULL){
    argument=0;
    line++;
    if(O_partial!=NULL){
      temp=(char *)__ogg_calloc(strlen(O_partial)+(delim-in)+1,1,OM_TEMP);
      strcat(temp,O_partial);
      __ogg_free(O_partial);
      O_partial=NULL;
      strncat(temp,in,(delim-in));
    }else{
      temp=(char *)__ogg_calloc((delim-in)+1,1,OM_TEMP);
      strncat(temp,in,(delim-in));
    }

    /* We need to expand and eliminate = lines */
    ptr=temp;
    token=tokenS(&ptr);
    if(token!=NULL && !strcmp(token,"=")){
      __ogg_free(token);
      token=tokenS(&ptr);
      argument++;
      if(token==NULL)
	VOMerror(1);
      token2=tokenS(&ptr);
      argument++;
      if(token2==NULL)
	VOMerror(1);
      if(searchlist(Vd,token)!=-1)
	VOMerror(2);
      expandlist(&Vd,token,OM_PARSE);
      __ogg_free(token);
      expandlist(&Ve,token2,OM_PARSE);
      __ogg_free(token2);
    }else{
      statement++;
      __ogg_free(token);
      expandlist(&O_lines,temp,OM_PARSE); /* lines are OM_PARSE */
    }
    __ogg_free(temp);
    if(*delim=='%'){
      if(*(delim+1)!='\0'){
	argument=0;
	VOMerror(3);
      }
      __ogg_script_complete=-1;
      statement=-1;
      argument=-1;
      line=-1;
      return;
    }
    in=delim+1;
  }
  if(*in!='\0'){
    strcatA(&O_partial,in); /* strcatA allocs from OM_TEMP */
    __ogg_ripmove(O_partial,OM_PARSE);
  }
  return;
}

/** look up a module entry in the registry */

static mod *findmodule(char *module){
  mod *ptr=registry;
  while(ptr->name!=NULL){
    if(!strcmp(ptr->name,module))
      return(ptr);
    ptr++;
  }
  return(NULL);
}

/* expand a macro.  nop if the token is not a macro, frees the input
   and returns the expansion if it is */

static char *subDEF(char *token){
  int lookup;
  
  if(token==NULL)return(NULL);

  if(token[0]=='<'){
    lookup=searchlist(Vd,token+1);
    if(lookup!=-1){
      __ogg_free(token);
      token=NULL;
      strcatA(&token,"<");
      strcatA(&token,Ve[lookup]);
      return(token);
    }else
      return(token);
  }else{
    lookup=searchlist(Vd,token);
    if(lookup!=-1){
      __ogg_free(token);
      token=NULL;
      strcatA(&token,Ve[lookup]);
      return(token);
    }else
      return(token);
  }
}

/* We need to pick though an arithmetic/typed expression, peeling out
   alphanumeric strings, and macro matching without types; also don't
   match within "".  We don't this in a recursive 'evaluation' style
   step as ogg macro defs are specified to not be recursive. */

/* Later we will need to actually substitute the symbols, simplify the
   algebra (simple collapsing of binary expressions that can be
   evaluated) and evaluate the algebra */

/* Operates in OM_TEMP */
static char *arithDEF(char *token){
  char *ret=NULL,*pre=token,*post=token,*temp,suffix[2];

  if(pre==NULL)return(NULL);
  while(*pre!='\0'){
    /* Get past any symbols; pay particular attention to double quotes */
    while(*post!='\0'){
      if(*post=='"')
	while(*(++post)!='\0')
	  if(*post=='"')break;
      if(*post=='\0' || isalnum(*post) || *post=='_'){
	if(pre!=post)
	  strncatA(&ret,pre,post-pre);
	pre=post;
	break;
      }
      post++;
    }

    /* we're either at the end, or at a name */
    while(*post!='\0'){
      if((!isalnum(*post)) && !(*post=='_'))break;
      post++;
    }
    if(pre!=post){
      /* make our token, strip off the type */
      temp=NULL;
      strncatA(&temp,pre,post-pre);
      suffix[0]=temp[strlen(temp)-1];
      switch(suffix[0]){
      case 'M':case 'P': case 'V':case 'L':case 'B':
	temp[strlen(temp)-1]='\0';
	temp=subDEF(temp);
	suffix[1]='\0';
	strcatA(&temp,suffix);
	break;
      default:
	temp=subDEF(temp);
	break;
      }
      strcatA(&ret,temp);
      __ogg_free(temp);
    }
    pre=post;
  }
  __ogg_free(token);
  return(ret);
}

/* substitute symbols out of arith expression */
/* pre-evaluate simple numeric expressions out of the algebra */
/* Aw, heck... do the variable name subs here too */
/* returns 0 for a scalar, -1 if there's still symbol content */
static int simplify_algebra(char *in,char **out,long chars){
  int parenlevel=0;
  int position=-1;
  int priority=9;  /*  *5 /5 +4 -4 <3 >3 &2 ^1 |0 */
  int unaryflag=-1;
  int count=0;
  long lookup;
  Trim(in);
  for(count=0;count<chars;count++){
    if(in[count]=='(')parenlevel++;
    if(in[count]==')'){
      parenlevel--;
      if(parenlevel<0)
	VOMerror(6); /* can only happen in parsing... */
      unaryflag=0;
      continue;
    }
    if(!parenlevel){
      if(!isspace(in[count])){
	if(isalnum(in[count]))
	  unaryflag=0;
	else{
	  switch(in[count]){
	  case '|':
	    if(!unaryflag){
	      if(priority>=0)
		position=count;
	      priority=1;
	    }else
	      VOMerror(4);
	    break;
	  case '^':
	    if(!unaryflag){
	      if(priority>=1)
		position=count;
	      priority=1;
	    }else
	      VOMerror(4);
	    break;
	  case '&':
	    if(!unaryflag){
	      if(priority>=2)
	      position=count;
	      priority=2;
	    }else
	      VOMerror(4);
	    break;
	  case '<':case '>':
	    if(!unaryflag){
	      if(priority>=3)
	      position=count;
	      priority=3;
	    }else
	      VOMerror(4);
	    break;
	  case '+':case '-':
	    if(!unaryflag){
	      if(priority>=4)
	      position=count;
	      priority=4;
	    }
	    break;
	  case '*':case '/':
	    if(!unaryflag){
	      if(priority>=5)
	      position=count;
	      priority=5;
	    }else
	      VOMerror(4);
	    break;
	  case '_':
	    unaryflag=0;
	    break;
	  default:
	    VOMerror(4);
	  }
	}
      }
    }
  }
  if(parenlevel!=0)
    VOMerror(6);

  if(position==-1){
    switch(in[0]){
    case '(':
      if(in[chars-1]==')'){
	char *ret=NULL;
	char *temp=NULL;
	/* We may not need the parens anymore */
	if(simplify_algebra(in+1,&temp,chars-2)){
	  strcatA(&ret,"(");
	  strcatA(&ret,temp);
	  __ogg_free(temp);
	  strcatA(&ret,")");
	  *out=ret;
	  return(-1);
	}else{
	  *out=temp;
	  return(0);
	}
      }else
	VOMerror(4);
      break;
    case '-':
      {
	char *temp;
	char *ret=NULL;
	int flag=simplify_algebra(in+1,&temp,chars-1);
	strcatA(&ret,"-");
	strcatA(&ret,temp);
	__ogg_free(temp);
	*out=ret;
	return(flag);
      }
    case '+':
      return(simplify_algebra(in+1,out,chars-1));
    default:
      /* Might this be a variable? */

      switch(in[chars-1]){
      case 'M':case 'P':case 'L':case 'B':
	VOMerror(5);
      case 'V':
	in[chars-1]='\0';
	lookup=registerlist(&O_vars,in);
	in[chars-1]='V';

	*out=NULL;
	aA(out,lookup);
	strcatA(out,"V");
	return(-1);
      default:
	*out=NULL;
	aA(out,atoi(in));
	return(0);
      }
    }
  }else{
    char *temp1,*temp2;
    long sym1=simplify_algebra(in,&temp1,position);
    long sym2=simplify_algebra(in+position+1,&temp2,chars-position-1);

    if(sym1 || sym2){
    
      *out=NULL;
      strcatA(out,temp1);
      
      switch(in[position]){
      case '|':
	strcatA(out,"|");break;
      case '^':
	strcatA(out,"^");break;
      case '&':
	strcatA(out,"&");break;
      case '<':
	strcatA(out,"<");break;
      case '>':
	strcatA(out,">");break;
      case '*':
	strcatA(out,"*");break;
      case '/':
	strcatA(out,"/");break;
      case '+':
	strcatA(out,"+");break;
      case '-':
	strcatA(out,"-");break;
      default:
	VOMerror(4);
      }

      strcatA(out,temp2);
      __ogg_free(temp1);
      __ogg_free(temp2);
      return(-1);
    }else{
      /* we can further simplify... */
      sym1=atol(temp1);
      sym2=atol(temp2);
      __ogg_free(temp1);
      __ogg_free(temp2);

      switch(in[position]){
      case '|':
	sym1|=sym2;break;
      case '^':
	sym1^=sym2;break;
      case '&':
	sym1&=sym2;break;
      case '<':
	sym1<<=sym2;break;
      case '>':
	sym1>>=sym2;break;
      case '*':
	sym1*=sym2;break;
      case '/':
	sym1/=sym2;break;
      case '+':
	sym1+=sym2;break;
      case '-':
	sym1-=sym2;break;
      default:
	VOMerror(4);
      }

      *out=NULL;
      aA(out,sym1);
      return(0);
      
    }
    /* Never reach here */
  }
  return(0);/* to make gcc shut up */
}

/* the proc that really evaluates the math */
/* error checking is stripped out; it was done during simplification */
static long __ogg_evaluate_algebra(char *in,long chars){
  /* parse left to right, breaking at the lowest priority */
  int parenlevel=0;
  int position=-1;
  int priority=9;  /*  *5 /5 +4 -4 <3 >3 &2 ^1 |0 */
  int unaryflag=-1;
  int count=0;
  char op;
  Trim(in);
  for(count=0;count<chars;count++){
    if(in[count]=='(')parenlevel++;
    if(in[count]==')'){
      parenlevel--;
      unaryflag=0;
      continue;
    }
    if(!parenlevel){
      if(!isspace(in[count])){
	if(isalnum(in[count]))
	  unaryflag=0;
	else{
	  switch(in[count]){
	  case '|':
	    if(priority>=0)
	      position=count;
	    priority=0;
	    break;
	  case '^':
	    if(priority>=1)
	      position=count;
	    priority=1;
	    break;
	  case '&':
	    if(priority>=2)
	      position=count;
	    priority=2;
	    break;
	  case '<':case '>':
	    if(priority>=3)
	      position=count;
	    priority=3;
	    break;
	  case '+':case '-':
	    if(priority>=4)
	      position=count;
	    priority=4;
	    break;
	  case '*':case '/':
	    if(priority>=5)
	      position=count;
	    priority=5;
	    break;
	  case '_':
	    unaryflag=0;
	    break;
	  }
	}
      }
    }
  }

  if(position==-1){
    switch(in[0]){
    case '(':
      return(__ogg_evaluate_algebra(in+1,chars-2));
    case '-':
      return(-__ogg_evaluate_algebra(in+1,chars-1));
    case '+':
      return(__ogg_evaluate_algebra(in+1,chars-1));
    default:
      /* Might this be a variable? */

      if(in[chars-1]=='V')
	return(__ogg_readfromV(atoi(in)));
      else
	return(atoi(in));
    }
  }else{
    op=in[position];
    switch(op){
    case '|':
      return(__ogg_evaluate_algebra(in,position)|
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    case '^':
      return(__ogg_evaluate_algebra(in,position)^
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    case '&':
      return(__ogg_evaluate_algebra(in,position)&
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    case '<':
      return(__ogg_evaluate_algebra(in,position)<<
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    case '>':
      return(__ogg_evaluate_algebra(in,position)>>
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    case '*':
      return(__ogg_evaluate_algebra(in,position)*
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    case '/':
      return(__ogg_evaluate_algebra(in,position)/
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    case '+':
      return(__ogg_evaluate_algebra(in,position)+
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    case '-':
      return(__ogg_evaluate_algebra(in,position)-
	     __ogg_evaluate_algebra(in+position+1,chars-position-1));
    }
  }
  return(0);
}

/* register labels, tokenize args and substitute macros */

void __ogg_firstpass(void){
  int lines=listsize(O_lines),i;
  char *cline,*token;
  parsed *pptr;
  line=-1;

  for(i=0;i<lines;i++){
    statement=i+1;
    argument=0;

    Trim(cline=O_lines[i]);

    token=tokenR(&cline);
    if(token==NULL)
      VOMerror(7);

    if(token[strlen(token)-1]==':'){
      /* label */
      if(token[0]=='<')
	VOMerror(4);
      token[strlen(token)-1]='\0';
      token=subDEF(token);

      if(searchlist(O_labels,token)!=-1)
	VOMerror(8);

      registerlist(&O_labels,token);
      expandlistN(&O_locs,numparsed);
      __ogg_free(token);
      token=tokenR(&cline);
      if(token==NULL)
	continue;
    }

    /* At this point we have a statement to store */
    /* Alloc new parse line entry.  OM_ASSEM */

    O_link=(parsed **)__ogg_realloc(O_link,sizeof(parsed *)*(numparsed+2),
				    OM_ASSEM);
    O_link[numparsed]=(pptr=(parsed *)__ogg_calloc(1,sizeof(parsed),OM_ASSEM));
    O_link[++numparsed]=NULL;
    
    /* The token could be holding a module name or an indirection */
    if(token[0]=='<'){
      if(token[strlen(token)-1]!='M')
	VOMerror(5);
      token[strlen(token)-1]='\0';
      memmove(token,token+1,strlen(token)+1);
      token=subDEF(token);
      strcatA(&(pptr->name),"<");
      strcatA(&(pptr->name),token);
      __ogg_free(token);
      __ogg_ripmove(pptr->name,OM_ASSEM);
      /* end of statement */
    }else{
      int required;
      /* module type here now */
      token=subDEF(token);
      if((pptr->regentry=findmodule(token))==NULL)
	VOMerror(9);
      pptr->name=token; /* Don't free this token!! */
      __ogg_ripmove(token,OM_ASSEM);
      required=-1;
      argument=0;
      while((token=tokenR(&cline))!=NULL){
	argument++;
	/* macro substitution first; it could expand to '#' */
	/* Don't forget to strip typing off */
	/* should be safe to use the arithmetic substitutor for all */
	token=arithDEF(token);

	if(*token=='#'){
	  if(required){
	    __ogg_free(token);
	    required=0;
	  }else{
	    expandlist(&(pptr->variable),token,OM_PARSE);
	    __ogg_free(token);
	  }
	  argument--;
	}else{
	  if(required){
	    expandlist(&(pptr->fixed),token,OM_PARSE);
	    __ogg_free(token);
	  }else{
	    expandlist(&(pptr->variable),token,OM_PARSE);
	    __ogg_free(token);
	  }
	}
      }	  
    }
  }
}

/* Support for the second pass */

static int islegalname(char *token){
  while(*token!='\0')
    if(!isalpha(*token) && !(*token)=='_')
      return(0);
    else
      token++;
  return(-1);
}

/* typecheck args, register and replace all symbols */

void __ogg_secondpass(void){
  char *token,*new;
  long i,lookup;

  line=-1;
  for(i=0;O_link[i]!=NULL;i++){
    statement=i+1;
    argument=0;

    if((token=O_link[i]->name)){
    /* The token could be holding a module name or an indirection */
      if(token[0]=='<'){
	lookup=searchlist(O_labels,token+1);
	if(lookup==-1)
	  VOMerror(10);
	new=NULL;
	strcatA(&new,"<");
	aA(&new,O_locs[lookup]);
	__ogg_free(token);
	__ogg_ripmove(new,OM_ASSEM);
	O_link[i]->name=new;
      }else{
	/* module type here now */
	char *argtypes=O_link[i]->regentry->typelist;
	char **arglist=O_link[i]->fixed;
	long fv=-1,arg=0,argc=0,tokenpos;
	__ogg_ripmove(O_link[i]->fixed,OM_ASSEM);
	__ogg_ripmove(O_link[i]->variable,OM_ASSEM);

	for(;arg<(long)strlen(argtypes);){
	  char typecheck=argtypes[arg];
	  argument=++argc;
	  switch(typecheck){
	  case 'm':case 'p':case 'l':case 'a':case 'c':case 'v':case 'b':
	    if(fv<0){
	      if(listsize(O_link[i]->fixed)!=arg)
		VOMerror(11);
	      arglist=O_link[i]->variable;
	    }
	    tokenpos=++fv;
	    break;
	  default:
	    if(fv>=0)
	      VOMerror(11);
	    tokenpos=arg;
	    break;
	  }

	  token=arglist[tokenpos];
	  if(token==NULL){
	    if(fv!=-1)arg++;
	    if(arg<(long)strlen(argtypes))
	      VOMerror(11);
	    break;
	  }

	  if(*token=='#')
	    /* If fv<0, we're in trouble */
	    if(fv<0)
	      VOMerror(11);
	    else{
	      __ogg_ripmove(token,OM_ASSEM);
	      arg++;
	      argc--;
	    }
	  else{
	    char type=token[strlen(token)-1];
	    switch(typecheck){
	    case 'm':case 'M':
	    case 'b':case 'B':
	    case 'v':case 'V':
	    case 'p':case 'P':
	    case 'l':case 'L':
	      /* First off, make certain it's a legal name */
	      if(!islegalname(token))VOMerror(12);

	      token[strlen(token)-1]='\0';
	      if(toupper(typecheck)!=type)VOMerror(5);
	      switch(type){
	      case 'v':case 'V':
		lookup=registerlist(&O_vars,token);
		break;
	      case 'b':case 'B':
		lookup=registerlist(&O_bpipes,token);
		break;
	      case 'p':case 'P':
		lookup=registerlist(&O_pipes,token);
		break;
	      case 'm':case 'M':
		lookup=searchlist(O_labels,token);
		if(lookup==-1)
		  VOMerror(13);
		lookup=O_locs[lookup];
		break;
	      case 'l':case 'L':
		lookup=searchlist(O_labels,token);
		if(lookup==-1)
		  VOMerror(10);
		lookup=O_locs[lookup];
		break;
	      }
	      new=NULL;
	      aA(&new,lookup);
	      __ogg_free(token);
	      __ogg_ripmove(new,OM_ASSEM);
	      arglist[tokenpos]=new;
	      break;
	    case 'a':case 'A':
	      /* the really painful case; take the parsing outside */
	      /* this will also register variables */
	      simplify_algebra(token,&new,strlen(token));
	      __ogg_free(token);
	      __ogg_ripmove(new,OM_ASSEM);
	      arglist[tokenpos]=new;
	      break;
	    /*case 'C':case 'c':*/
	      /* nothing to do :) */
	    }
	    argc++;
	    if(fv<0)arg++;
	  }
	}
	/* Check for excess */
	if(fv<0){
	  if(listsize(O_link[i]->fixed)!=arg){
	    argument++;
	    VOMerror(11);
	  }
	}else{
	  if(listsize(O_link[i]->variable)!=fv){
	    argument++;
	    VOMerror(11);
	  }
	}
      }
    }
  }
}

/* Argument reading interface for the modules; these also handle
   dynamic evaluation of arithmetic expressions */

/* support first */

int __ogg_fixedARG(module *m,int num){
  if(m->argtypes[num]=='A')
    return(__ogg_evaluate_algebra(m->fixed[num],strlen(m->fixed[num])));
  else
    return(atoi(m->fixed[num])); /* Of course, it would be silly to call this
				    on a non-arithmetic arg runtime :) */
}

int __ogg_variableARG(module *m,int num){
  return(__ogg_evaluate_algebra(m->variable[num],strlen(m->variable[num])));
}

int __ogg_variableBOUND(module *m,int num){
  if(m->variable[num]==NULL)return(0);
  if(m->variable[num][0]=='#')return(-1);
  return(0);
}

int __ogg_variableEND(module *m,int num){
  if(m->variable[num]==NULL)return(-1);
  return(0);
}

char *__ogg_CfixedARG(module *m,int num){
  return(m->fixed[num]);
}

char *__ogg_CvariableARG(module *m,int num){
  return(m->variable[num]);
}

void __ogg_Cerror(int num,int arg){
  /* Set the error flags */
  ogg_errno=101;
  ogg_strerr=top100_errors[ogg_errno-100];
  ogg_script_err_number=num;
  ogg_script_err_line=-1;
  ogg_script_err_statement=statement;
  ogg_script_err_argument=arg;
  ogg_script_err_string=creation_errors[num-100];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
}

static void registerbytype(char *token,char type){
  char *pre,*post;
  if(token==NULL)return;
  switch(type){
  case 'V':case 'v':
    __ogg_registerV(atoi(token));
    break;
  case 'B':case 'b':
    __ogg_Bregister(atoi(token));
    break;
  case 'P':case 'p':
    __ogg_registerP(atoi(token));
    break;
  case 'A':case 'a':
    pre=post=token;

    while(*pre!='\0'){
      /* Get past any symbols; pay particular attention to double quotes */
      while(*post!='\0'){
	if(*post=='"')
	  while(*(++post)!='\0')
	    if(*post=='"')break;
	if(*post=='\0' || isalnum(*post)){
	  pre=post;
	  break;
	}
	post++;
      }
      
      /* we're either at the end, or at a number */
      while(*post!='\0'){
	if(!isalnum(*post))break;
	post++;
      }
      
      if(pre!=post)
	/* do we have a variable? */
	if(*(post-1)=='V')__ogg_registerV(atoi(pre));

      pre=post;
    }
  }
}

/* Make the module, create pipes and variables, call the module init */
static void makeline(int statement){
  module *m;
  parsed *p=O_link[statement];
  mod *r=p->regentry;
  char *types;
  int i,j=0;
 
  /* indirection? */
  if(p->name[0]=='<'){
    __ogg_indirectM(atoi(p->name+1));
  }else{
    m=__ogg_createM();
    m->name=p->name;
    __ogg_ripmove(m->name,OM_MACHINE);
    types=(m->argtypes=r->typelist);
    m->fixed=p->fixed;
    if(m->fixed){
      for(i=0;m->fixed[i]!=NULL;i++){
	__ogg_ripmove(m->fixed[i],OM_MACHINE);
	registerbytype(m->fixed[i],types[j++]);
      }
      __ogg_ripmove(m->fixed,OM_MACHINE);
    }

    m->variable=p->variable;
    if(m->variable){
      for(i=0;m->variable[i]!=NULL;i++){
	__ogg_ripmove(m->variable[i],OM_MACHINE);
	if(m->variable[i][0]=='#')
	  j++;
	else
	  registerbytype(m->variable[i],types[j]);
      }
      __ogg_ripmove(m->variable,OM_MACHINE);
    }
    if(r->create)
      (*(r->create))(m);
  }

  /* free the parsed structure */
  __ogg_free(p);
  O_link[statement]=NULL;
}

void __ogg_makemachine(void){
  int creation_module=0;
  statement=1;
  while(O_link[creation_module]!=NULL)
    makeline(creation_module++),statement++;
}

/* instead of making a machine, we pack a header */

static void writearith(char *token){
  long temp;
  while(*token!='\0'){
    switch(*token){
    case '(':case ')':case '+':case '-':case '/':case '*':case '<':
    case '>':case '&':case '|':case '^':
      __ogg_Bout(*token,8);
      token++;
      break;
    default:
      /* Search until we find a delimiterer; if the area ends with a
	 'V', it's a var, else a number */
      temp=strcspn(token,"()+-*/<>&|^ ");
      if(temp==0){
	ogg_errno=5;
	if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
	exit(1);
      }
      if(token[temp-1]=='V'){
	__ogg_Bout('V',8);
	__ogg_Bout(atol(token),16);
      }else{
	__ogg_Bout('n',8);
	__ogg_Bout(atol(token),32);
      }
      token+=temp;
      break;
    }
  }
  __ogg_Bout(0,8);
}

static char *readarith(void){
  long temp;
  char *ret=NULL,buffer[2];

  buffer[1]='\0';
  while((buffer[0]=(char)__ogg_Bread(8))!='\0'){
    switch(buffer[0]){
    case '(':case ')':case '+':case '-':case '/':case '*':case '<':
    case '>':case '&':case '|':case '^':
      strcatA(&ret,buffer);
      break;
    case 'V':
      temp=__ogg_Bread(16);
      aA(&ret,temp);
      strcatA(&ret,"V");
      break;
    case 'n':
      temp=__ogg_Blook(32);
      __ogg_Badv(32);
      aA(&ret,temp);
      break;      
    default:
      ogg_errno=8;
      if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
      exit(1);
    }
  }
  return(ret);
}

void __ogg_packheader(char **ret,long *bytes){ /* returns buffer in no group */
  parsed **highp=O_link;
  parsed *p=*highp;
  char *oldret;

  __ogg_Btempalloc(); /* We can use __ogg_Bout and __ogg_Bwflush with bpipenums of -1 */

  /* OGG header sig */
  __ogg_BoutSl("OggSQUISH\n");

  /* conforming to which spec version */

  __ogg_Bout(MAJOR,16);
  __ogg_Bout(MINOR,16);
  __ogg_Bout(REVISION,16);

  /* Compressing app?? */

  __ogg_BoutS("OggSquish");

  /* The script */

  while(p!=NULL){
    if(p->name[0]=='<'){
      __ogg_Bout('<',8);
      __ogg_Bout(atol(p->name+1),16);
    }else{
      char **fptr=p->fixed;
      char **vptr=p->variable;
      char *token;
      char *args=p->regentry->typelist;
      int i;

      __ogg_BoutS(p->name);

      for(i=0;i<(int)strlen(args);i++){
	switch(args[i]){
	case 'p':case 'v': case 'm': case 'l':case 'c': case 'b':case 'a':

	  token=*(vptr++);
	  while(token!=NULL && *token!='#'){
	    __ogg_Bout(' ',8);
	    switch(args[i]){
	    case 'p':
	    case 'v':
	    case 'm':
	    case 'b':
	    case 'l':
	      __ogg_Bout(atoi(token),16);
	      break;
	    case 'c':
	      __ogg_BoutSl(token);
	      __ogg_Bout(0,8);
	    break;
	    case 'a':
	      /* This one bites a little */
	      writearith(token);
	      break;
	    }
	    token=*(vptr++);
	  }
	  __ogg_Bout('#',8);
	  break;
	case 'P':
	case 'V':
	case 'M':
	case 'L':
	case 'B':
	  token=*(fptr++);
	  __ogg_Bout(atoi(token),16);
	  break;
	case 'C':
	  token=*(fptr++);
	  __ogg_BoutS(token);
	break;
	case 'A':
	  /* This one bites a little */
	  token=*(fptr++);
	  writearith(token);
	  break;
	}
      }
    }
    p=*(++highp);
  }
  __ogg_Bout('%',8);
  
  ogg_bitstream_major=MAJOR;
  ogg_bitstream_minor=MINOR;
  ogg_bitstream_revision=REVISION;
  ogg_bitstream_vendor="OggSquish";
	  
  /* well, *this* is a little clumsy.  Clean it up later */
  __ogg_Bwflush(-1,ret,bytes); /* in group 'OM_TEMP' now */
  __ogg_ripremove(*ret);
  oldret=*ret;
  *ret=(char *)(((long *)(*ret))-2);
  memmove(*ret,oldret,*bytes);
}

static int header_begun=0;

long __ogg_readheader(char *lines,long bytes){
  long last=0,i;
  char *token,*args;
  jmp_buf old_jump;
  int old_jump_set;

  line=-1;
  statement=0;
  argument=0;

  __ogg_Btempalloc();
  __ogg_Bfill(-1,lines,bytes);

  /* We don't want to longjump exit on 'out of data'... */
  old_jump_set=__ogg_jmp_env_set; /* set back to normal before return */
  memcpy(old_jump,__ogg_jmp_env,sizeof(jmp_buf));
  if(setjmp(__ogg_jmp_env)){
    if(ogg_errno==2){ 
      /* just an out of data; handle it. */
      ogg_errno=0;
      /* Don't set the script complete flag; just exit */
      __ogg_Breset(-1);
      return(last);
    }else{
      /* real error! */
      __ogg_jmp_env_set=old_jump_set;
      memcpy(__ogg_jmp_env,old_jump,sizeof(jmp_buf));
      __ogg_Breset(-1);
      if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
      exit(1);
    }
  }
  __ogg_jmp_env_set=-1;

  if(!header_begun){
    /* Check the header fields */
    /* Set header replies (globals) */
    
    if(__ogg_Bread(8)=='O'){
      if(toupper(__ogg_Bread(8))=='G')
	if(toupper(__ogg_Bread(8))=='G')
	  if(__ogg_Bread(8)=='S')
	    if(__ogg_Bread(8)=='Q')
	      if(__ogg_Bread(8)=='U')
		if(__ogg_Bread(8)=='I')
		  if(__ogg_Bread(8)=='S')
		    if(__ogg_Bread(8)=='H')
		      if(__ogg_Bread(8)=='\n'){
			int major,minor;
			major=__ogg_Bread(16);
			minor=__ogg_Bread(16);
			ogg_bitstream_major=major;
			ogg_bitstream_minor=minor;
			ogg_bitstream_revision='x';
			if(major==0 && minor<98){
			  ogg_errno=6; /* Incompatable Ogg bitstream version */
			  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
			  exit(1);
			} 

			/* Set header replies (globals) */
			
			ogg_bitstream_revision=__ogg_Bread(16);
			ogg_bitstream_vendor=__ogg_BinSS();
			__ogg_ripmove(ogg_bitstream_vendor,OM_ASSEM);
			header_begun=-1;
			last=__ogg_Btest(-1);
		      }
    
    }else{
      if(__ogg_Bread(8)=='Q')
	if(__ogg_Bread(8)=='U')
	  if(__ogg_Bread(8)=='I')
	    if(__ogg_Bread(8)=='S')
	      if(__ogg_Bread(8)=='H')
		if(__ogg_Bread(8)=='\n'){
		  int major,minor;
		  major=__ogg_Bread(8);
		  minor=__ogg_Bread(8);
		  ogg_bitstream_major=major;
		  ogg_bitstream_minor=minor;
		  ogg_bitstream_revision='x';
		  ogg_errno=6; /* Incompatable Ogg bitstream version */
		  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
		  exit(1);
		} 
    }
		
    if(!header_begun){
      /* Gee, must have not have had an 'OggSQUISH' sig */
      ogg_errno=7; /* Not an ogg bitstream */
      __ogg_Breset(-1);
      if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
      exit(1);
    }
  }

  /* Grab statements one at a time; commit to mem and 'last' after
     getting each whole statement */

  while(1){
    char chartemp,*strtemp,*build;
    parsed *pptr;
    /* If we run out of data, Bread will longjump.  If not, we exit on
       the '%'. */

    /* indirection? */
    chartemp=(char)__ogg_Bread(8);
    if(chartemp=='%'){
      /* Yay!  Complete header read! */
      last=__ogg_Btest(-1);
      break;
    }
    /* alloc a parsed struct, but don't insert it in the parsed list
       until we complete the statement read */
    /* alloc from OM_ASSEM; 'leaked' memory will be freed once we
       reset OM_ASSEM*/

    pptr=(parsed *)__ogg_calloc(1,sizeof(parsed),OM_ASSEM);
    statement=numparsed+1;
    argument=0;

    if(chartemp=='<'){
      int tempi;
      build=NULL;
      strcatA(&build,"<");
      tempi=__ogg_Bread(16);
      aA(&build,tempi);

      __ogg_ripmove(build,OM_ASSEM);
      pptr->name=build;
    }else{
      int required=-1; 

      /* Grab the rest of the module name */
      /* Splice the mess together */
      build=NULL;
      strcatA(&build," ");
      build[0]=chartemp;
      strtemp=__ogg_BinSS();
      strcatA(&build,strtemp);
      __ogg_free(strtemp);

      __ogg_ripmove(build,OM_ASSEM);
      pptr->name=build;

      pptr->regentry=findmodule(build);
      /* Yeah, VOMerror jumps back within the function, but it will
	 be forwarded to the proper place */
      if(pptr->regentry==NULL)VOMerror(9);
      args=pptr->regentry->typelist;

      required=-1;
      for(i=0;i<(long)strlen(args);i++){
	char *strtemp,ctoken;
	long temp;

	switch(args[i]){
	case 'p':case 'b': case 'm': case 'l':case 'c': case 'v':case 'a':
	  if(required==-1)
	    required=0;
	  else
	    expandlist(&(pptr->variable),"#",OM_ASSEM);
	  
	  while((ctoken=(char)__ogg_Bread(8))==' '){
	    switch(args[i]){
	    case 'p':
	    case 'v':
	    case 'm':
	    case 'b':
	    case 'l':
	      temp=__ogg_Bread(16);
	      strtemp=NULL;
	      aA(&strtemp,temp);
	      expandlist(&(pptr->variable),strtemp,OM_ASSEM);
	      __ogg_free(strtemp);
	      break;
	    case 'c':
	      strtemp=__ogg_BinSS();
	      expandlist(&(pptr->variable),strtemp,OM_ASSEM);
	      break;
	    case 'a':
	      /* This one bites a little */
	      strtemp=readarith();
	      expandlist(&(pptr->variable),strtemp,OM_ASSEM);
	      break;
	    }
	  }
	  /* Is the token a proper terminator? */
	  if(ctoken!='#'){
	    ogg_errno=8;
	    if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
	    exit(1);
	  }
	  break;
	case 'P':
	case 'V':
	case 'M':
	case 'L':
	case 'B':
	  if(required!=-1){
	    ogg_errno=8;
	    if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
	    exit(1);
	  }
	  temp=__ogg_Bread(16);
	  strtemp=NULL;
	  aA(&strtemp,temp);
	  expandlist(&(pptr->fixed),strtemp,OM_ASSEM);
	  __ogg_free(strtemp);
	  break;
	case 'C':
	  if(required!=-1){
	    ogg_errno=8;
	    if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
	    exit(1);
	  }
	  token=__ogg_BinSS();
	  expandlist(&(pptr->fixed),token,OM_ASSEM);
	break;
	case 'A':
	  if(required!=-1){
	    ogg_errno=8;
	    if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
	    exit(1);
	  }
	  token=readarith();
	  expandlist(&(pptr->fixed),token,OM_ASSEM);
	  break;
	}
      }
    }
    
    /* OK; got the whole thing */
    O_link=(parsed **)__ogg_realloc(O_link,sizeof(parsed *)*(numparsed+2),
				    OM_ASSEM);
    O_link[numparsed]=pptr;
    O_link[++numparsed]=NULL;
    last=__ogg_Btest(-1);
  }

  /* Got everything.  Set the completion flag */
  __ogg_script_complete=-1;
  
  /* Reset the longjump */
  __ogg_jmp_env_set=old_jump_set;
  memcpy(__ogg_jmp_env,old_jump,sizeof(jmp_buf));
  __ogg_Breset(-1);
  return(last);
}

void __ogg_CleanupParser(void){
  /* free the macro lists, parse structures, and symbol lists */

  __ogg_ripfree(OM_PARSE);
  O_partial=NULL;
  O_lines=NULL;
  
  O_vars=NULL;
  O_pipes=NULL;
  O_bpipes=NULL;
  O_labels=NULL;
  O_locs=NULL;

  Vd=NULL; 
  Ve=NULL;
  
  line=-1;
  statement=-1;
  argument=-1;
}

void __ogg_CleanupAssembler(void){
  /* free OM_ASSEM, reset parsed list */
  header_begun=0;
  __ogg_ripfree(OM_ASSEM);

  ogg_bitstream_vendor=NULL;

  O_link=NULL;
  numparsed=0;
  line=-1;
  statement=-1;
  argument=-1;
}

