/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: quantize.c
  function: vector msb quantization routines
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Jun 11 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "ilog.h"
#include "quantize.h"

/* Quantize a vector to the msb and squash the resulting value */
void __ogg_msb_quantize(long *v,int n){
  int count;

  for(count=0;count<n;count++){
    if(v[count]>0)
      v[count]=ilog(v[count]);
    else
      if(v[count]<0)
	v[count]=-(ilog(-v[count]));
  }
}

static unsigned long maskq[]=
                    {0x00000000,0x00000001,0x00000003,0x00000007,0x0000000f,
		     0x0000001f,0x0000003f,0x0000007f,0x000000ff,0x000001ff,
		     0x000003ff,0x000007ff,0x00000fff,0x00001fff,0x00003fff,
		     0x00007fff,0x0000ffff,0x0001ffff};

void __ogg_msb_quantizeN(long *v,int n,int bits){
  int count,temp,temp2;

  for(count=0;count<n;count++){
    temp=0;
    if(v[count]>0){
      temp=ilog(v[count]);
      temp2=v[count];
    }else
      if(v[count]<0){
	temp=ilog(-v[count]);
	temp2=-v[count];
      }

    if(temp>bits){
      temp2=(temp2>>(temp-bits))&maskq[bits-1];
      temp=(temp-bits+2)<<(bits-1);

      temp|=temp2;
      if(v[count]<0)temp=-temp;
      v[count]=temp;

    }/* Otherwise leave it alone */
  }
}

/* reverse msb_quantize */
void __ogg_msb_unquantize(long *v,int n){
  int i;
  for(i=0;i<n;i++){
    if(v[i]<0)
      v[i]=-(1<<(-v[i]-1));
    else
      if(v[i])
	v[i]=1<<(v[i]-1);
  }
}

void __ogg_msb_unquantizeN(long *v,int n,int bits){
  int count,temp,temp2,thresh,pre;

  thresh=(1<<bits)-1;
  pre=1<<(bits-1);
  for(count=0;count<n;count++){
    temp=0;
    if(v[count]>0)
      temp=v[count];
    else
      if(v[count]<0)
	temp=-v[count];

    if(temp>thresh){
      temp2=pre|(temp&maskq[bits-1]);
      temp2<<=(temp>>(bits-1))-2;

      if(v[count]<0)temp2=-temp2;
      v[count]=temp2;
    }/* else leave it alone */
  }
}
