/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: registry.c
  function: Virtual machine module registration struct
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 09 1996

 ********************************************************************/

#include "A.h"
#include "B.h"
#include "C.h"
#include "D.h"
#include "E.h"
#include "F.h"
#include "J.h"
#include "S.h"
#include "W.h"

typedef struct mod {
  char *name;
  char *typelist;
  void (*create)(module *);
} mod;

/* B bitpipe
   P pipe
   V scalar variable
   M module
   A arith expression
   C character string
   L location (execution order)
   p,v,a,b,m,c,l same in context of a vararg list
*/

static mod registry[]={ 
  {"Aid","AAPAAAA",__oggc_Aid},
  {"Aib","AABAA",__oggc_Aib},
  {"Aod","AAPAAA",__oggc_Aod},
  {"Aob","AABA",__oggc_Aob},

  {"B","",NULL}, 
  {"Bh","",NULL}, 
  {"Bdpl","AMAAA",__oggc_Bdpl}, 
  {"Bdpc","AAMAAA",__oggc_Bdpc}, 
  {"Bxc","M",__oggc_Bxc}, 
  {"Bxu","M",__oggc_Bxu}, 

  {"Covr","PAMB",__oggc_Covr}, 
  {"Civrs","PAAAMB",__oggc_Civrs}, 
  {"Conr","PB",__oggc_Conr}, 
  {"Cinr","PB",__oggc_Cinr}, 
  {"Conrl","PB",__oggc_Conrl}, 
  {"Cinrl","PB",__oggc_Cinrl}, 

#ifdef PRE_COMPAT
  {"Civta","PBAAAma",__oggc_Civta}, 
  {"Civt","PBAAAma",__oggc_Civt}, 
#endif

  {"Covtl","PBAm",__oggc_Covtl}, 
  {"Civtl","PBAAAm",__oggc_Civtl}, 
  {"Conc","PBAAMA",__oggc_Conc}, 
  {"Cinc","PBAAMA",__oggc_Cinc}, 
  {"Coa","AAB",__oggc_Coa}, 
  {"Cia","VAB",__oggc_Cia}, 

  {"Dca","PPVA",__oggc_Dca}, 
  {"Dr","PPPPAAP",__oggc_Dr}, 
  {"Dp","PPPPP",__oggc_Dp}, 
  {"Dr32","PPPP",__oggc_Dr32}, 
  {"Dp32","PPPP",__oggc_Dp32}, 
  {"Dr64","PPPP",__oggc_Dr64}, 
  {"Dp64","PPPP",__oggc_Dp64}, 
  {"Ddf","PPP",__oggc_Ddf}, 
  {"Ddi","PPP",__oggc_Ddi}, 

  {"Es","PPPA",__oggc_Es}, 
  {"Ez","PAAA",__oggc_Ez}, 
  {"Ezo","PAAA",__oggc_Ezo}, 
  {"Ec","PPP",__oggc_Ec}, 
  {"Eq","PA",__oggc_Eq}, 
  {"Eqi","PA",__oggc_Eqi}, 
  {"El","PP",__oggc_El}, 
  {"Edi","PPP",__oggc_Edi}, 
  {"Eaa","PP",__oggc_Eaa}, 
  {"Eas","PP",__oggc_Eas}, 
  {"Ean","P",__oggc_Ean}, 
  {"Eav","PAV",__oggc_Eav}, 
  {"Ecs","PAP",__oggc_Ecs}, 
  {"Ecr","PAP",__oggc_Ecr}, 

  {"Ff","PP",__oggc_Ff}, 
  {"Ffi","PP",__oggc_Ffi}, 
  {"Fcq","PP",__oggc_Fcq}, 
  {"Fcqi","PP",__oggc_Fcqi}, 
  {"Ff32","PP",__oggc_Ff32}, 
  {"Ffi32","PP",__oggc_Ffi32}, 
  {"Fcq32","PP",__oggc_Fcq32}, 
  {"Fcqi32","PP",__oggc_Fcqi32}, 
  {"Ffs","PPAA",__oggc_Ffs}, 
  {"Ffis","PPAA",__oggc_Ffis}, 
  {"Fcqs","PPAA",__oggc_Fcqs}, 
  {"Fcqis","PPAA",__oggc_Fcqis}, 
  {"Fm","PPPAA",__oggc_Fm},
  {"Fmc","PPPAA",__oggc_Fmc},
  {"Fmi","PPPA",__oggc_Fmi},
  {"Fn","P",__oggc_Fn}, 

  {"J","L",__oggc_J},
  {"Jle","AAL",__oggc_Jle},
  {"Jl","AAL",__oggc_Jl},
  {"Jge","AAL",__oggc_Jge},
  {"Jg","AAL",__oggc_Jg},
  {"Jne","AAL",__oggc_Jne},
  {"Je","AAL",__oggc_Je},
  {"Jz","AL",__oggc_Jz},
  {"Jnz","AL",__oggc_Jnz},
  {"Jwle","AALB",__oggc_Jle},
  {"Jwl","AALB",__oggc_Jl},
  {"Jwge","AALB",__oggc_Jge},
  {"Jwg","AALB",__oggc_Jg},
  {"Jwne","AALB",__oggc_Jne},
  {"Jwe","AALB",__oggc_Je},
  {"Jwz","ALB",__oggc_Jz},
  {"Jwnz","ALB",__oggc_Jnz},
  {"Jr","LB",__oggc_Jr},
  {"Jfor","VAAL",__oggc_Jfor},

  {"Sof","PPBAV",__oggc_Sof}, 
  {"Sif","BAV",__oggc_Sif}, 
  {"Sofr","PPBAAAAAVV",__oggc_Sofr}, 
  {"Sifr","BAAVV",__oggc_Sifr}, 
  {"Shalt","AA",__oggc_Shalt},
  {"Spm","PP",__oggc_Spm},
  {"Sbm","BB",__oggc_Sbm},
  {"Spc","PP",__oggc_Spc},
  {"Sbc","BB",__oggc_Sbc},
  {"Sba","BB",__oggc_Sba},
  {"Sbt","BV",__oggc_Sbt},
  {"Sbr","B",__oggc_Sbr},
  {"Sv","AV",__oggc_Sv},
  {"Spush","AP",__oggc_Spush},
  {"Spull","PV",__oggc_Spull},

  {"Whl","PA",__oggc_Whl},
  {"Whr","PA",__oggc_Whr},
  {"Whli","PA",__oggc_Whli},
  {"Whri","PA",__oggc_Whri},

  {NULL,NULL,NULL}};
