/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS RESTRICTED  *
 * BY DISTRIBUTION TERMS THAT SHOULD BE INCLUDED WITH THIS SOURCE.  *
 * PLEASE READ THESE TERMS BEFORE USING OR DISTRIBUTING.            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1996             *
 * XIPHOPHORUS                                                      *
 *                                                                  *
 ********************************************************************

  file: virtual.c
  function: Virtual machine loop
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Aug 06 1996

 ********************************************************************/

#include "ogg.h"
#include "memory.h"
#include "buffer.h"
#include "bitwise.h"
#include "huffman.h"
#include "parser.h"
#include "strerrs.h"
#include "memtypes.h"
#include "virtual.h"

#ifndef SELFTESTS
extern int ogg_errno;
extern char *ogg_strerr;
extern jmp_buf __ogg_jmp_env;
extern int __ogg_jmp_env_set;
extern int __ogg_precision;
extern int ogg_run_loaded;
#else
int ogg_run_loaded;
int __ogg_memory=0;
int ogg_errno=0;
char *ogg_strerr=NULL;
jmp_buf __ogg_jmp_env;
int __ogg_jmp_env_set=0;
int __ogg_precision=0;
#include <stdio.h>
#endif

/* Also handle bank switching for multiscripting here */

/* These are set according to bank */
int ogg_run_ready=0; /* nonzero also signifies state */
int ogg_run_scriptnum=0;
char *ogg_comment=NULL;

int OM_PIPE=OM_BASE+POM_PIPE;
int OM_VAR=OM_BASE+POM_VAR;
int OM_BPIPE=OM_BASE+POM_BPIPE;
int OM_HUFF=OM_BASE+POM_HUFF;
int OM_MACHINE=OM_BASE+POM_MACHINE;

extern mpipe **__ogg_pipelist;
extern long __ogg_pipes;
extern mvar **__ogg_varlist;
extern long __ogg_mvars;

extern bpipe **__ogg_bpipelist;
extern int __ogg_bpipes;

static module **modulelist=NULL;
static int modulenum=0;
static int MC;

typedef struct bankstate{
  int run_ready;
  module **modulelist;
  int modulenum;

  /* Named buffers are part of the state */
  dspb **dspin;
  dspb **dspout;
  bitb **bitin;
  bitb **bitout;
  
  long dspinN;
  long dspoutN;
  long bitinN;
  long bitoutN;

  mpipe **pipelist;
  long pipes;
  mvar **varlist;
  long mvars;
  bpipe **bpipelist;
  int bpipes;

} bankstate;

static struct bankstate **banklist=NULL;
static int banks=0;

/* I/O buffer stuff */

dspb **__ogg_dspin=NULL;
dspb **__ogg_dspout=NULL;
bitb **__ogg_bitin=NULL;
bitb **__ogg_bitout=NULL;

long __ogg_dspinN=0;
long __ogg_dspoutN=0;
long __ogg_bitinN=0;
long __ogg_bitoutN=0;

/* Cleans up only the active bank */
void __ogg_cleanupM(void){
  __ogg_ripfree(OM_MACHINE);
  modulelist=NULL;
  modulenum=0;
  ogg_run_ready=0;
  ogg_run_loaded=0;
  __ogg_dspin=NULL;
  __ogg_dspout=NULL;
  __ogg_bitin=NULL;
  __ogg_bitout=NULL;
  __ogg_dspinN=0;
  __ogg_dspoutN=0;
  __ogg_bitinN=0;
  __ogg_bitoutN=0;
  ogg_comment=NULL;
}

/* Bank switcher; only save banks with state */
/* Very ugly, but best way to handle multiscripts with the current
   code setup */

/* This will leave the bank state clean if we abort on a mem error */
void __ogg_bankswitch(int to){
  /* save current if state */
  if(ogg_run_loaded){
    /* Have state... is a previous state allocated? */
    if(ogg_run_scriptnum>=banks){
      /* have to expand the bank list */
      int i;
      banklist=(bankstate **)__ogg_realloc(banklist,sizeof(bankstate *)*
					   (ogg_run_scriptnum+1),OM_BANK);
      for(i=banks;i<=ogg_run_scriptnum;i++)banklist[i]=NULL;
      banks=ogg_run_scriptnum+1;
    }
    if(banklist[ogg_run_scriptnum]==NULL){
      /* alloc our state */
      banklist[ogg_run_scriptnum]=
	(bankstate *)__ogg_calloc(1,sizeof(bankstate),OM_BANK);
    }

    banklist[ogg_run_scriptnum]->run_ready=ogg_run_ready;
    banklist[ogg_run_scriptnum]->modulelist=modulelist;
    banklist[ogg_run_scriptnum]->modulenum=modulenum;

    banklist[ogg_run_scriptnum]->dspin=__ogg_dspin;
    banklist[ogg_run_scriptnum]->dspout=__ogg_dspout;
    banklist[ogg_run_scriptnum]->bitin=__ogg_bitin;
    banklist[ogg_run_scriptnum]->bitout=__ogg_bitout;
    banklist[ogg_run_scriptnum]->dspinN=__ogg_dspinN;
    banklist[ogg_run_scriptnum]->dspoutN=__ogg_dspoutN;
    banklist[ogg_run_scriptnum]->bitinN=__ogg_bitinN;
    banklist[ogg_run_scriptnum]->bitoutN=__ogg_bitoutN;
    
    banklist[ogg_run_scriptnum]->pipelist=__ogg_pipelist;
    banklist[ogg_run_scriptnum]->bpipelist=__ogg_bpipelist;
    banklist[ogg_run_scriptnum]->varlist=__ogg_varlist;
    banklist[ogg_run_scriptnum]->pipes=__ogg_pipes;
    banklist[ogg_run_scriptnum]->bpipes=__ogg_bpipes;
    banklist[ogg_run_scriptnum]->mvars=__ogg_mvars;

  }else{
    /* no current state; dealloc any old state */
    __ogg_cleanupP();
    __ogg_cleanupV();
    __ogg_cleanupB();
    __ogg_cleanupH();
    __ogg_cleanupM();
    if(banklist)
      if(ogg_run_scriptnum<banks){
      __ogg_free(banklist[ogg_run_scriptnum]);
      banklist[ogg_run_scriptnum]=NULL;
    }
  }

  /* load new state */
  if(to<banks && banklist[to]!=NULL){
    /* We have a state to load */
    ogg_run_ready=banklist[to]->run_ready;
    ogg_run_loaded=-1;
    modulelist=banklist[to]->modulelist;
    modulenum=banklist[to]->modulenum;

    __ogg_dspin=banklist[to]->dspin;
    __ogg_dspout=banklist[to]->dspout;
    __ogg_bitin=banklist[to]->bitin;
    __ogg_bitout=banklist[to]->bitout;
    __ogg_dspinN=banklist[to]->dspinN;
    __ogg_dspoutN=banklist[to]->dspoutN;
    __ogg_bitinN=banklist[to]->bitinN;
    __ogg_bitoutN=banklist[to]->bitoutN;

    __ogg_pipelist=banklist[to]->pipelist;
    __ogg_bpipelist=banklist[to]->bpipelist;
    __ogg_varlist=banklist[to]->varlist;
    __ogg_pipes=banklist[to]->pipes;
    __ogg_bpipes=banklist[to]->bpipes;
    __ogg_mvars=banklist[to]->mvars;

  }else{
    /* The requested bank is currently unused;
       initialize vars */
    ogg_run_ready=0;
    ogg_run_loaded=0;
    modulelist=NULL;
    modulenum=0;
    __ogg_dspin=NULL;
    __ogg_dspout=NULL;
    __ogg_bitin=NULL;
    __ogg_bitout=NULL;
    __ogg_dspinN=0;
    __ogg_dspoutN=0;
    __ogg_bitinN=0;
    __ogg_bitoutN=0;

    __ogg_pipelist=NULL;
    __ogg_bpipelist=NULL;
    __ogg_varlist=NULL;
    __ogg_pipes=0;
    __ogg_bpipes=0;
    __ogg_mvars=0;
  }
  ogg_run_scriptnum=to;
  OM_PIPE=OM_BASE+POM_PIPE+OM_OFF*to;
  OM_VAR=OM_BASE+POM_VAR+OM_OFF*to;
  OM_BPIPE=OM_BASE+POM_BPIPE+OM_OFF*to;
  OM_HUFF=OM_BASE+POM_HUFF+OM_OFF*to;
  OM_MACHINE=OM_BASE+POM_MACHINE+OM_OFF*to;
}

module *__ogg_createM(void){
  module *m;

  m=(module *)__ogg_calloc(1,sizeof(module),OM_MACHINE);
  modulelist=(module **)__ogg_realloc(modulelist,sizeof(module *)*
					(modulenum+1),OM_MACHINE);
  m->handle=modulenum;
  modulelist[modulenum++]=m;
  return(m);
}

module *__ogg_indirectM(int mod){
  module *m;
  
  if(mod<0||mod>=modulenum)return(NULL);
  m=modulelist[mod];

  modulelist=(module **)__ogg_realloc(modulelist,sizeof(module *)*
					(modulenum+1),OM_MACHINE);
  modulelist[modulenum++]=m;
  return(m);
}

/* We can't pass pointers through sockets, but we can pass handles */
/* We are assuming the handle is valid */
void __ogg_callMbyhandle(int h){
  (*(modulelist[h]->action))(modulelist[h]);
}

int __ogg_getMhandle(module *m){
  return(m->handle);
}

module *__ogg_gethandleM(int handle){
  return(modulelist[handle]);
}

static void clear_ouput_buffers(void){
  int i;

  for(i=0;i<__ogg_dspoutN;i++)
    if(__ogg_dspout[i])
      if(__ogg_dspout[i]->ptr){
	__ogg_freeMB(__ogg_dspout[i]->ptr);
	__ogg_dspout[i]->ptr=NULL;
      }
  for(i=0;i<__ogg_bitoutN;i++)
    if(__ogg_bitout[i])
      if(__ogg_bitout[i]->ptr){
	__ogg_free(__ogg_bitout[i]->ptr);
	__ogg_bitout[i]->ptr=NULL;
      }
}

static void postprocess_input_buffers(void){
  int i;

  for(i=0;i<__ogg_dspinN;i++)
    if(__ogg_dspin[i]){
      __ogg_dspin[i]->ptr=__ogg_readfromP(__ogg_dspin[i]->pipe);
      __ogg_ripmoveMB(__ogg_dspin[i]->ptr,OM_MACHINE);
    }
  for(i=0;i<__ogg_bitinN;i++)
    if(__ogg_bitin[i]){
      __ogg_bitin[i]->bytes=__ogg_Brremaining(__ogg_bitin[i]->bpipe,&(__ogg_bitin[i]->ptr));
      __ogg_ripmove(__ogg_bitin[i]->ptr,OM_MACHINE);
    }
}

void __ogg_virtual_machine_pass(void){
  clear_ouput_buffers();
  for(MC=0;MC<modulenum && ogg_run_ready;MC++)
    if(modulelist[MC])
      if(modulelist[MC]->action)
	(*(modulelist[MC]->action))(modulelist[MC]);
  postprocess_input_buffers();
}

/* Well, this is systems programming, so we're allowed to do really
   ugly things */

void __ogg_virtual_machine_jump(int pos){
  MC=pos-1;
}

void __ogg_virtual_machine_halt(void){
  ogg_run_ready=0;
}

/* I/O buffer stuff ***************************/

int __ogg_buffer_used(int group,int name){
  int maxname;
  void **ptrs;

  if(group<0 || group>MAX_IOBUFFER)return(0);
  switch(group){
  case 0:
    maxname=__ogg_dspinN;
    ptrs=(void **)__ogg_dspin;
    break;
  case 2:
    maxname=__ogg_bitinN;
    ptrs=(void **)__ogg_bitin;
    break;
  case 1:
    maxname=__ogg_dspoutN;
    ptrs=(void **)__ogg_dspout;
    break;
  case 3:
    maxname=__ogg_bitoutN;
    ptrs=(void **)__ogg_bitout;
    break;
  }
  if(ptrs==NULL)return(0);
  if(name<0 || name>=maxname)return(0);
  if(ptrs[name]==NULL)return(0);
  return(-1);
}

dspb *__ogg_registerDSP(dspb ***ptr,long *nptr,int name){
#ifndef SELFTESTS
  if(name<0)__ogg_Cerror(100,2);
#endif
  if(*nptr<=name){
    int i;
    *ptr=(dspb **)__ogg_realloc(*ptr,sizeof(dspb *)*(name+1),OM_MACHINE);
    for(i=*nptr;i<=name;i++)(*ptr)[i]=NULL;
    *nptr=name+1;
  }
  if((*ptr)[name]==NULL)
    (*ptr)[name]=(dspb *)__ogg_calloc(1,sizeof(dspb),OM_MACHINE);
#ifndef SELFTESTS
  else
    __ogg_Cerror(101,2);
#endif
  return((*ptr)[name]);
}

bitb *__ogg_registerBIT(bitb ***ptr,long *nptr,int name){
#ifndef SELFTESTS
  if(name<0)__ogg_Cerror(100,2);
#endif
  if(*nptr<=name){
    int i;
    *ptr=(bitb **)__ogg_realloc(*ptr,sizeof(bitb *)*(name+1),OM_MACHINE);
    for(i=*nptr;i<=name;i++)(*ptr)[i]=NULL;
    *nptr=name+1;
  }
  if((*ptr)[name]==NULL)
    (*ptr)[name]=(bitb *)__ogg_calloc(1,sizeof(bitb),OM_MACHINE);
#ifndef SELFTESTS
  else
    __ogg_Cerror(101,2);
#endif
  return((*ptr)[name]);
}

static void Verror(int num){
  /* Set the error flags */
  ogg_errno=num;
  ogg_strerr=top_errors[ogg_errno];
  if(__ogg_jmp_env_set)longjmp(__ogg_jmp_env,1);
  exit(1);
}

void __ogg_buffer_dsp(int group,int name,struct ogg_dspinfo *s){
  dspb **ptr,*ptr2;
  switch(group){
  case 0: case 1:
    if(!__ogg_buffer_used(group,name))
      Verror(15);
    switch(group){
    case 0:
      ptr=__ogg_dspin;
      break;
    case 1:
      ptr=__ogg_dspout;
      break;
    }
    break;
  default:
    Verror(16);
  }
  ptr2=ptr[name];
  s->rate=ptr2->rate;
  s->channels=ptr2->channels;
  s->dynrange=ptr2->dynrange;
}

void __ogg_buffer_bit(int group,int name,struct ogg_bitinfo *s){
  bitb **ptr,*ptr2;
  switch(group){
  case 2: case 3:
    if(!__ogg_buffer_used(group,name))
      Verror(15);
    switch(group){
    case 2:
      ptr=__ogg_bitin;
    case 3:
      ptr=__ogg_bitout;
    break;
    }
    break;
  default:
    Verror(16);
  }
  ptr2=ptr[name];
  s->seekable=ptr2->seekable;
}

long __ogg_input_query(int group,int name){
  /* Postprocessing has restored buffered input to the temp buffer */
  switch(group){
  case 0:case 2:
    if(!__ogg_buffer_used(group,name))
      Verror(15);
    switch(group){
    case 0:
      {
	dspb *ptr=__ogg_dspin[name];
	long elements=(ptr->hint-__ogg_sizeMB(ptr->ptr))*ptr->channels;
	if(elements<0)return(0);
	return(elements);
      }
    case 2:
      {
	bitb *ptr=__ogg_bitin[name];
	
	long elements=(ptr->hint-ptr->bytes);
	if(elements<0)return(0);
	return(elements);
      }
    }
    break;
  default:
    Verror(16);
  }
  return(-1);/* make the compiler stop complaining */
}

void __ogg_input_reset(int group,int name){
  /* Postprocessing has restored buffered input to the temp buffer */
  switch(group){
  case 0:case 2:
    if(!__ogg_buffer_used(group,name))return;
    switch(group){
    case 0:
      __ogg_freeMB(__ogg_dspin[name]->ptr);
      __ogg_dspin[name]->ptr=NULL;
      break;
    case 2:
      __ogg_free(__ogg_bitin[name]->ptr);
      __ogg_bitin[name]->ptr=NULL;
      __ogg_bitin[name]->bytes=0;
      break;
    }
    break;
  default:
    Verror(16);
  }
}


/* I wonder how many alignment issues this is gonna trip in the
   future...  it shouldn't trip any...  I guess we'll find out :) */

static int bigendianp(void){
  int test=1;
  char *hack=(char *)(&test);
  if(hack[0])return(0);
  return(1);
}

/* Padding out fractional bytes is done AIFC style */

void __ogg_input_dsp(int group,int name,void *data,long elements,
		     int width,int endian,int sign,int scale){
  if(width<2)Verror(18);
  switch(endian){
  case -1:
    endian=bigendianp();
    break;
  case 0:case 1:
    break;
  default:
    endian=!bigendianp();
    break;
  }
  switch(group){
  case 0:
    if(!__ogg_buffer_used(group,name))
      Verror(15);
    else{
      dspb *ptr=__ogg_dspin[name];
      int ch;
      void *m;
      elements=(elements+ptr->channels-1)/ptr->channels;
      m=__ogg_allocMB(ptr->channels,elements);

      /* Move the data into the mbuffer */
      for(ch=0;ch<ptr->channels;ch++){
	long *mptr=__ogg_ptrMB(m,ch);
	int i,j;
	/* We want to be doing math on longs, not chars... get the data into
	   the destination array, then swap/sign any bytes */
	/* Assume that if we're handling an odd bit size (like 20)
           that it will be in an integer number of bytes (20 in 24
           bits or 3 bytes) */
	
	switch((width-1)/8){
	case 0:
	  /* 1 byte wide */
	  {
	    signed char *tdata=(signed char *)data;
	    j=ch;
	    for(i=0;i<elements;i++){
	      mptr[i]=tdata[j];
	      j+=ptr->channels;
	    }
	  }
	  break;
	case 1:
	  /* 2 bytes wide */
	  {
	    size16 *tdata=(size16 *)data;
	    j=ch;
	    for(i=0;i<elements;i++){
	      mptr[i]=tdata[j];
	      j+=ptr->channels;
	    }
	    if(bigendianp()!=endian)
	      for(i=0;i<elements;i++)mptr[i]=(short)(((mptr[i]&0xff)<<8)|
						     ((mptr[i]>>8)&0xff));
	  }	  
	  break;
	case 2:
	  /* 3 bytes wide */
	  {
	    signed char *tdata=(signed char *)data;
	    unsigned char *udata=(unsigned char *)data;
	    j=ch*3;
	    if(endian)
	      for(i=0;i<elements;(i++),(j+=(3*ptr->channels)))
		mptr[i]=(((long)(tdata[j]))<<16)|
		  (((long)(udata[j+1]))<<8)|
		  udata[j+2];
	    else
	      for(i=0;i<elements;(i++),(j+=(3*ptr->channels)))
		mptr[i]=(((long)(tdata[j+2]))<<16)|
		  (((long)(udata[j+1]))<<8)|
		  udata[j];
		
	  }	  
	  break;
	case 3:
	  /* 4 bytes wide */
	  {
	    size32 *tdata=(size32 *)data;
	    j=ch;
	    for(i=0;i<elements;i++){
	      mptr[i]=tdata[j];
	      j+=ptr->channels;
	    }
	    if(bigendianp()!=endian)
	      for(i=0;i<elements;i++)mptr[i]=(size32)(((mptr[i]&0xff)<<24)|
						      ((mptr[i]&0xff00)<<8)|
						      ((mptr[i]>>8)&0xff00)|
						      ((mptr[i]>>24)&0xff));
	  }
	  break;
	default:
	  Verror(18);
	}

	/* Are we signed or unsigned? */
	if(sign){
	  if(width%8){
	    /* AIFC style; we're left padded right now */
	    int move1=32-(((width-1)/8+1)*8);
	    int move2=32-width;
	    for(i=0;i<elements;i++)mptr[i]=(mptr[i]<<move1)>>move2;
	  }
	}else{
	  unsigned long *uptr=(unsigned long *)mptr;
	  if(width%8){
	    /* AIFC style; we're left padded right now */
	    int move1=32-(((width-1)/8+1)*8)-width;
	    for(i=0;i<elements;i++)mptr[i]>>=move1;
	  }
	  
	  /* Sign the data */
	  /* also need to strip off the incorrect extention */

	  if(width==32)
	    for(i=0;i<elements;i++)
	      mptr[i]=(-2147483647L)+uptr[i]-1; /* gcc warning bug */
	  else{
	    long c=-(1<<(width-1));
	    unsigned long mask=0x7fffffffL>>(31-width);
	    for(i=0;i<elements;i++)mptr[i]=c+(uptr[i]&mask);
	  }
	}

	/* scale the data if necessary */
	if(scale){
	  if(ptr->dynrange!=width)
	    if(width<ptr->dynrange)
	      for(i=0;i<elements;i++)mptr[i]<<=(ptr->dynrange-width);
	    else{
	      if(__ogg_precision==0){
		/* round properly */
		long offset=(1<<(width-ptr->dynrange-1))-1;
		for(i=0;i<elements;i++)mptr[i]+=offset;
	      }
	      for(i=0;i<elements;i++)
		mptr[i]>>=(width-ptr->dynrange);
	    }
	}
      }

      /* Append the new mbuffer to the old */
      ptr->ptr=__ogg_copyMB(ptr->ptr,m);
      __ogg_freeMB(m);
      __ogg_ripmoveMB(ptr->ptr,OM_MACHINE);
    }
    break;
  default:
    Verror(16);
  }
}

void __ogg_input_bit(int group,int name,char *data,long bytes){
  switch(group){
  case 2:
    if(!__ogg_buffer_used(group,name))
      Verror(15);
    else{
      bitb *ptr=__ogg_bitin[name];

      /* Append the new temp buffer to the old */
      ptr->ptr=__ogg_realloc(ptr->ptr,ptr->bytes+bytes,OM_MACHINE);
      memcpy(ptr->ptr+ptr->bytes,data,bytes);
      ptr->bytes+=bytes;
    }
    break;
  default:
    Verror(16);
  }
}

void __ogg_output_dsp(int group,int name,void **data,long *el,
		     int width,int endian,int sign,int scale){
  if(width<2)Verror(18);
  switch(endian){
  case -1:
    endian=bigendianp();
    break;
  case 0:case 1:
    break;
  default:
    endian=!bigendianp();
    break;
  }
  switch(group){
  case 0:
  case 1:
    if(!__ogg_buffer_used(group,name))
      Verror(15);
    else{
      dspb *ptr;
      int ch;
      long elements;
      switch(group){
      case 0:
	ptr=__ogg_dspin[name];
	break;
      case 1:
	ptr=__ogg_dspout[name];
	break;
      }
      *el=(elements=__ogg_sizeMB(ptr->ptr))*ptr->channels;
      /* Just align for size32; no types should cause hiccups for that */
      /* Straight alloc; not in a ripcord group, no header */
      *data=__ogg_malloc((*el)*(((width-1)>>3)+1),NOGROUP);
      
      /* Move the data into the buffer */
      for(ch=0;ch<ptr->channels;ch++){
	long *mptr=__ogg_ptrMB(ptr->ptr,ch);
	int i,j;
	/* We want to be doing math on longs, not chars... */
	/* Assume that if we're handling an odd bit size (like 20)
           that it will be in an integer number of bytes (20 in 24
           bits or 3 bytes) */
	
	/* scale the data if necessary */
	if(scale){
	  if(ptr->dynrange!=width)
	    if(width>ptr->dynrange)
	      for(i=0;i<elements;i++)mptr[i]<<=(width-ptr->dynrange);
	    else{
	      if(__ogg_precision==0){
		/* round properly */
		long offset=(1<<(ptr->dynrange-width-1))-1;
		for(i=0;i<elements;i++)mptr[i]+=offset;
	      }
	      for(i=0;i<elements;i++)
		mptr[i]>>=(ptr->dynrange-width);
	    }
	}

	/* Now limit and unsign the value if requested */
	{
	  long c=-(1<<(width-1));
	  long d=(1<<(width-1))-1;
	  for(i=0;i<elements;i++){
	    if(mptr[i]<c)
	      mptr[i]=c;
	    else
	      if(mptr[i]>d)mptr[i]=d;
	  }
	}

	if(!sign){
	  unsigned long *uptr=(unsigned long *)mptr;
	  if(width==32)
	    for(i=0;i<elements;i++)
	      mptr[i]=2147483648UL+uptr[i];
	  else{
	    long c=(1<<(width-1));
	    for(i=0;i<elements;i++)mptr[i]+=c;
	  }
	}

	/* bitpad AIFC style if fractional byte output */

	if(width%8){
	  int pad=8-(width%8);
	  for(i=0;i<elements;i++)
	    mptr[i]<<=pad;
	}

	/* Now move the mbuffer numbers into the output array */
	switch((width-1)/8){
	case 0:
	  /* 1 byte wide */
	  {
	    signed char *tdata=(signed char *)(*data);
	    j=0;
	    for(i=ch;i<*el;i+=ptr->channels)tdata[i]=(signed char)mptr[j++];
	  }
	  break;
	case 1:
	  /* 2 bytes wide */
	  {
	    size16 *tdata=(size16 *)(*data);
	    if(bigendianp()!=endian)
	      for(i=0;i<elements;i++)mptr[i]=(short)(((mptr[i]&0xff)<<8)|
						     ((mptr[i]>>8)&0xff));
	    j=0;
	    for(i=ch;i<*el;i+=ptr->channels)tdata[i]=(size16)mptr[j++];
	  }	  
	  break;
	case 2:
	  /* 3 bytes wide */
	  {
	    signed char *tdata=(signed char *)(*data);
	    j=0;
	    if(endian)
	      for(i=(ch*3);i<((*el)*3);i+=(3*ptr->channels)){
		tdata[i]=(char)mptr[j]>>16;
		tdata[i+1]=(char)mptr[j]>>8;
		tdata[i+2]=(char)mptr[j++];
	      }
	    else
	      for(i=(ch*3);i<((*el)*3);i+=(3*ptr->channels)){
		tdata[i]=(char)mptr[j];
		tdata[i+1]=(char)mptr[j]>>8;
		tdata[i+2]=(char)mptr[j++]>>16;
	      }		
	  }	  
	  break;
	case 3:
	  /* 4 bytes wide */
	  {
	    size32 *tdata=(size32 *)(*data);
	    if(bigendianp()!=endian)
	      for(i=0;i<elements;i++)mptr[i]=(size32)(((mptr[i]&0xff)<<24)|
						      ((mptr[i]&0xff00)<<8)|
						      ((mptr[i]>>8)&0xff00)|
						      ((mptr[i]>>24)&0xff));
	    j=0;  
	    for(i=ch;i<(*el);i+=ptr->channels)tdata[i]=mptr[j++];
	  }
	  break;
	default:
	  Verror(18);
	}
      }
    }
    break;
  default:
    Verror(16);
  }
}

void __ogg_output_bit(int group,int name,char **data,long *el){
  switch(group){
  case 2:
  case 3:
    if(!__ogg_buffer_used(group,name))
      Verror(15);
    else{
      bitb *ptr;
      switch(group){
      case 2:
	ptr=__ogg_bitin[name];
	break;
      case 3:
	ptr=__ogg_bitout[name];
	break;
      }
      *el=ptr->bytes;
      *data=__ogg_malloc(*el,NOGROUP);
      
      /* Move the data into the buffer */
      memcpy(*data,ptr->ptr,ptr->bytes);
    }
    break;
  default:
    Verror(16);
  }
}

void __ogg_run_clear(void){
  /* clear all the pipes, bpipes and set vars to -1 */
  /* also call the soft reset arm of each module;
     some (like A,dicts and variables) need to reset to their 
     original state */
  int i;

  for(i=0;i<__ogg_bitinN;i++)
    if(__ogg_bitin[i])
      if(!__ogg_bitin[i]->seekable)Verror(14);
  for(i=0;i<__ogg_bitoutN;i++)
    if(__ogg_bitout[i])
      if(!__ogg_bitout[i]->seekable)Verror(14);

  for(i=0;i<__ogg_pipes;i++)__ogg_resetP(i);
  for(i=0;i<__ogg_mvars;i++)__ogg_writetoV(i,-1);
  for(i=0;i<__ogg_bpipes;i++)__ogg_Breset(i);
  
  for(MC=0;MC<modulenum;MC++)
    if(modulelist[MC])
      if(modulelist[MC]->softreset)
	(*(modulelist[MC]->softreset))(modulelist[MC]);
  __ogg_ripfree(OM_TEMP);
  ogg_run_ready=-1;
}

#ifdef SELFTESTS

dspb *selftestptr;

void signedtest(void *data,void *data2,int width,int el,long *test,int endian){
  long *test2;
  int i;
  char *dataout;
  long elout;
  
  selftestptr->dynrange=width;
  selftestptr->channels=2;
  __ogg_input_reset(0,0);
  __ogg_input_dsp(0,0,data,el,width,endian,-1,0);
  test2=__ogg_ptrMB(selftestptr->ptr,0);
  for(i=0;i<el;i+=2)
    if(test[i]!=test2[i>>1]){
      fprintf(stderr," INPUT FAILED! (%ld != %ld)\n\n",test[i],test2[i>>1]);
      exit(1);
    }
  test2=__ogg_ptrMB(selftestptr->ptr,1);
  for(i=1;i<el;i+=2)
    if(test[i]!=test2[i>>1]){
      fprintf(stderr," INPUT FAILED! (%ld != %ld)\n\n",test[i],test2[i>>1]);
      exit(1);
    }
  __ogg_output_dsp(0,0,(void *)(&dataout),&elout,width,endian,-1,0);
  for(i=0;i<elout;i++)
    if(((char *)data2)[i]!=dataout[i]){
      fprintf(stderr," OUTPUT FAILED! (%d != %ld)\n\n",(int)(((char *)data)[i]),
	      (char)(dataout[i]));
      exit(1);
    }
  free(dataout);
  
  fprintf(stderr," passed.");
}

void unsignedtest(void *data,void *data2,int width,int el,long *test,int endian){
  long *test2;
  int i;
  char *dataout;
  long elout;
  
  selftestptr->dynrange=width;
  selftestptr->channels=2;
  __ogg_input_reset(0,0);
  __ogg_input_dsp(0,0,data,el,width,endian,0,0);
  test2=__ogg_ptrMB(selftestptr->ptr,0);
  for(i=0;i<el;i+=2)
    if(test[i]!=test2[i>>1]){
      fprintf(stderr," INPUT FAILED! (%ld != %ld)\n\n",test[i],test2[i>>1]);
      exit(1);
    }
  test2=__ogg_ptrMB(selftestptr->ptr,1);
  for(i=1;i<el;i+=2)
    if(test[i]!=test2[i>>1]){
      fprintf(stderr," INPUT FAILED! (%ld != %ld)\n\n",test[i],test2[i>>1]);
      exit(1);
    }
  __ogg_output_dsp(0,0,(void *)(&dataout),&elout,width,endian,0,0);
  for(i=0;i<elout;i++)
    if(((char *)data2)[i]!=dataout[i]){
      fprintf(stderr," OUTPUT FAILED! (%d != %ld)\n\n",(int)(((char *)data)[i]),
	      (char)(dataout[i]));
      exit(1);
    }
  free(dataout);
  
  fprintf(stderr," passed.");
}

int main(void){
  /* The buffer endianness/signedness conversions do a lot of subtle
     tricks with ANSI C.... we should make sure that they're all
     working properly with whatever compiler actually built us. */

  fprintf(stderr,"\nEndianness / type / sign input conversions: \ttesting...");
  if(bigendianp())
    fprintf(stderr,"\n\tThis machine is big endian...");
  else
    fprintf(stderr,"\n\tThis machine is little endian...");

  selftestptr=__ogg_registerDSP(&__ogg_dspin,&__ogg_dspinN,0);
  selftestptr->channels=1;

  /* first all signed */
  /* <1 byte */
  fprintf(stderr,"\n\tTesting <8 bits (signed)...\t\t\t");
  {
    char in[]={'\000','\010','\370','\377','\200','\100','\203','\003'};
    char in2[]={'\000','\010','\370','\370','\200','\100','\200','\000'};
    long test[]={0,1,-1,-1,-16,8,-16,0};
    signedtest(in,in2,5,8,test,0);
  }    

  /* 1 byte */
  fprintf(stderr,"\n\tTesting 8 bits (signed)...\t\t\t");
  {
    char in[]={'\000','\001','\077','\377','\040','\010','\340','\200'};
    long test[]={0,1,63,-1,32,8,-32,-128};
    signedtest(in,in,8,8,test,0);
  }    

  /* 1.n byte */
  fprintf(stderr,"\n\tTesting 8<n<16 %s endian (signed)...\t",bigendianp()?
	  "   big":"little");
  {
    size16 in[]={0,0x10,0xfff0,0xffff,0x8000,0x1001,0x800f,0x000f};
    size16 in2[]={0,0x10,0xfff0,0xfff0,0x8000,0x1001,0x800f,0x0000};
    long test[]={0,1,-1,-1,-2048,256,-2048,0};
    signedtest(in,in2,12,8,test,bigendianp());
  }    
  fprintf(stderr,"\n\tTesting 8<n<16 %s endian (signed)...\t",!bigendianp()?
	  "   big":"little");
  {
    size16 in[]={0,0x2000,0xf0ff,0xffff,0x0080,0x0110,0xf80,0x0f00};
    size16 in2[]={0,0x2000,0xf0ff,0xf0ff,0x0080,0x0110,0xf80,0};
    long test[]={0,2,-1,-1,-2048,256,-2048,0};
    signedtest(in,in2,12,8,test,!bigendianp());
  }    

  /* 2 byte */
  fprintf(stderr,"\n\tTesting 16 bit %s endian (signed)...\t",bigendianp()?
	  "   big":"little");
  {
    size16 in[]={0,1,0x0fff,0xffff,0x4000,0x0100,0x8000,0x7fff};
    long test[]={0,1,4095,-1,16384,256,-32768,32767};
    signedtest(in,in,16,8,test,bigendianp());
  }    
  fprintf(stderr,"\n\tTesting 16 bit %s endian (signed)...\t",!bigendianp()?
	  "   big":"little");
  {
    size16 in[]={0,0x0200,0xff0f,0xffff,0x0040,0x0001,0x0080,0xff7f};
    long test[]={0,2,4095,-1,16384,256,-32768,32767};
    signedtest(in,in,16,8,test,!bigendianp());
  }    

  /* 2.n byte */
  fprintf(stderr,"\n\tTesting 16<n<24 %s endian (signed)...\t","little");
  {
    char in[]={0,0,0,0x40,0,0,0x4f,0x20,0x80,0xe0,0xff,0x7f};
    char in2[]={0,0,0,0x40,0,0,0x40,0x20,0x80,0xe0,0xff,0x7f};
    long test[]={0,2,-261886,262143};
    signedtest(in,in2,19,4,test,0);
  }    
  fprintf(stderr,"\n\tTesting 16<n<24 %s endian (signed)...\t","   big");
  {
    char in[]={0,0,0,0,0,0x20,0x80,0x20,0x4f,0x7f,0xff,0xe0};
    char in2[]={0,0,0,0,0,0x20,0x80,0x20,0x40,0x7f,0xff,0xe0};
    long test[]={0,1,-261886,262143};
    signedtest(in,in2,19,4,test,1);
  }    

  /* 3 byte */
  fprintf(stderr,"\n\tTesting 24 bit %s endian (signed)...\t","little");
  {
    char in[]="\000\000\000""\002\000\000""\002\001\200""\377\377\177";
    long test[]={0,2,-8388350,8388607};
    signedtest(in,in,24,4,test,0);
  }    
  fprintf(stderr,"\n\tTesting 24 bit %s endian (signed)...\t","   big");
  {
    char in[]="\000\000\000""\000\000\001""\200\001\002""\177\377\377";
    long test[]={0,1,-8388350,8388607};
    signedtest(in,in,24,4,test,1);
  }    

  /* 3.n byte */
  fprintf(stderr,"\n\tTesting 24<n<32 %s endian (signed)...\t",bigendianp()?
	  "   big":"little");
  {
    size32 in[]={0,0x10,0xfffffff0,0xffffffff,0x40000000,0x10203040,
		 0x8000000f,0x7ffffff0};
    size32 in2[]={0,0x10,0xfffffff0,0xfffffff0,0x40000000,0x10203040,
		 0x80000000,0x7ffffff0};
    long test[]={0,1,-1,-1,67108864,16909060,-134217728,134217727};
    signedtest(in,in2,28,8,test,bigendianp());
  }    
  fprintf(stderr,"\n\tTesting 24<n<32 %s endian (signed)...\t",!bigendianp()?
	  "   big":"little");
  {
    size32 in[]={0,0x20000000,0xf0ffffff,0xffffffff,0x00000040,0x40302010,
		 0x00000080,0xf0ffff7f};
    size32 in2[]={0,0x20000000,0xf0ffffff,0xf0ffffff,0x00000040,0x40302010,
		 0x00000080,0xf0ffff7f};
    long test[]={0,2,-1,-1,67108864,16909060,-134217728,134217727};
    signedtest(in,in2,28,8,test,!bigendianp());
  }    

  /* 4 byte */
  fprintf(stderr,"\n\tTesting 32 bit %s endian (signed)...\t",bigendianp()?
	  "   big":"little");
  {
    size32 in[]={0,1,0x0fffffff,0xffffffff,0x04000000,0x01020304,
		 0x80000001,0x7fffffff};
    long test[]={0,1,268435455,-1,67108864,16909060,-2147483647L,2147483647L};
    signedtest(in,in,32,8,test,bigendianp());
  }    
  fprintf(stderr,"\n\tTesting 32 bit %s endian (signed)...\t",!bigendianp()?
	  "   big":"little");
  {
    size32 in[]={0,0x02000000,0xffffff0f,0xffffffff,0x00000004,0x04030201,
		 0x01000080,0xffffff7f};
    long test[]={0,2,268435455,-1,67108864,16909060,-2147483647L,2147483647L};
    signedtest(in,in,32,8,test,!bigendianp());
  }    

  /* now unsigned */
  /* 16 bit */
  fprintf(stderr,"\n\tTesting <32 bit unsigned...\t\t\t");
  {
    size16 in[]={0,1,0x0fff,0xffff,0x4000,0x0100,0x8000,0x7fff};
    long test[]={-32768,-32767,4095-32768,32767,16384-32768,256-32768,0,-1};
    unsignedtest(in,in,16,8,test,bigendianp());
  }    
  /* 32 bit */
#define MIN32 2147483647-1 /* gcc warning bug */
  fprintf(stderr,"\n\tTesting 32 bit unsigned...\t\t\t");
  {
    unsigned size32 in[]={0x02000000,0xffffff0f,0xffffffff,0x00000004,
			  0x04030201,0x01000080,0xffffff7f,0xffffff7f};
    long test[]={2-MIN32,268435455-MIN32,-(1-MIN32),67108864-MIN32,
		 16909060-MIN32,1,-1,-1};
    unsignedtest(in,in,32,8,test,!bigendianp());
  }    

  /* test mbuffer append */
  selftestptr->channels=1;
  __ogg_input_reset(0,0);
  fprintf(stderr,"\n\tTesting append buffering...\t\t\t");
  {
    size16 in[]={0,1,0x0fff,0xffff,0x4000,0x0100,0x8000,0x7fff};
    selftestptr->dynrange=16;
    __ogg_input_dsp(0,0,(void *)in,8,16,bigendianp(),0,0);
  }
  {
    long *test2,i;
    size16 in[]={0,2,0x0fff,0xffff,0x4000,0x0100,0x8000,0x7fff};
    long test[]={-32768,-32767,4095-32768,32767,16384-32768,256-32768,0,-1,
		 -32768,-32766,4095-32768,32767,16384-32768,256-32768,0,-1};
    __ogg_input_dsp(0,0,(void *)in,8,16,bigendianp(),0,0);
    test2=__ogg_ptrMB(selftestptr->ptr,0);
    for(i=0;i<__ogg_sizeMB(selftestptr->ptr);i++)
      if(test[i]!=test2[i]){
	fprintf(stderr," FAILED! (%ld != %ld)\n\n",test[i],test2[i]);
	exit(1);
      }
    fprintf(stderr," passed.");
  }    

  /* test scaling */
  __ogg_input_reset(0,0);
  fprintf(stderr,"\n\tTesting integer scaling...\t\t\t");
  {
    long *test2,i;
    size16 in[]={0,1,2,3,4,5,6,7,8,9,10,0,-1,-2,-3,-4,-5,-6,-7,-8,-9,-10};
    long test[]={0,0,0,1,1,1,1,2,2,2,2,0,0,-1,-1,-1,-1,-2,-2,-2,-2,-3};
    selftestptr->dynrange=14;
    __ogg_input_dsp(0,0,(void *)in,22,16,bigendianp(),1,1);
    test2=__ogg_ptrMB(selftestptr->ptr,0);
    for(i=0;i<__ogg_sizeMB(selftestptr->ptr);i++)
      if(test[i]!=test2[i]){
	fprintf(stderr," FAILED! (%ld != %ld)\n\n",test[i],test2[i]);
	exit(1);
      }
    fprintf(stderr," passed.");
  }    

  fprintf(stderr,"\nEndian / type / sign tests pass.\n");
  return(0);
}

#endif
